#ifndef _DARGUMENTS_H
#define _DARGUMENTS_H

// This file was automatically generated by objch.

//==============================================================================
//
//       DArguments - the program arguments parser in the ofc-library
//
//             Copyright (C) 2003-2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:58 $ $Revision: 1.15 $
//
//==============================================================================


#include "ofc/DList.h"
  
//
// Structure for adding a group of program options
//

typedef struct
{
  char           *longOption;     // the long option incl. parameter (date or date=DATE)
  char            shortOption;    // the short option (d)
  char           *description;    // the help text
  id              target;         // the target object (reference)
} DArgOption;

//
// The DArguments class implements methods for parsing the arguments of a
// program. After adding the definition of the arguments, the arguments
// are parsed. The results are stored in boolean objects, parsable objects and a list of 
// texts. <br>
// Some notes: <br><ul>
// <li>Default options: -?/--help (show the help text) --version (show the version) -- 
// (stop parsing)
// <li>The target object for switches (options without parameter) must be the class DBool
// (or inherite from DBool). The target objects for parameter options must comply to
// the protocol DParsable.
// <li>The help format looks like the help text from the gnu tools (with column width=80)
// <li>The parameter for an option is required; optional parameters are not supported.
// <ul>
//
// @example
//
// #include <stdio.h>
// #include "ofc/DArguments.h"
// #include "ofc/DBool.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DArguments *args    = [DArguments new]; // the argument parser
// 
//   DBool      *verbose = [DBool new];     // the argument destination objects
//   DText      *date    = [DText new];
// 
//   DList      *extra;                     // the list with non options
// 
//                                          // add the -d or --date option
//   [args option :"date=DATE" :'d' :"change default data" :date];
//                                          // add the -v or --verbose switch
//   [args option :"verbose"   :'v' :"be verbose"          :verbose];
// 
//                                          // parse the arguments
//   extra = [args parse :"example"
//                       :"Usage: example [OPTION] ... [FILES]"
//                       :"example v1.7\n\n(c) 2008 Example soft"
//                       :"Report bugs to bugs@example.com"
//                       :argv
//                       :argc];
//                                          // Process the result
//   if ([verbose get])
//     printf("Verbose!!\n");
//   if ([date length] > 0)
//     printf("Date=%s\n", [date cstring]);
// 
//   printf("There are %ld non-options\n", (extra != nil ? [extra length] : 0));
// 
//   [args      free];                      // Cleanup
// 
//   [verbose   free];
//   [date      free];
//   [extra     free];
// 
//   return 0;
// }
// 

@interface DArguments : Object
{
@private
  DList *_options;      // the options
  int    _longColumn;   // the lenght of column with long options
}


//// Constructors

//
// Initialise an empty argument parser
//
// @return the object
//
- (DArguments *) init;


//// Copy related methods

//
// Do a shallow copy of the parser (not implemented)
//
// @return the object
//
- shallowCopy;


//// Destructor

//
// Free the parser
//
// @return the object
//
- free;


//// Option definition methods

//
// Add an option
// 
// @param longOption  the long option description (date or date=DATE)
// @param shortOption the short option description (d)
// @param description the help text for the option
// @param target      the target object (reference)
//
// @return the object
//
- (DArguments *) option :(const char *) longOption
                        :(char)         shortOption 
                        :(const char *) description 
                        :(id)           target;

//
// Add an array of options
//  
// @param options   the array with options
// @param number    the number of elements in options
//
// @return the object
//
- (DArguments *) options :(const DArgOption *) options :(int) number;


//// Main method

//
// Parse the program arguments, leave the result in the target objects and the non-option list
//
// @param name     the program name
// @param usage    the usage text (or NULL)
// @param version  the version text (or NULL)
// @param tail     the tail text (or NULL)
// @param argv     the arguments array
// @param argc     the number of elements in argv
//
// @return the (new) list with all the non-option arguments (or nil if not found)
//
- (DList *) parse :(const char *) name 
                  :(const char *) usage 
                  :(const char *) version 
                  :(const char *) tail
                  :(char **)      argv 
                  :(int)          argc;


//// Help methods

//
// Print the version on stdout
// 
// @param version  the version text
// 
// @return the object
// 
- (DArguments *) printVersion :(const char *) version;

//
// Print the option help on stdout
// 
// @param shortOption   the short option (or EOS)
// @param longColumn    the length of the long column
// @param longOption    the long option (or NULL)
// @param description   the description (or NULL)
// 
// @return none
//
- (void) printOptionHelp :(char) shortOption
                         :(int) longColumn
                         :(const char *) longOption
                         :(const char *) description;

//
// Print the standard help on stdout
//
// @param usage    the usage text
// @param tail     the tail text
//
// @return the object
//
- (DArguments *) printHelp :(const char *) usage :(const char *) tail;

@end
#endif
