#ifndef _DARRAY_H
#define _DARRAY_H

// This file was automatically generated by objch.

//==============================================================================
//
//        DArray - the dynamic array collection in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:58 $ $Revision: 1.11 $
//
//==============================================================================

#include "ofc/config.h"

// The array collection stores objects in an array that can grow. Objects
// can be accessed by an index. Due to the easy access there is no iterator.
//
// @example
//
// #include "ofc/DArray.h"
// #include "ofc/DText.h"
// #include "ofc/DInt.h"
// 
// int main(int argc, char *argv[])
// {
//   DArray *array = [DArray alloc];
//   DText  *str;
//   DInt   *nr;
//   id      obj;
// 
//   [array init : 10];      // Start with length 10
// 
//                           // Put objects in the array
//   str = [DText new]; [str set :"String1"];
//   [array set :0 :str];
// 
//   str = [DText new]; [str set :"String2"];
//   [array set :1 :str];
// 
//   nr = [DInt alloc]; [nr init :3];
//   [array set :2 :nr];
// 
//   nr = [DInt alloc]; [nr init :4];
//   [array set :3 :nr];
// 
//                           // Get objects from the array
//   obj = [array get :4];   // id = nil
// 
//   obj = [array get :1];   // id = DText("String2")
//   if ([obj isKindOf :[DText class]])
//     printf("Text:%s\n", [obj cstring]);
// 
//   if ([obj isKindOf :[DInt class]])
//     printf("Number:%d\n", [obj get]);
// 
//   [array free];           // Free the array and the stored objects
// 
//   return 0;
// }
// 

@interface DArray : Object
{
@private
  id            *_objects;  // the stored objects
  long           _length;   // the length of the array
}

//// Constructors

//
// Initialise to an empty array
//
// @return the object
//
- (DArray *) init;

//
// Initialise to an array with an initial length
//
// @param length   the length of the array
//
// @return the object
//
- (DArray *) init :(long) length;


//// Copy related methods

//
// Do a shallow copy of the array
// 
// @return the object
// 
- shallowCopy;

//
// Deepen a copy of the array
//
// @return the object
//
- deepen;


//// Destructor

//
// Free the array, not the stored objects
//
// @return the object
//
- shallowFree;

//
// (Deep) free the array (incl. the objects)
//
// @return the object
//
- free;


//// Member methods

//
// Return the length of the array
//
// @return the length
//
- (long) length;

//
// Set the length for the array
//
// @param length   the new length 
//
// @return the object
//
- (DArray *) length :(long) length;

//
// Test if an index is valid
//
// @param index   the index to be tested
//
// @return is it valid ?
//
- (BOOL) isValid :(long) index;


//// Main methods

//
// Set an object in the array
//
// @param index    the index in the array
// @param obj      the object to be set
//
// @return the previous object in the array (or nil)
//
- (id) set :(long) index :(id) obj;

//
// Get an object from the array
//
// @param index    the index in the array
//
// @return the object in the array (or nil)
//
- (id) get :(long) index;


//// Object location methods

//
// Find the first occurence of an object
//
// @param obj      the object to be found
//
// @return the first occurence of the object (or -1 for not found)
//
- (long) index :(id) obj;

//
// Find the last occurence of an object
//
// @param obj      the object to be found
//
// @return the last occurence of the object (or -1 for not found)
//
- (long) rindex :(id) obj;

//
// Count the number of occurences of an object
//
// @param obj      the object to be counted
//
// @return the number of occurences
//
- (long) count :(id) obj;

//
// Test if an object is in an array
//
// @param obj      the object to be checked
//
// @return is it in the array ?
//
- (BOOL) has :(id) obj;

//
// Perform a selector on each object in the array
//
// @param sel      the selector
//
// @return the object
//
- (DArray *) each :(SEL) sel;

@end
#endif
