#ifndef _DAVLTREE_H
#define _DAVLTREE_H

// This file was automatically generated by objch.

//==============================================================================
//
//        DAvlTree - the avl tree collection class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:58 $ $Revision: 1.9 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DComparable.h"

#include "ofc/DList.h"

typedef struct _DAvlNode
{
  id                   _key;
  id                   _object;
  struct _DAvlNode    *_left;
  struct _DAvlNode    *_right;
  struct _DAvlNode    *_parent;
  short                _balance;
  short                _trace;
} DAvlNode;


// The avl tree collection class implements a number of methods for creating, 
// inserting, searching and removing (elements in) an avl tree. The implementation is 
// inspired by the book "Algorithms & Data structures" by N. Wirth.
//
// @example
//
// #include <stdio.h>
// #include "ofc/DAvlTree.h"
// 
// int main(int argc, char *argv[])
// {
//   DAvlTree *tree = [DAvlTree alloc];
//   DText    *key  = [DText    new  ];
//   DText    *val;
// 
//   [tree init :[DText class]];            // Init tree, key is a DText object
// 
//   [key set :"Key1"];                     // Set key1 and create value1 for tree
//   val = [DText new]; [val set :"Value1"];
// 
//   [tree insert :key :val];
// 
//   [key set :"Key2"];                     // Set key1 and create value1 for tree
//   val = [DText new]; [val set :"Value2"];
// 
//   [tree insert :key :val];               // Create key2 and value2 for tree
// 
//   printf("Number of elements:%ld\n", [tree length]);  // Print tree length
// 
//   [key set :"Key2"];
//                                          // Check key present
//   printf("%s is %s present in tree\n", [key cstring], ([tree has :key] ? "" : "not"));
// 
//   val = [tree get :key];
//   printf("%s has value %s\n", [key cstring], (val != nil ? [val cstring] : "???"));
// 
//   [tree free];                           // Free tree and stored objects
//   [key  free];
// 
//   return 0;
// }
// 


@interface DAvlTree : Object
{
@private
  DAvlNode     *_root;    // the root of the avl-tree
  Class         _class;   // the class of the key
  long          _count;   // the number of nodes in the avl tree
}

#if 0
#endif

//// Constructors

//
// Initialise to an unitialised avl tree (not implemented)
//
// @return the object
//
- (DAvlTree *) init;

//
// Initialise to an empty avl tree
//
// @param key       the class of the key
//
// @return the object
//
- (DAvlTree *) init :(Class) key;


//// Copy related methods

//
// Do a shallow copy of the avl tree
// 
// @return the object
// 
- shallowCopy;

//
// Deepen a copied avl tree
//
// @return the object
//
- deepen;


//// Deconstructors

//
// Free the avl tree (keys are freed, objects are *NOT* freed)
//
// @return the object
//
- shallowFree;

//
// (Deep) free the avl tree (keys are freed, objects are *FREED*)
//
// @return the object
//
- free;


//// Member methods

//
// Return the number of objects in the tree
//
// @return the length
//
- (long) length;

//
// Check if the avl tree is empty
//
// @return is it?
//
- (BOOL) isEmpty;


//// Avl tree manipulation

//
// Insert a key,object in the avl tree
//
// @param key      the key to be inserted
// @param object   the object to be inserted
//
// @return success
//
- (BOOL) insert :(id) key :(id) object;

//
// Remove a key,object from the avl tree
//
// @param key      the key to be removed
//
// @return the object that is removed (or nil for not found)
//
- (id) delete :(id) key;

//
// Check if the avl tree contains a key
//
// @param key      the key to be checked
//
// @return is the key present
//
- (BOOL) has :(id) key;

//
// Return the object related to a key
//
// @param key      the key 
//
// @return the object related to the key (or nil)
//
- get :(id) key;


//// List related functions

//
// Return a list with copies of all keys (sorted) in the avl tree
//
// @return a (new) sorted list with all the keys
//
- (DList *) keys;

//
// Return a list with copies of all keys (reverse sorted) in the avl tree
//
// @return a (new) reverse sorted list with all the keys
//
- (DList *) rkeys;

//
// Return a list with copies of all the objects in the avl tree
//
// @return a (new) list with all the objects
//
- (DList *) objects;

#if 0
//
// Show the tree
//
- (void) show;

//
// Count the nodes
//
// @return the number of nodes
//
- (int) count;

//
// Check the parent pointers
//
// @return the number of wrong parent pointers
//
- (int) check;

#endif
@end

//
// The DAvlIterator class contains a number of methods for iterating a avl tree.
//
// @example
// #include <stdio.h>
// #include "ofc/DAvlTree.h"
//
// int main(int argc, char *argv[])
// {
//   DAvlTree     *tree = [DAvlTree     alloc];
//   DAvlIterator *iter = [DAvlIterator alloc];
//
//   DText        *key;
//   DText        *value;
//
//   [tree init :[DText class]];            // Init tree, key is a DText object
//
//   // Insert DText keys and values in the tree ...
//
//   [iter init :tree];                     // Init the iterator with the tree
//
//   key = [iter key];
//   while (key != nil)                     // Iterate the tree
//   {
//     val = [iter object];
//
//     [iter next];
//     key = [iter key];
//   }
//
//   [iter free];                           // Cleanup
//   [tree free];
//   [key  free];
//
//   return 0;
// }
//

@interface DAvlIterator : Object
{
@private
  DAvlTree       *_tree;    // the tree on which the iterator operates
  DAvlNode       *_node;    // the current node in the list
}


//// the constructors

//
// Initialises an avl tree iterator without a tree
//
// @return the iterator
//
- (DAvlIterator *) init;

//
// Initialises an iterator on a tree (the iterator is moved to the first node)
//
// @param tree    the tree on which the iterator operates
//
// @return the iterator
//
- (DAvlIterator *) init :(DAvlTree *) tree;


//// the iterator methods

//
// Set the avl tree (iterator is moved to the first object)
//
// @return the object
//
- (DAvlIterator *) tree :(DAvlTree *) tree;

//
// Get the first object in the tree
//
// @return a reference to the first object (or nil)
//
- (id) first;

//
// Get the next object in the tree
//
// @return the next object (or nil)
//
- (id) next;

//
// Get the previous object in the tree
//
// @return the previous object (or nil)
//
- (id) prev;

//
// Get the last object in the tree
//
// @return the last object (or nil)
//
- (id) last;

//
// Set the current object to another object
//
// @param object   the object to be placed in current node
//
// @return the previous stored object
//
- (id) object :(id) object;

//
// Get the current object
//
// @return the current object (or nil)
//
- (id) object;

//
// Get the key of the current object
//
// @return a referencee to the key of the current object (or nil)
//
- (id) key;

@end
#endif
