#ifndef _DBZIPFILE_H
#define _DBZIPFILE_H

// This file was automatically generated by objch.

//==============================================================================
//
//           DBZipFile - the bzip2 file class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-24 05:37:06 $ $Revision: 1.11 $
//
//==============================================================================
#ifdef HAVE_DBZIPFILE
#endif

#include "ofc/config.h"

#include "ofc/DText.h"
#include "ofc/DData.h"
#include "ofc/DList.h"
#include "ofc/DTextReadable.h"
#include "ofc/DTextWritable.h"
#include "ofc/DDataReadable.h"
#include "ofc/DDataWritable.h"

#if defined(HAVE_LIBBZ2) && defined(HAVE_BZLIB_H)
#define HAVE_DBZIPFILE 1
#endif

#ifdef HAVE_DBZIPFILE

//
// The DBZipFile class implements a number of methods for opening of, writing to, 
// reading from and closing of bzip2 files. See the open method for used settings for 
// the bzip2 library.
//
// @example
// #include <stdio.h>
// #include "ofc/DBZipFile.h"
// 
// int main(int argc, char *argv[])
// {
//   DBZipFile *file = [DBZipFile new];
//   DText     *str;
// 
//   char       name[] = "output.bz2";
// 
//   if ([file open :name :"w"])          // Open a bzipped text file for writing
//   {
//     if (![file writeLine :"A line full of text .."])
//       printf("\"%s\" could not be written: %d\n", name, [file error]);
// 
//     [file close];
//     
//     printf("\"%s\" succesfull written.\n", name);
//   }
//   else
//     printf("\"%s\" could not be opened: %d\n", name, [file error]);
//   
//   
//   if ([file open :name :"r"])          // Open a bzipped text file for reading
//   {
//     while (![file isEof])                // Read all lines
//     {
//       str = [file readLine];
// 
//       printf("%s\n", [str cstring]);
// 
//       [str free];
//     }
//     
//     [file close];
// 
//     printf("\"%s\" succesfull read.\n", name);
//   }
//   else
//     printf("\"%s\" could not be opened: %d\n", name, [file error]);
// 
//   [file free];
// 
//   return 0;
// }
// 

@interface DBZipFile : Object <DTextReadable,DTextWritable,DDataReadable,DDataWritable>
{
@private
  void           *_file;    // the file pointer
  void           *_bzfile;  // the bzip file pointer
  BOOL            _read;    // is the file opened for reading ?
  BOOL            _eof;     // is end of file reached
  int             _error;   // the reported error
}

#endif
#ifdef HAVE_DBZIPFILE

//// Constructors

//
// Initialise an empty file object
//
// @return the object
//
- (DBZipFile *) init;

//
// Open a bzip file with default level and strategy
//
// @param name     the filename (with extension .bz2)
// @param mode     the file mode (r,w)
//
// @return the object
//
- (DBZipFile *) init :(const char *) name :(const char *) mode;

//
// Open a bzip file
//
// @param name      the filename (with extension .bz2)
// @param mode      the file mode (r,w)
// @param small     should a small memory footprint be used
//
// @return the object
//
- (DBZipFile *) init :(const char *) name :(const char *) mode :(BOOL) small;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the file object
//
- free;


//// File info methods

//
// Check if the file is open
//
// @return is the file open ?
//
- (BOOL) isOpen;

//
// Return the last error
//
// @return the last error
//
- (int) error;


//// File opening methods

//
// Open a bzip file with big memory usage
//
// @param name     the filename (with extension .bz2)
// @param mode     the file mode (r,w)
//
// @return is the file open ?
//
- (BOOL) open :(const char *) name :(const char *) mode;

//
// Open a bzip file (note: the defaults for the bzip2 library: verbosity = 0 (silent), workFactor = 0,
// blockSize100k = 5 (for small) and 9 (for not small)
//
// @param name      the filename (with extension .bz2)
// @param mode      the file mode (r,w)
// @param small     should a small memory footprint be used ?
//
// @return is the file open ?
//
- (BOOL) open :(const char *) name :(const char *) mode :(BOOL) small;


//// TextReadable protocol implementation

//
// Read a text string (all available text)
//
// @return a (new) text string
//
- (DText *) readText;

//
// Read a text string 
//
// @param length   the length of the text to be read
//
// @return a (new) text string
//
- (DText *) readText :(long) length;

//
// Read a line of text (till Eof or \n)
//
// @return a (new) text string (or nil for no data)
//
- (DText *) readLine;;

//
// Read a character
//
// @return the character read (EOS for EOF)
//
- (char) readChar;;

//
// Tell the current position in the readable
//
// @return -1 (not implemented)
//
- (unsigned long) tell;;

//
// Move the current position in the readable
//
// @param offset   the offset from the origin (in bytes)
// @param origin   the origin for the offset 
//
// @return NO (not implemented)
//
- (BOOL) seek :(unsigned long) offset :(int) origin;

//
// Skip a number of positions
//
// @param offset   the number of bytes to skip
//
// @return NO (not implemented)
//
- (BOOL) skip :(unsigned long) offset;

//
// Check if the end of the bzip2 file is reached
//
// @return is it?
//
- (BOOL) isEof;


//// Implementation of DTextWritable

//
// Write text string 
//
// @param text     the text to be written
//
// @return success
//
- (BOOL) writeText :(const char *) text;

//
// Write line (appending a '\n')
//
// @param text     the text to be written
//
// @return success
//
- (BOOL) writeLine :(const char *) text;

#ifdef WIN32
#endif
//
// Write a character
//
// @param ch       the character to be written
//
// @return success
//
- (BOOL) writeChar :(char) ch;


//// Methods for DDataReadable protocol

//
// Read a data string 
//
// @param length   the length of the data
//
// @return a (new) data string object
//
- (DData *) readData :(unsigned long) length;

//
// Read a data string 
//
// @param dest     the destination buffer
// @param length   the length of the data (and minimum size of dest)
//
// @return the actual number of bytes read
//
- (unsigned long) readData :(unsigned char *) dest :(unsigned long) length;

//
// Read a byte
//
// @return the byte read (0 for error)
//
- (unsigned char) readByte;

//
// Read a short
//
// @return the short
//
- (short) readShort;

//
// Read a long
//
// @return the long
//
- (long) readLong;;

//
// Read a double
//
// @return the double
//
- (double) readDouble;;


//// DDataWritable protocol implementation

//
// Write a data string 
//
// @param data     the data to be written
// @param length   the length of the data
//
// @return success
//
- (BOOL) writeData :(const unsigned char *) data :(unsigned long) length;;

//
// Write a byte
//
// @param byte     the byte to be written
//
// @return success
//
- (BOOL) writeByte :(unsigned char) byte;;

//
// Write a short 
//
// @param nr       the short to be written
//
// @return success
//
- (BOOL) writeShort :(short) nr;;

//
// Write a long
//
// @param nr       the long to be written
//
// @return success
//
- (BOOL) writeLong :(long) nr;

//
// Write a double
//
// @param nr       the double to be written
//
// @return success
//
- (BOOL) writeDouble :(double) nr;


//// List related methods

//
// Read all lines from a file into a list of strings
//
// @return a (new) list of (new) DText objects
//
- (DList *) readLines;

//
// Writes all the elements ('lines') of text from a list to file
//
// @param list     the list with objects that responds to textable protocol
//
// @return success
//
- (BOOL) writeLines :(DList *) list;


//// File manipulation

//
// Flush the output buffers of the file 
//
// @return NO (not implemented for BZipFile)
//
- (BOOL) flush;


//// File closing methods

//
// Close the file
//
// @return the file object
//
- (DBZipFile *) close;

@end
#endif
#endif
