#ifndef _DBIGDOUBLE_H
#define _DBIGDOUBLE_H

// This file was automatically generated by objch.

//==============================================================================
//
//   DBigDouble - the arbitrary length double number class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:59 $ $Revision: 1.4 $
//
//==============================================================================
#ifdef HAVE_DBIGDOUBLE
#endif

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DParsable.h"
#include "ofc/DSimpleArithmetic.h"
#include "ofc/DComparable.h"
#include "ofc/DText.h"

#if defined(HAVE_LIBGMP) && defined(HAVE_GMP_H)
#define HAVE_DBIGDOUBLE  1
#endif

#ifdef HAVE_DBIGDOUBLE

//
// The big double class implements a number of methods for working with arbitrary large double 
// numbers. The class uses the excellent gmp library for the actual calculations. Note: setting
// the precision (as is done by move: and point:) will result in a realloc)
// 
// @example
//
// #include <stdio.h>
// #include "ofc/DBigDouble.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DBigDouble *dbl1 = [DBigDouble alloc];
//   DBigDouble *dbl2 = [DBigDouble new  ];
//   DBigDouble *dbl3 = [DBigDouble new  ];
//   DText      *str;
// 
//   [dbl1 init :3.14143123423234322343223432 :256]; // Init with pi, with 256 precision bits 
// 
//   [dbl2 move :dbl1];                     // Move dbl1 in dbl2
// 
//   [dbl2 mul  :dbl1];                     // Multiply dbl2 with dbl1 -> PI^2
// 
//   [dbl3 sub  :dbl2 :dbl1];               // dbl3 = dbl2 - dbl1 = PI^2 - PI
// 
//   printf("PI^2 - PI=%f\n", [dbl3 get]);  // Get dbl3 as double number
// 
//   str = [dbl3 get :10 :15];              // Convert to string, base 10, 15 digits
// 
//   printf("PI^2 - PI=%s\n", [str cstring]);
// 
//   if ([dbl1 compare :dbl2] == 0)         // Compare numbers
//     printf("Double1 is equal to double2\n");
//   else if ([dbl1 compare :dbl2] < 0)
//     printf("Double1 is smaller than double2\n");
//   else
//     printf("Double1 is greater than double2\n");
// 
//   [str  free];                           // Cleanup
//   [dbl3 free];
//   [dbl2 free];
//   [dbl1 free];
// 
//   return 0;
// }
// 


@interface DBigDouble : Object <DTextable,DParsable,DSimpleArithmetic,DComparable>
{
@private
  void        *_value;      // the big double value
}

#endif
#ifdef HAVE_DBIGDOUBLE

//// Constructors

//
// Initialise to double zero with default precision 
//
// @return the object
//
- (DBigDouble *) init;

//
// Initialise to double zero with a precision
//
// @param point    the number of bits in the precision
//
// @return the object
//
- (DBigDouble *) init :(unsigned long) point;

//
// Initialise to a double with a precision
//
// @param value  the double value
// @param point  the number of bits in the precision
//
// @return the object
//
- (DBigDouble *) init :(double) value :(unsigned long) point;

//
// Initialise with a string
// 
// @param cstring  the cstring with the double
// @param base     the base for the double
// @param point    the number of bits in the precision
// 
// @return the object
// 
- (DBigDouble *) init :(const char *) cstring :(int) base :(unsigned long) point;


//// Copy constructor

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free a big double
// 
// @return the object
//
- free;

// Member methods
//
// Check if the big double is negative
// 
// @return is it?
// 
- (BOOL) isNegative;

//
// Check if the big double is zero
// 
// @return is it ?
// 
- (BOOL) isZero;

//
// Get the current precision of the big double
// 
// @return the number of bits for the precision
// 
- (unsigned long) point;

//
// Set the precision for the big double
// 
// @param point    the new value for the precision (in bits)
// 
// @return the objectc
// 
- (DBigDouble *) point :(unsigned long) point;


//// Set methods

//
// Set the big double to zero
// 
// @return the object
// 
- (DBigDouble *) clear;

//
// Set the big double with a (small) double
//
// @param value   the (small) double
//
// @return the object
//
- (DBigDouble *) set :(double) value;

//
// Set the big double with a text number (format: MeN)
//
// @param cstring   the cstring with the double
// @param base      the base of the double (2..36)
//
// @return success
//
- (BOOL) set :(const char *) cstring :(int) base;

//
// Move an other object in this object (including the precision)
//
// @param other    the other object
//
// @return this object
//
- (DBigDouble *) move :(DBigDouble *) other;


//// Get methods

//
// Get the big double as (small) double
// 
// @return the (small) double
// 
- (double) get;

//
// Get the big double as a text string
// 
// @param base     the base for the text string (2..36)
// @param digits   the number of digits in the string (0 = all)
// 
// @return a (new) text string (or nil)
// 
- (DText *) get :(int) base :(unsigned) digits;


//// Methods on big double numbers

//
// Add with a big double 
//
// @param other    the big double
//
// @return the object
//
- (DBigDouble *) add :(DBigDouble *) other;

//
// Add two big doubles and store the result in the object
//
// @param src1     the first big double
// @param src2     the second big double
//
// @return the object
//
- (DBigDouble *) add :(DBigDouble *) src1 :(DBigDouble *) src2;

//
// Subract with a big double
//
// @param other     the big double
//
// @return the object
//
- (DBigDouble *) sub :(DBigDouble *) other;

//
// Subract two big doubles and store the result in the object
//
// @param src1      the first big double
// @param src2      the second big double
//
// @return the object
//
- (DBigDouble *) sub :(DBigDouble *) src1 :(DBigDouble *) src2;

//
// Multiply with a big double
//
// @param other    the big double
//
// @return the object
//
- (DBigDouble *) mul :(DBigDouble *) other;

//
// Multiply two big doubles and store the result in the object
//
// @param src1    the first big double
// @param src2    the second big double
//
// @return the object
//
- (DBigDouble *) mul :(DBigDouble *) src1 :(DBigDouble *) src2;

//
// Divide with a big double
//
// @param other    the divider
//
// @return the object
//
- (DBigDouble *) div :(DBigDouble *) other;

//
// Divide two big doubles and store the result in the object
//
// @param src1    the number
// @param src2    the divider
//
// @return the object
//
- (DBigDouble *) div :(DBigDouble *) src1 :(DBigDouble *) src2;

//
// Absolute the big double
// 
// @return the object
// 
- (DBigDouble *) abs;

//
// Negate the big double
// 
// @return the object
// 
- (DBigDouble *) negate;

//
// Do a square root on the big double
// 
// @return the object
// 
- (DBigDouble *) sqrt;

//
// Raise to the power of the big double
// 
// @param factor   the factor 
// 
// @return the object
// 
- (DBigDouble *) power :(unsigned long) factor;

//
// Ceil the big double
// 
// @return the object
// 
- (DBigDouble *) ceil;

//
// Floor the big double
// 
// @return the object
// 
- (DBigDouble *) floor ;

//
// Truncate the big double
// 
// @return the object
// 
- (DBigDouble *) trunc;


//// DComparable protocol implementation

//
// Compare with another big double
//
// @param other    the other big double
// 
// @return the compare result (-1,0,1)
//
- (int) compare :(DBigDouble *) other;


//// Textable protocol implementation

//
// Convert to a decimal text string
//
// @return a (new) text string with the big double (or nil)
//
- (DText *) toText;


//// DParsable protocol implementation

//
// Parse a string for a big double. Formats: 
// [space]*[+-][0..9]*{.[0..9]*}{[eE][+-][0..9]*}
// 
// @param cstr     the string to be parsed (moved to the first non-parsable char)
// 
// @return the result (0,ERANGE,ENODATA)
// 
- (int) fromString :(char **) cstr;

@end
#endif
#endif
