#ifndef _DBIGFRACTION_H
#define _DBIGFRACTION_H

// This file was automatically generated by objch.

//==============================================================================
//
// DBigFraction - the arbitrary length fraction number class in the ofc-library
//
//               Copyright (C) 2006  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:59 $ $Revision: 1.7 $
//
//==============================================================================
#ifdef HAVE_DBIGFRACTION
#endif

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DParsable.h"
#include "ofc/DSimpleArithmetic.h"
#include "ofc/DComparable.h"
#include "ofc/DText.h"

#if defined(HAVE_LIBGMP) && defined(HAVE_GMP_H)
#define HAVE_DBIGFRACTION  1
#endif

#ifdef HAVE_DBIGFRACTION

//
// The big fraction class implements a number of methods for working with 
// arbitrary large fraction numbers, also called rational numbers.
// The class uses the excellent gmp library for the actual calculations.
// 
// @example
// #include <stdio.h>
// #include "ofc/DBigFraction.h"
// 
// int main(int argc, char *argv[])
// {
//   DBigFraction *fra1 = [DBigFraction alloc];
//   DBigFraction *fra2 = [DBigFraction new  ];
//   DBigFraction *fra3 = [DBigFraction new  ];
//   DText        *str;
// 
//   [fra1 init :"6/4" :10];           // Init fraction with 6/4 (base 10)
// 
//   [fra2 fraction :8 :10];           // Set fraction with 8/10
// 
//   [fra2 mul :fra1];                 // fra2 = fra2 * fra1
// 
//   [fra3 sub :fra2 :fra1];           // fra3 = fra2 - fra1
// 
//   printf("Fraction (as double):%f\n", [fra3 get]);  // Get fraction as double
// 
//   str = [fra3 get :10];             // Get fraction as string, base 10
//   printf("Fraction (as string):%s\n", [str cstring]);
// 
//   if ([fra1 compare :fra2] == 0)    // Compare numbers
//     printf("Fraction1 is equal to fraction2\n");
//   else if ([fra1 compare :fra2] < 0)
//     printf("Fraction1 is smaller than fraction2\n");
//   else
//     printf("Fraction1 is greater than fraction2\n");
// 
//   [str  free];                      // Cleanup
//   [fra3 free];
//   [fra2 free];
//   [fra1 free];
// 
//   return 0;
// }
// 


@interface DBigFraction : Object <DTextable,DParsable,DSimpleArithmetic,DComparable>
{
@private
  void        *_value;      // the big fraction value
}
#endif
#ifdef HAVE_DBIGFRACTION

//// Constructors

//
// Initialise to fraction zero
//
// @return the object
//
- (DBigFraction *) init;

//
// Initialise the fraction with a double
// 
// @param value    the double value
// 
// @return the object
// 
- (DBigFraction *) init :(double) value;

//
// Initialise the fraction with a text string
// 
// @param cstring  the cstring with the fraction
// @param base     the base for the fraction
// 
// @return the object
// 
- (DBigFraction *) init :(const char *) cstring :(int) base;


//// Copy constructor

//
// Do a shallow copy of the big fraction
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free a big fraction
// 
// @return the object
//
- free;


//// Set methods

//
// Set the big fraction to zero
// 
// @return the object
// 
- (DBigFraction *) clear;

//
// Set the big fraction with a double
// 
// @param value    the double value
// 
// @return the object
// 
- (DBigFraction *) set :(double) value;

//
// Set the big fraction with a text string
//
// @param cstring   the cstring with the fraction
// @param base      the base of the number (2..36)
//
// @return success
//
- (BOOL) set :(const char *) cstring :(int) base;

//
// Set the big fraction with (small) numerator and denominator 
//
// @param num      the numerator
// @param denom    the denominator
//
// @return the object
//
- (DBigFraction *) fraction :(int) num :(int) denom;

//
//
// Move an other object in this object
//
// @param other    the other object
//
// @return this object
//
- (DBigFraction *) move :(DBigFraction *) other;


//// Get methods

//
// Get the big fraction as a double
// 
// @return the double
// 
- (double) get;

//
// Convert to a double
//
// @return the double
//
- (double) toDouble;

//
// Get the big fraction as a text string
// 
// @param base     the base for the text string (2..36)
// 
// @return a (new) text string (or nil)
// 
- (DText *) get :(int) base;

//
// Get the numerator (as small integer)
//
// @return the numerator
//
- (long) numerator;

//
// Get the denominator (as small integer)
//
// @return the denominator
//
- (long) denominator;


//// Methods on big fraction numbers

//
// Add with an other big fraction
//
// @param other    the big fraction
//
// @return the object
//
- (DBigFraction *) add :(DBigFraction *) other;

//
// Add two big fractions and store the result in the object
//
// @param src1     the first big fraction
// @param src2     the second big fraction
//
// @return the object
//
- (DBigFraction *) add :(DBigFraction *) src1 :(DBigFraction *) src2;

//
// Subract with an other big fraction
//
// @param other     the big fraction
//
// @return the object
//
- (DBigFraction *) sub :(DBigFraction *) other;

//
// Subract two big fractions and store the result in the object
//
// @param src1      the first big fraction
// @param src2      the second big fraction
//
// @return the object
//
- (DBigFraction *) sub :(DBigFraction *) src1 :(DBigFraction *) src2;

//
// Multiply with an other big fraction
//
// @param other    the big fraction
//
// @return the object
//
- (DBigFraction *) mul :(DBigFraction *) other;

//
// Multiply two big fractions and store the result in the object
//
// @param src1    the first big fraction
// @param src2    the second big fraction
//
// @return the object
//
- (DBigFraction *) mul :(DBigFraction *) src1 :(DBigFraction *) src2;

//
// Divide with an other big fraction
//
// @param other    the divisor big fraction
//
// @return the object
//
- (DBigFraction *) div :(DBigFraction *) other;

//
// Divide two big fractions and store the result in the object
//
// @param src1    the dividend big fraction
// @param src2    the divisor big fraction
//
// @return the object
//
- (DBigFraction *) div :(DBigFraction *) src1 :(DBigFraction *) src2;

//
// Absolute the big fraction
// 
// @return the object
// 
- (DBigFraction *) abs;

//
// Negate the big fraction
// 
// @return the object
// 
- (DBigFraction *) negate;

//
// Invert the big fraction
// 
// @return the object
// 
- (DBigFraction *) invert;

//
// Normalize the big fraction
//
// @return the object
//
- (DBigFraction *) norm;


//// DComparable protocol implementation

//
// Compare with an other big fraction
//
// @param other    the other big fraction
// 
// @return the compare result ( <0, 0, 0> )
//
- (int) compare :(DBigFraction *) other;


//// Textable protocol implementation

//
// Convert to a decimal text string
//
// @return a (new) text string with the big fraction (or nil)
//
- (DText *) toText;


//// DParsable protocol implementation

//
// Parse a string for a big fraction. Format: [space]*[+-]number/number
// and number is one of:
// decimal:     [1..9][0..9]*
// hexadecimal: 0x[0..9a..f]*
// octal:       0[0..7]*
// 
// @param cstr     the string to be parsed (moved to the first non-parsable char)
// 
// @return the result (0,ERANGE,ENODATA)
// 
- (int) fromString :(char **) cstr;

@end
#endif
#endif
