#ifndef _DBIGINT_H
#define _DBIGINT_H

// This file was automatically generated by objch.

//==============================================================================
//
//      DBigInt - the arbitrary length integer number class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:59 $ $Revision: 1.9 $
//
//==============================================================================
#ifdef HAVE_DBIGINT
#endif

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DDatable.h"
#include "ofc/DParsable.h"
#include "ofc/DSimpleArithmetic.h"
#include "ofc/DBitArithmetic.h"
#include "ofc/DComparable.h"
#include "ofc/DText.h"

#if defined(HAVE_LIBGMP) && defined(HAVE_GMP_H)
#define HAVE_DBIGINT  1
#endif

#ifdef HAVE_DBIGINT

//
// The big int class implements a number of methods for working with arbitrary large integer numbers.
// The class uses the excellent gmp library for the actual calculations.
// 
// @example
// #include <stdio.h>
// #include "ofc/DBigInt.h"
// 
// int main(int argc, char *argv[])
// {
//   DBigInt *nr1 = [DBigInt alloc];
//   DBigInt *nr2 = [DBigInt alloc];
//   DBigInt *nr3 = [DBigInt new  ];
//   DText   *str;
// 
//   [nr1 init :1000];                 // Initialise with an integer number
//   [nr2 init :"7000" :10];           // Initialise with a string and a base
// 
//   [nr2 mul :nr1];                   // nr2 = nr2 * nr1
// 
//   [nr3 sub :nr2 :nr1];              // nr3 = nr2 - nr1
// 
//   printf("Value (as long):%ld\n", [nr3 get]);
// 
//   str = [nr3 get :10];              // Convert to a string, base 10
//   printf("Value (as string):%s\n", [str cstring]);
// 
//   if ([nr1 compare :nr2] == 0)      // Compare numbers
//     printf("Int1 is equal to int2\n");
//   else if ([nr1 compare :nr2] < 0)
//     printf("Int1 is smaller than int2\n");
//   else
//     printf("Int1 is greater than int2\n");
// 
//   [str free];                       // Cleanup 
//   [nr3 free];
//   [nr2 free];
//   [nr1 free];
// 
//   return 0;
// }
// 

@interface DBigInt : Object <DTextable,DParsable,DDatable,DSimpleArithmetic,DBitArithmetic,DComparable>
{
@private
  void        *_value;      // the big integer value
}

#endif
#ifdef HAVE_DBIGINT

//// Constructors

//
// Initialise to integer zero
//
// @return the object
//
- (DBigInt *) init;

//
// Initialise to a (small) number
//
// @param number   the (small) number
//
// @return the object
//
- (DBigInt *) init :(long) number;

//
// Initialise with a string
// 
// @param cstring  the cstring with the number
// @param base     the base for the number
// 
// @return the object
// 
- (DBigInt *) init :(const char *) cstring :(int) base;


//// Copy constructor

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free a big integer
// 
// @return the object
//
- free;

// Member methods
//
// Check if the big integer is negative
// 
// @return is it?
// 
- (BOOL) isNegative;

//
// Check if the big integer is zero
// 
// @return is it ?
// 
- (BOOL) isZero;


//// Set methods

//
// Set the big integer to zero
// 
// @return the object
// 
- (DBigInt *) clear;

//
// Set the big integer with a (small) number
//
// @param number   the (small) number
//
// @return the object
//
- (DBigInt *) set :(long) number;

//
// Set the big integer with an array of bytes
//
// @param numbers   the number in an array of bytes [0] = least significant (NULL = number 0)
// @param length    the number of bytes in number (length 0 = number 0)
// @param negative  should the number be negative ?
//
// @return the object
//
- (DBigInt *) set :(const unsigned char *) number :(unsigned long) length :(BOOL) negative;

//
// Set the big integer with a text number
//
// @param cstring   the cstring with the number
// @param base      the base of the number (2..36)
//
// @return success
//
- (BOOL) set :(const char *) cstring :(int) base;

//
//
// Move an other object in this object
//
// @param other    the other object
//
// @return this object
//
- (DBigInt *) move :(DBigInt *) other;


//// Get methods

//
// Get the big integer as (small) integer
// 
// @return the (small) integer
// 
- (long) get;

//
// Get the big integer as a text string
// 
// @param base     the base for the text string (2..36)
// 
// @return a (new) text string (or nil)
// 
- (DText *) get :(int) base;

//
// Return the big number as data string (note: keep the negative state seperate)
// 
// @return the number in a (new) data string (empty for number 0)
// 
- (DData *) toData;


//// Methods on big integer numbers

//
// Add with a big integer 
//
// @param other    the big integer
//
// @return the object
//
- (DBigInt *) add :(DBigInt *) other;

//
// Add two big integers and store the result in the object
//
// @param src1     the first big integer
// @param src2     the second big integer
//
// @return the object
//
- (DBigInt *) add :(DBigInt *) src1 :(DBigInt *) src2;

//
// Subract with a big integer
//
// @param other     the big integer
//
// @return the object
//
- (DBigInt *) sub :(DBigInt *) other;

//
// Subract two big integers and store the result in the object
//
// @param src1      the first big integer
// @param src2      the second big integer
//
// @return the object
//
- (DBigInt *) sub :(DBigInt *) src1 :(DBigInt *) src2;

//
// Multiply with a big integer
//
// @param other    the big integer
//
// @return the object
//
- (DBigInt *) mul :(DBigInt *) other;

//
// Multiply two big integers and store the result in the object
//
// @param src1    the first big integer
// @param src2    the second big integer
//
// @return the object
//
- (DBigInt *) mul :(DBigInt *) src1 :(DBigInt *) src2;

//
// Divide with a big integer
//
// @param other    the divider
//
// @return the object
//
- (DBigInt *) div :(DBigInt *) other;

//
// Divide two big integers and store the result in the object
//
// @param src1    the number
// @param src2    the divider
//
// @return the object
//
- (DBigInt *) div :(DBigInt *) src1 :(DBigInt *) src2;

//
// Determine the module with a big integer
//
// @param other    the divider
//
// @return the object
//
- (DBigInt *) mod :(DBigInt *) other ;

//
// Determine the module from two big integers and store the result in the object
//
// @param src1     the number
// @param src2     the divider
//
// @return the object
//
- (DBigInt *) mod :(DBigInt *) src1 :(DBigInt *) src2;

//
// Absolute the big integer
// 
// @return the object
// 
- (DBigInt *) abs;

//
// One's complement the big integer
// 
// @return the object
// 
- (DBigInt *) com;

//
// Negate the big integer
// 
// @return the object
// 
- (DBigInt *) negate;

//
// And with a big integer
// 
// @param other    the other big integer
// 
// @return the object
// 
- (DBigInt *) and :(DBigInt *) other;

//
// And two big integers and store the result in the object
// 
// @param src1     the first big integer
// @param src2     the second big integer
// 
// @return the object
// 
- (DBigInt *) and :(DBigInt *) src1 :(DBigInt *) src2;

//
// Or with a big integer
// 
// @param other    the other big integer
// 
// @return the object
// 
- (DBigInt *) or :(DBigInt *) other;

//
// Or two big integers and store the result in the object
// 
// @param src1     the first big integer
// @param src2     the second big integer
// 
// @return the object
// 
- (DBigInt *) or :(DBigInt *) src1 :(DBigInt *) src2;

//
// Xor with a big integer
// 
// @param other    the other big integer
// 
// @return the object
// 
- (DBigInt *) xor :(DBigInt *) other;

//
// Xor two big integers and store the result in the object
// 
// @param src1     the first big integer
// @param src2     the second big integer
// 
// @return the object
// 
- (DBigInt *) xor :(DBigInt *) src1 :(DBigInt *) src2;

//
// Left shift the big integer
// 
// @param shifts   the number of shifts
// 
// @return the object
// 
- (DBigInt *) lshift :(unsigned long) shifts;

//
// Left shift a big integer and store the result in the object
//
// @param src      the number to be shifted
// @param shifts   the number of shifts
// 
// @return the object
// 
- (DBigInt *) lshift :(DBigInt *) src :(unsigned long) shifts;

//
// Right shift the big integer
// 
// @param shifts   the number of shifts
// 
// @return the object
// 
- (DBigInt *) rshift :(unsigned long) shifts;

//
// Right shift a big integer and store the big integer
// 
// @param src      the number to be shifted
// @param shifts   the number of shifts
// 
// @return the object
// 
- (DBigInt *) rshift :(DBigInt *) src :(unsigned long) shifts;


//// DComparable protocol implementation

//
// Compare with a big integer
//
// @param other    the other big integer
// 
// @return the compare result (-1,0,1)
//
- (int) compare :(DBigInt *) other;


//// Textable protocol implementation

//
// Convert to a decimal text string
//
// @return a (new) text string with the big integer (or nil)
//
- (DText *) toText;


//// DParsable protocol implementation

//
// Parse a string for a big integer. Formats:
// decimal:     [space]*[+-][1..9][0..9]*
// hexadecimal: [space]*[+-]0x[0..9a..f]*
// octal:       [space]*[+-]0[0..7]*
// 
// @param cstr     the string to be parsed (moved to the first non-parsable char)
// 
// @return the result (0,ERANGE,ENODATA)
// 
- (int) fromString :(char **) cstr;

@end
#endif
#endif
