#ifndef _DCOLOR_H
#define _DCOLOR_H

// This file was automatically generated by objch.

//==============================================================================
//
//             DColor - the color class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-19 12:28:59 $ $Revision: 1.11 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DParsable.h"

#include "ofc/DText.h"

#define DCLR_UNKNOWN         (-1)   // the unknown text color
#define DCLR_BLACK           (0)    // the black text color
#define DCLR_RED             (1)    // the red text color
#define DCLR_GREEN           (2)    // the green text color
#define DCLR_YELLOW          (3)    // the yellow text color
#define DCLR_BLUE            (4)    // the blue text color
#define DCLR_MAGENTA         (5)    // the magenta text color
#define DCLR_CYAN            (6)    // the cyan text color
#define DCLR_WHITE           (7)    // the white text color

//
// The color class implements a number of methods for using colors.
//
// @example
//
// #include <stdio.h>
// #include "ofc/DColor.h"
// 
// int main(int argc, char *argv[])
// {
//   DColor *red = [DColor alloc];
//   DColor *grn = [DColor alloc];
//   DColor *blu = [DColor new  ];
//   DColor *col = [DColor new  ];
// 
//   double  H,S,V;
//   DText  *str;
// 
//   [red init :"red"];                   // Init with a color name
// 
//   [grn init :0 :255 :0 :80];           // Init with green color and alpha 80
// 
//   [blu set  :0 :0 :255];               // Set the blue color
// 
//   [col move :red];                     // Move the color
// 
//   [col blend :123 :145 :78 :20];       // Blend the color with another color, alpha 20
// 
//   [col lighter :1.3];                  // Makes the color lighter
// 
//   [col toHSV :&H :&S :&V];             // Convert the color to HSV
// 
//   str = [col toText];                  // Convert the color to text
// 
//   printf("Color (%0d,%0d,%0d) or \"%s\" in HSV:%.3f, %.3f, %.3f\n", 
//      [col red], [col green], [col blue], [str cstring], H, S, V);
//   
//   [str free];                          // Cleanup
//   [col free];
//   [blu free];
//   [grn free];
//   [red free];
// 
//   return 0;
// }
// 

@interface DColor : Object <DTextable,DParsable>
{
@public
  unsigned char _red;         // the red color 255 = full
  unsigned char _green;       // the green color 255 = full
  unsigned char _blue;        // the blue color 255 = full
  unsigned char _alpha;       // the alpha factor 255 = solid
  int           _text;        // the (alternative) text color (DCLR_... or -1)
}


//// Constructors

//
// Initialise an empty (black) color
//
// @return the object
//
- (DColor *) init;

//
// Initialise to a solid named color
//
// @param name     the color name
//
// @return the object
//
- (DColor *) init :(const char *) name;

//
// Initialise to a solid rgb color
//
// @param red      the red factor (0..255)
// @param green    the green factor (0..255)
// @param blue     the blue factor (0..255)
//
// @return the object
//
- (DColor *) init :(unsigned char) red :(unsigned char) green :(unsigned char) blue;

//
// Initialise to a transparent rgb color
//
// @param red      the red factor (0..255)
// @param green    the green factor (0..255)
// @param blue     the blue factor (0..255)
// @param alpha    the alpha factor (0..255)
//
// @return the object
//
- (DColor *) init :(unsigned char) red :(unsigned char) green :(unsigned char) blue :(unsigned char) alpha;


//// Member methods

//
// Get the red factor
//
// @return the red factor
//
- (unsigned char) red;

//
// Get the green factor
//
// @return the green factor
//
- (unsigned char) green;

//
// Get the blue factor
//
// @return the blue factor
//
- (unsigned char) blue;

//
// Get the alpha factor
//
// @return the solid factor
//
- (unsigned char) alpha;

//
// Set the alpha factor
// 
// @param alpha    the transparent factor (0..255)
// 
// @return the object
// 
- (DColor *) alpha :(unsigned char) alpha;

//
// Get the text color
//
// @return the text color
//
- (int) textColor;

//
// Set the text color
// 
// @param color    the text color (DCLR_..)
// 
// @return the object
// 
- (DColor *) textColor :(int) color;


//// Set the color

//
// Set the color with another color
// 
// @param other     the other color
// 
// @return the object
// 
- (DColor *) move :(DColor *) other;

//
// Set a color based on a name
// (alpha is not changed)
// 
// @param name     the name of the color
// 
// @return is the name found ?
// 
- (BOOL) set :(const char *) name;

//
// Set a color based on rgb 
// (alpha and text color are not changed)
// 
// @param red      the red color (0..255)
// @param green    the green color (0..255)
// @param blue     the blue color (0..255)
// 
// @return the object
// 
- (DColor *) set :(unsigned char) red :(unsigned char) green :(unsigned char) blue;

//
// Set a transparent color based on rgb
// (text color is not changed)
// 
// @param red      the red color (0..255)
// @param green    the green color (0..255)
// @param blue     the blue color (0..255)
// @param alpha    the transparent factor (0..255)
//
// @return the object
//  
- (DColor *) set :(unsigned char) red :(unsigned char) green :(unsigned char) blue :(unsigned char) alpha;

//
// Set a transparent color based on rgb and a text color
// 
// @param red      the red color (0..255)
// @param green    the green color (0..255)
// @param blue     the blue color (0..255)
// @param alpha    the transparent factor (0..255)
// @param color    the text color (DCLR_..)
//
// @return the object
// 
- (DColor *) set :(unsigned char) red :(unsigned char) green :(unsigned char) blue :(unsigned char) alpha :(int) color;


//// Color manipulation

//
// Make the color lighter or darker
// 
// @param factor   the lighter (<1.0..>) or darker [0.0..1.0>) factor
//
// @return the object
// 
- (DColor *) lighter :(double) factor;

//
// Alpha blending with a color
// 
// @param red      the red factor of the color (0..255)
// @param green    the green factor of the color (0..255)
// @param blue     the blue factor of the color (0..255)
// @param alpha    the alpha blending factor (0..255)
// 
// @return the object
// 
- (DColor *) blend :(unsigned char) red :(unsigned char) green :(unsigned char) blue :(unsigned char) alpha;


//// Color systems

//
// 'Convert' to RGB
// 
// @param red      the destination for red (0.0 .. 1.0)
// @param green    the destination for green (0.0 .. 1.0)
// @param blue     the destination for blue (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) toRGB :(double *) red :(double *) green :(double *) blue;

//
// 'Convert' from RGB
// 
// @param red      the red color (0.0 .. 1.0)
// @param green    the green color (0.0 .. 1.0)
// @param blue     the blue color (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) fromRGB :(double) red :(double) green :(double) blue;

//
// Convert to the composite video signals YIQ
// (Y = grey level, I+Q color components)
// 
// @param Y        the destination for Y (0.0 .. 1.0)
// @param I        the destination for I (0.0 .. 1.0)
// @param Q        the destination for Q (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) toYIQ :(double *) Y :(double *) I :(double *) Q;

//
// Convert from YIQ
// 
// @param Y        the Y (0.0 .. 1.0)
// @param I        the I (0.0 .. 1.0)
// @param Q        the Q (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) fromYIQ :(double) Y :(double) I :(double) Q;

//
// Convert to HLS
// (H = Hue, L=Lightness, S=Saturation)
// 
// @param H        the destination for H (0.0 .. 359.9)
// @param L        the destination for L (0.0 .. 1.0)
// @param S        the destination for S (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) toHLS :(double *) H :(double *) L :(double *) S;

//
// Convert from HLS
// 
// @param H        the H (0.0 .. 359.9)
// @param L        the L (0.0 .. 1.0)
// @param S        the S (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) fromHLS :(double) H :(double) L :(double) S;

//
// Convert to HSV
// (H = Hue, S=Saturation, Value)
// 
// @param H        the destination for H (0.0 .. 359.9)
// @param S        the destination for S (0.0 .. 1.0)
// @param V        the destination for V (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) toHSV :(double *) H :(double *) S :(double *) V;

//
// Convert from HSV
// 
// @param H        the H (0.0 .. 359.9)
// @param S        the S (0.0 .. 1.0)
// @param V        the V (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) fromHSV :(double) H :(double) S :(double) V;

//
// Convert to CMY
// (C = Cyan, M = Magenta, Y = Yellow)
// 
// @param C        the destination for C (0.0 .. 1.0)
// @param M        the destination for M (0.0 .. 1.0)
// @param Y        the destination for Y (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) toCMY :(double *) C :(double *) M :(double *) Y;

//
// Convert from CMY
// 
// @param C        the C (0.0 .. 1.0)
// @param M        the M (0.0 .. 1.0)
// @param Y        the Y (0.0 .. 1.0)
// 
// @return the object
// 
- (DColor *) fromCMY :(double) C :(double) M :(double) Y;


//// Textable protocol methods

//
// Convert to a text string (RGB coded in six hex digits, 
// if a text color is known, it is added to the six hex
// digits)
//
// @return a (new) text string with the color
//
- (DText *) toText;


//// Parsable protocol members

//
// Parse a string to set a color object.
// This method supports the following formats: <br>
//  #RGB        -- color is defined by three hex digits <br>
//  #RRGGBB     -- color is defined by six hex digits <br>
//  R%,G%,B%    -- color is defined by three percentage digits <br>
//  R.R,G.G,B.B -- color is defined by three floats (0.0 .. 1.0) <br>
//  R,G,B       -- color is defined by three bytes (0 .. 255) <br>
//  name        -- color is defined by an internal name <br>
// the color can be followed by a comma and a text color name <br>
// Example: #8042AF,black <br>
//
// @param cstr     the string to be parsed (moved to first non-parsable char)
// 
// @return the result (0, ERANGE, ENODATA)
//
- (int) fromString :(char **) cstr;

@end
#endif
