#ifndef _DCOMPLEX_H
#define _DCOMPLEX_H

// This file was automatically generated by objch.

//==============================================================================
//
//           DComplex - the complex number class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-20 06:23:36 $ $Revision: 1.13 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DSimpleArithmetic.h"

#include "ofc/DText.h"

// The complex class implements a number of methods for working with complex numbers.
//
// Heavily 'inspired' by the excellent ccmath library by Daniel A. Atkinson (c) Copyright 2000.
//
// @example
// 
// #include <stdio.h>
// #include "ofc/DComplex.h"
// 
// int main(int argc, char *argv[])
// {
//   DComplex *cp1 = [DComplex alloc];
//   DComplex *cp2 = [DComplex new  ];
//   DComplex *cp3 = [DComplex new  ];
//   DText    *str;
// 
//   [cp1 init :0.5 :0.5];                // Init with 0.5+0.5j
// 
//   [cp2 set  :2.0 :3.0];                // Set with 2+3j
//   
//   [cp3 move :cp1];                     // cp3 = cp1
// 
//   [cp2 mul  :cp1];                     // cp2 = cp2 * cp1
// 
//   [cp3 sub  :cp2 :cp1];                // cp3 = cp2 - cp1
// 
//   str = [cp3 toText];
// 
//   printf("Complex3 = %.3f + j %0.3f or \"%s\".\n", [cp3 re], [cp3 im], [str cstring]);
//   printf("ABS(complex3) = %.3f\n", [cp3 abs]);
// 
//   [str free];
//   [cp1 free];
//   [cp2 free];
//   [cp3 free];
// 
//   return 0;
// }
// 

@interface DComplex : Object <DTextable,DSimpleArithmetic>
{
@private
  double       _re;
  double       _im;
}


//// Constructors

//
// Initialise to complex number zero
//
// @return the object
//
- (DComplex *) init;

//
// Initialise to a complex number
//
// @param re       the real part
// @param im       the imaginary part
//
// @return the object
//
- (DComplex *) init :(double) re :(double) im;


//// Member methods

//
// Get the real part
//
// @return the real part
//
- (double) re;

//
// Set the real part
//
// @param re       the real part of a complex number
//
// @return the object
//
- (DComplex *) re :(double) re;

//
// Get the imaginary part of the number
//
// @return the imaginary part
//
- (double) im;

//
// Set the imaginary part of the number
//
// @param im       the imaginary part
//
// @return the object
//
- (DComplex *) im :(double) im;

//
// Set the real and imaginary part of the number
//
// @param re       the real part
// @param im       the imaginary part
//
// @return the object
//
- (DComplex *) set :(double) re :(double) im;

//
//
// Move an other object in this object
//
// @param other    the other object
//
// @return this object
//
- (DComplex *) move :(DComplex *) other;


//// Methods on complex numbers

//
// Add with a complex number 
//
// @param other    the complex number
//
// @return the object
//
- (DComplex *) add :(DComplex *) other;

//
// Add two complex numbers and store the result in the object
//
// @param src1     the first complex number
// @param src2     the second complex number
//
// @return the object
//
- (DComplex *) add :(DComplex *) src1 :(DComplex *) src2;

//
// Subract with a complex number
//
// @param other     the complex number
//
// @return the object
//
- (DComplex *) sub :(DComplex *) other;

//
// Subract two complex numbers and store the result in the object
//
// @param src1      the first complex number
// @param src2      the second complex number
//
// @return the object
//
- (DComplex *) sub :(DComplex *) src1 :(DComplex *) src2;

//
// Multiply with an other complex number
//
// @param other    the complex number
//
// @return the object
//
- (DComplex *) mul :(DComplex *) other;

//
// Multiply two complex numbers and store the result in the object
//
// @param src1     the first complex number
// @param src2     the second complex number
//
// @return the object
//
- (DComplex *) mul :(DComplex *) src1 :(DComplex *) src2;

//
// Multiply with a real number
//
// @param re       the real number
//
// @return the object
//
- (DComplex *) rmul :(double) re;

//
// Multiply with a imaginary number
//
// @param im       the imaginary number
//
// @return the object
//
- (DComplex *) imul :(double) im;

//
// Divide with a complex number
//
// @param other    the divider
//
// @return the object
//
- (DComplex *) div :(DComplex *) other;

//
// Divide two complex numbers and store the result in the object
//
// @param src1     the first complex number
// @param src2     the divider
//
// @return the object
//
- (DComplex *) div :(DComplex *) src1 :(DComplex *) src2;

//
// Conjugate the complex number
//
// @return the object
//
- (DComplex *) cng ;

//
// Calculate the modulus of the complex number
//
// @return the modulus
//
- (double) abs ;

//
// Calculate the square of the modulus of the complex number
//
// @return the square of the modulus
//
- (double) nrm;

//
// Calculate the square root for the complex number
//
// @return the object
//
- (DComplex *) sqrt;

//
// Calculate the exponent function for the complex number
//
// @return the object
//
- (DComplex *) exp;

//
// Calculate the natural logarithm for the complex number
//
// @return the object
//
- (DComplex *) log;

//
// Calculate the trigonometric functions sine for the complex number
//
// @return the object
//
- (DComplex *) sin;

//
// Calculate the trigonometric functions cosine for the complex number
//
// @return the object
//
- (DComplex *) cos;

//
// Calculate the trigonometric functions trangent for the complex number
//
// @return the object
//
- (DComplex *) tan;

//
// Calculate the inverse trigonometric function sine for the complex number
//
// @return the object
//
- (DComplex *) asin;

//
// Calculate the inverse trigonometric function cosine for the complex number
//
// @return the object
//
- (DComplex *) acos;

//
// Calculate the inverse trigonometric function tangent for the complex number
//
// @return the object
//
- (DComplex *) atan;

//
// Calculate the hyperbolic function sine for the complex number
//
// @return the object
//
- (DComplex *) sinh;

//
// Calculate the hyperbolic function cosine for the complex number
//
// @return the object
//
- (DComplex *) cosh;

//
// Calculate the hyperbolic function tangent for the complex number
//
// @return the object
//
- (DComplex *) tanh;

//
// Calculate the inverse hyperbolic function sine for the complex number
//
// @return the object
//
- (DComplex *) asinh;

//
// Calculate the inverse hyperbolic function cosine for the complex number
//
// @return the object
//
- (DComplex *) acosh;

//
// Calculate the inverse hyperbolic function tangent for the complex number
//
// @return the object
//
- (DComplex *) atanh;


//// Textable protocol methods

//
// Convert to a text string 
//
// @return a (new) text string with the complex number
//
- (DText *) toText;

@end
#endif
