#ifndef _DDATA_H
#define _DDATA_H

// This file was automatically generated by objch.

//==============================================================================
//
//            DData - the data string class in the ofc-library
//
//              Copyright (C) 2003-2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-20 06:23:36 $ $Revision: 1.25 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DDatable.h"
#include "ofc/DTextable.h"
#include "ofc/DTextScannable.h"
#include "ofc/DComparable.h"
#include "ofc/DSizeable.h"
#include "ofc/DDataReadable.h"
#include "ofc/DDataWritable.h"

#include "ofc/DText.h"


//
// The DData class implements a number of methods for manipulating data strings.
// @example
//
// #include <stdio.h>
// #include "ofc/DData.h"
// 
// int main(int argc, char *argv[])
// {
//   DData *dat1 = [DData alloc];
//   DData *dat2 = [DData new  ];
//   DText *str;
// 
//   [dat1 init :"DData stores data strings" :25]; // Init with a data string
// 
//   printf("Length of data string1:%ld\n", [dat1 length]);
// 
//   str = [dat1 tohexString];
//   printf("Hex string of data string:%s.\n", [str cstring]);
//   [str free];
// 
//   str = [dat1 toBase64];
//   printf("Base64 string of data string:%s.\n", [str cstring]);
//   [str free];
// 
//   printf("Hash value of data string:%ld.\n", [dat1 hash]);
// 
//   printf("Index of \"data\" in data string is:%ld.\n", [dat1 index :"data" :4 :0 :-1]);
// 
//   [dat2 set :"DData stores data strings" :1 :17]; // Set with substring
// 
//   [dat2 put :16 :'!'];                 // Put a character in the string
// 
//   [dat2 insert :16 :16 :" strings" :9]; // Insert a data string
// 
//   [dat2 delete :11 :15];               // Delete a substring
// 
//   str = [dat2 toText];
//   printf("Data string2:%s.\n", [str cstring]);
//   [str free];
// 
//   printf("Data string1 is ");          // Compare strings
//   if ([dat1 compare :dat2] > 0)
//     printf("greater");
//   else if ([dat1 compare :dat2] < 0)
//     printf("smaller");
//   else
//     printf("equal");
//   printf(" than data string2.\n");
// 
//   [dat1 free];                         // Cleanup
//   [dat2 free];
// 
//   return 0;
// }
//

@interface DData : Object <DDatable,DTextable,DTextScannable,DComparable,DSizeable,DDataReadable,DDataWritable>
{
@private
  unsigned long  _size;     // the allocated size for the data
  unsigned       _extra;    // the extra size amount
  unsigned long  _length;   // the length of the data
  unsigned char *_data;     // the data
  unsigned long  _pntr;     // the 'file pointer' for the protocols
  int            _error;    // the last reported error (or 0)
}


//// Constructors

//
// Initialise to an empty data string
//
// @return the object
//
- (DData *) init;

//
// Initialise the instance with data
//
// @param data     the data
// @param length   the length of the data
//
// @return the string object
//
- (DData *) init :(const unsigned char *) data :(unsigned long) length;


//// Copy related methods

//
// Do a shallow copy of the data string object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free;


//// Size methods

//
// Insure the size for the data string
//
// @param size     the requested size for the data
//
// @return the object
//
- (DData *) size :(unsigned long) size;

//
// Set the amount of extra size increase if the size must be extended
//
// @param extra    the amount (def. 0)
//
// @return the object
//
- (DData *) extra :(unsigned) extra;


//// Conversion methods

//
// Convert the data string to a hexidecimal text string (every
// character in the string is converted to two text characters
// hexadecimal)
//
// @return the (new) string object
//
- (DText *) tohexString;

//
// Calculate the hash of the data string
//
// @return the hash value
//
- (unsigned long) hash;

//
// Convert to Base64
// 
// @return a (new) text string object (or nil)
// 
- (DText *) toBase64;

//
// Convert to a text string with printable characters
// (by changing non-printable characters to a dot character)
//
// @return a (new) text string object
//
- (DText *) toPrintable;


//// DTextable protocol methods

//
// Convert the object to a text string without any
// conversion. (see toPrintable for a method that 
// convert the data object to a true printable text object)
//
// @return a (new) text string object
//
- (DText *) toText;


//// Data string setters

//
// Clear the data string
// 
// @return the object
// 
- (DData *) clear;

//
// Set data in the data string
//
// @param data     the data for the string ('\0' allowed)
// @param length   the length of the data
//
// @return the object
//
- (DData *) set :(const unsigned char *) data :(unsigned long) length;

//
// Set a substring of data in the data string
//
// @param data     the data for the string ('\0' allowed)
// @param from     the start in the data
// @param to       the end in the data (incl)
//
// @return the object
//
- (DData *) set :(const unsigned char *) data :(long) from :(long) to;

//
// Set the data string with Base64 converted data
// 
// @param cstring  the c-string with the base 64 codes
// 
// @return the object
// 
- (DData *) fromBase64 :(const char *) cstring;


//// Byte methods

//
// Put a byte at the given index
//
// @param index     the index (-length..+length)
// @param byte      the byte to be placed
//
// @return the object
//
- (DData *) put :(long) index :(unsigned char) byte;

//
// Get the byte at the given index
//
// @param index    the index 
//
// @return the byte from the data string
//
- (unsigned char) get :(long) index;

//
// Delete the byte at the given index
//
// @param index    the index
//
// @return the object
//
- (DData *) delete :(long) index;


//// Slice methods

//
// Insert a data string in a part of the data string
//
// @param from     the start index
// @param to       the end index (to >= (from-1))
// @param data     the new data
// @param length   the length of the new data
//
// @return the object
//
- (DData *) insert :(long) from :(long) to :(const unsigned char *) data :(unsigned long) length;

//
// Return a substring
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return a (new) data string object
//
- (DData *) get :(long) from :(long) to;

//
// Delete a substring
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the object
//
- (DData *) delete :(long) from :(long) to;


//// Data string member methods

//
// Check if the data string is empty
//
// @return Is it empty ?
//
- (BOOL) isEmpty;

//
// Return the length of the data string
//
// @return the length
//
- (unsigned long) length;

//
// Return the pointer to the data of the string
//
// @return the pointer to the data
//
- (const unsigned char *) data;

//
// Return the size of the data string in bytes
//
// @return the size
//
- (unsigned long) size;

//
// Return the last error
//
// @return the error (0, ENODATA)
//
- (int) error;


//// DDatable protocol methods

//
// Convert the object to a data string object
//
// @return a (new) data string object
//
- (DData *) toData;


//// DDataReadable protocol methods

//
// Read text string (not-print characters are replaced by a '.')
//
// @param length   the length of the text to be read
//
// @return a (new) text string object
//
- (DText *) readText :(unsigned long) length;

//
// Read a data string 
//
// @param length   the length of the data
//
// @return a (new) data object (empty if no data available)
//
- (DData *) readData :(unsigned long) length;;

//
// Read a data string 
//
// @param dest     the destination buffer
// @param length   the length of the data (and minimum size of dest)
//
// @return the actual number of bytes read
//
- (unsigned long) readData :(unsigned char *) dest :(unsigned long) length;

//
// Read a character
//
// @return the character read (or EOS)
//
- (char) readChar;

//
// Read a byte
//
// @return the byte read (or 0)
//
- (unsigned char) readByte;

//
// Read a short
//
// @return the short read (or 0)
//
- (short) readShort;

//
// Read a long
//
// @return the long read (or 0)
//
- (long) readLong;

//
// Read a double
//
// @return the double read (or 0.0)
//
- (double) readDouble;

//
// Is the end of the readable reached ?
//
// @return is it?
//
- (BOOL) isEof;


//// TextScannable protocol methods

//
// Scan for a text string until a separator character. The separator character is 
// is not part of the returned string and is skipped in the Readable.
//
// @return a (new) text string (or nil if the separator is not found)
//
- (DText *) scanText :(char) separator;

//
// Scan for a text string until one of the separator characters in the 
// separator string.  The matched separator character is is not part of 
// the returned string, is skipped in the text object and is returned in 
// the separator argument.
//
// @param separators  the c-string with the separators
// @param separator   the destination for the actual matched separator character (or NULL)
//
// @return a (new) text string (or nil if none of the separators is found)
//
- (DText *) scanText :(const char *) separators :(char *) separator;

//
// Match a string case sensitive; if the string matched, the current position in
// the readable is moved to the location after the matched string.
//
// @param cstring   the c-string to be matched
//
// @return does the string match ?
//
- (BOOL) cmatch :(const char *) cstring;

//
// Match a string case insensitive; if the string matched, the current position in
// the readable is moved to the location after the matched string.
//
// @param cstring   the c-string to be matched
//
// @return does the string match ?
//
- (BOOL) imatch :(const char *) cstring;

//
// Scan for a (decimal) integer value
//
// @param wrong   the value returned if there is no int to scan
//
// @return the integer read (or the wrong value)
//
- (int) scanInt :(int) wrong;

//
// Skip characters
//
// @param ch    the character to be skipped
//
// @return the number of characters skipped
//
- (int) skipChar :(char) ch;

//
// Skip whitespace
//
// @return the number of characters skipped
//
- (int) skipWhiteSpace;


//// DDataWritable protocol methods

//
// Write text string 
//
// @param text     the text to be written
//
// @return success
//
- (BOOL) writeText :(const char *) text;

//
// Write the data string 
//
// @param data     the data to be written
// @param length   the length of the data
//
// @return success
//
- (BOOL) writeData :(const unsigned char *) text :(unsigned long) length;

//
// Write the character
//
// @param ch       the character to be written
//
// @return success
//
- (BOOL) writeChar :(char) ch;

//
// Write the byte
//
// @param byte     the byte to be written
//
// @return success
//
- (BOOL) writeByte :(unsigned char) byte;

//
// Write the short
//
// @param sh       the short to be written
//
// @return success
//
- (BOOL) writeShort :(short) sh;

//
// Write the long
//
// @param val      the long to be written
//
// @return success
//
- (BOOL) writeLong :(long) val;

//
// Write the double
//
// @param val      the double to be written
//
// @return the object
//
- (BOOL) writeDouble :(double) val;

//
// Flush the output buffers
//
// @return success
//
- (BOOL) flush;


//// DDataReadable and DDataWritable protocol methods

//
// Tell the current position in the writeable
//
// @return the current position
//
- (unsigned long) tell;

//
// Move the current position to a location
//
// @param offset   the offset from the origin (in bytes)
// @param origin   the origin for the offset (0=begin, 1=current, 2=end,others = no change)
//
// @return the object
//
- (BOOL) seek :(unsigned long) offset :(int) origin;

//
// Skip a number of positions
//
// @param offset   the number of bytes to skip
//
// @return success
//
- (BOOL) skip :(unsigned long) offset;


//// Data string manipulation methods

//
// Append data to the data string
//
// @param data     the data to be appended
// @param length   the length of the data to be appended
//
// @return the object
//
- (DData *) append :(const unsigned char *) data :(unsigned long) length;

//
// Prepend data in front of the string
//
// @param data     the data to be prepended
// @param length   the length of the data to be prepended
//
// @return the object
//
- (DData *) prepend :(const unsigned char *) data :(unsigned long) length;

//
// Push one byte at the end of the data string
// 
// @param ch       the byte that is placed at the end
//
// @return the object
//
- (DData *) push :(unsigned char) ch;

//
// Pop one byte from the end of the data string
//
// @return the byte that is pulled (or 0)
//
- (unsigned char) pop;

//
// Multiply the data string a number of times
//
// @param times    the number of times the data string must be multiplied
//
// @return the object
//
- (DData *) multiply :(unsigned) times;


//// Comparable protocol methods

//
// Compare the data string object with another data string object
//
// @param other    the object that should be compared (!= nil)
//
// @return the compare result (-1, 0, 1)
//
- (int) compare :(DData *) other;


//// Comparason methods

//
// Compare the object with a data string
//
// @param data     the data string
// @param length   the length of the data string
//
// @return the compare result (-1, 0, 1)
//
- (int) bcompare :(unsigned char *) data :(unsigned long) length;


//// Search and replace methods

//
// Count the number of occurrences of data in the data string
//
// @param search   the data that should be counted
// @param length   the length of the data to be counted
// @param from     the start index 
// @param to       the end index (to >= from)
//
// @return the number of occurrences
//
- (unsigned long) count :(const unsigned char *) search :(unsigned long) length :(long) from :(long) to;

//
// Return the first index where the search data is found
//
// @param search   the data that should be searched
// @param length   the length of the data
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the first index or -1 = not found
//
- (long) index  :(const unsigned char *) search :(unsigned long) length :(long) from :(long) to;

//
// Return the last index where the search data is found
//
// @param search   the data that should be searched
// @param length   the length of the data
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the last index or -1 = not found
//
- (long) rindex :(const unsigned char *) search :(unsigned long) length :(long) from :(long) to;

//
// Replace in the string the occurrences of old with new, max times
//
// @param old      the data that is searched
// @param olength  the length of the searched data
// @param new      the data that is inserted
// @param nlength  the length of the inserted data
// @param max      the maximum number of replaces (-1 = all)
//
// @return the object
//
- (DData *) replace :(const unsigned char *) old :(unsigned long) olength 
                    :(const unsigned char *) new :(unsigned long) nlength :(long) max;

@end
#endif
