#ifndef _DDBM_H
#define _DDBM_H

// This file was automatically generated by objch.

//==============================================================================
//
//           DDbm - the database manager class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-20 14:03:51 $ $Revision: 1.11 $
//
//==============================================================================
#ifdef HAVE_DDBM
#endif

#include "ofc/config.h"

#include "ofc/DList.h"
#include "ofc/DData.h"

#if defined(HAVE_LIBGDBM) && defined(HAVE_GDBM_H)
#define HAVE_DDBM
#endif

#ifdef HAVE_DDBM

//
// The Dbm class implements a wrapper around the gdbm database manager.
// The naming of the methods is identical to the hashtable class, but this
// class is not a collection. It stores data, not objects. Open modes: r = read, 
// w = write, n = write in new database, c = write in new database if not exist.
//
// @example
// #include <stdio.h>
// #include "ofc/DDbm.h"
// 
// int main(int argc, char *argv[])
// {
//   DDbm  *dbm1 = [DDbm alloc];
//   DDbm  *dbm2 = [DDbm new  ];
//   DData *dat;
//   DText *str;
// 
//   [dbm1 init :"file.dbm" :"c"];        // Init and open with filename, writer
// 
//   if ([dbm1 isOpen])                   // Check if the file is open
//   {
//     if ([dbm1 insert :"key1" :4 :"data1" :5])
//       printf("Data: key1,data1 succesfull inserted.\n");
//     else
//       printf("Error inserting: key1,data1:%d.\n", [dbm1 error]);
// 
//     if ([dbm1 insert :"key2" :4 :"data2" :5])
//       printf("Data: key2,data2 succesfull inserted.\n");
//     else
//       printf("Error inserting: key2,data2:%d.\n", [dbm1 error]);
// 
//     [dbm1 close];
//   }
//   else
//     printf("Error opening file.dbm:%d.\n", [dbm1 error]);
// 
// 
//   if ([dbm2 open :"file.dbm" :"r"])    // Open the file for reading
//   {
//     dat = [dbm2 get :"key2" :4];       // Get the data for key2
// 
//     if (dat != nil)
//     {
//       str = [dat toText];
//       printf("Data for key2:%s.\n", [str cstring]);
//       [str free];
//       [dat free];
//     }
//     else
//       printf("Key2 is not present in the file.\n");
// 
//     if ([dbm2 get :"key3" :4])         // Check for key3
//       printf("Key3 is present in the file.\n");
//     else
//       printf("Key3 is not present in the file.\n");
// 
//     [dbm2 close];
//   }
//   else
//     printf("Error opening file.dbm:%d.\n", [dbm2 error]);
// 
//   [dbm1 free];                         // Cleanup
//   [dbm2 free];
// 
//   return 0;
// }
// 

@interface DDbm : Object
{
@private
  void        *_dbf;   // the database
  char         _mode;  // the database mode
}
#endif

#ifdef HAVE_DDBM

//// Constructors

//
// Initialise without a database
//
// @return the object
//
- (DDbm *) init;

//
// Initialise with a database
// 
// @param name     the database name
// @param mode     the mode for the database (r,w,n,c)
// 
// @return the object
// 
- (DDbm *) init :(const char *) name :(const char *) mode;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
// 
- free;


//// Member methods

//
// Check if the database is open
//
// @return is it ?
//
- (BOOL) isOpen;

//
// Check if the the database is read only ?
//
// @return is it ?
//
- (BOOL) isReadOnly;

//
// Return the last error
//
// @return the last error
//
- (int) error;


//// Main methods

//
// Open a database
//
// @param name     the name of the database
// @param mode     the mode for the database (r,w,c,n)
//
// @return is the dbm open ?
//
- (BOOL) open :(const char *) name :(const char *) mode;

 #ifdef WIN32
 #endif
//
// Insert data in the database (only writer)
// 
// @param key      the key
// @param klen     the length of the key
// @param data     the data
// @param dlen     the length of the data
// 
// @return success
//
- (BOOL) insert :(void *) key :(unsigned) klen :(void *) data :(unsigned) dlen;

//
// Return the data related to a key
// 
// @param key      the key
// @param klen     the length of the key
// 
// @return the data in a new DData object (or nil if not found)
// 
- (DData *) get :(void *) key :(unsigned) klen;

//
// Check if the database has a key
// 
// @param key      the key
// @param klen     the length of the key
// 
// @return is the key present ?
//
- (BOOL) has :(void *) key :(unsigned) klen;

//
// Delete a key,data from the database (only writer)
// 
// @param key      the key
// @param klen     the length of the key
// 
// @return success
//
- (BOOL) delete :(void *) key :(unsigned) klen;

//
// Reorganize the database (only writer)
//
// @return the object
//
- (BOOL) reorganize;

//
// Close the database
// 
// @return the object
// 
- (DDbm *) close;


//// List related methods

//
// Return a list with all the keys in the database
// 
// @return the (new) list with all the keys in DData objects (or nil for error)
// 
- (DList *) keys;

//
// Return a list with all the data in the database
// 
// @return the (new) list with all the data in DData objects (or nil for error)
// 
- (DList *) objects;

@end
#endif
#endif
