#ifndef _DDISCRETEDISTRIBUTION_H
#define _DDISCRETEDISTRIBUTION_H

// This file was automatically generated by objch.

//==============================================================================
//
// DDiscreteDistribution - the (discrete) distribution class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-23 18:59:07 $ $Revision: 1.4 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DScore.h"
#include "ofc/DList.h"

//
// The DDiscreteDistribution class implements a (discrete) distribution class.
// The distribution object is updated with new values. Its task is to determine
// the length, the average or mean, the variance, the standard deviation, the sum 
// and the squared sum of all values in the distribution. For the scores the length 
// and percentage of every score can be determined.
// The distribution object is not a collection: the values are not stored in the
// object; the state of the object is updated with the new value.
// 
// @example
// #include <stdio.h>
// #include "ofc/DDiscreteDistribution.h"
// 
// int main(int argc, char *argv[])
// {
//   DDiscreteDistribution *dis  = [DDiscreteDistribution new];
//   DListIterator         *iter;
// 
//   double d;
// 
//                                        // Setup the discrete ranges
//   for (d = 0.0; d < 5.0; d += 0.5)
//   {
//     [dis range :d :d+0.5];
//   }
// 
//                                        // Set the values for the distribution
//   if (![dis update :0.1]) printf("Value 0.1 is not valid for the distribution.\n");
//   if (![dis update :4.2]) printf("Value 4.2 is not valid for the distribution.\n");
//   if (![dis update :3.6]) printf("Value 4.2 is not valid for the distribution.\n");
//   if (![dis update :2.2]) printf("Value 4.2 is not valid for the distribution.\n");
//   if (![dis update :0.5]) printf("Value 4.2 is not valid for the distribution.\n");
//   if (![dis update :3.0]) printf("Value 4.2 is not valid for the distribution.\n");
//   if (![dis update :2.1]) printf("Value 4.2 is not valid for the distribution.\n");
//   if (![dis update :4.9]) printf("Value 4.2 is not valid for the distribution.\n");
// 
//                                        // Calculate some numbers
//   printf("Number of values in the distribution:%d.\n", [dis length]);
//   printf("Sum    of values in the distribution:%.1f.\n", [dis sum]);
//   printf("SqSum  of values in the distribution:%.1f.\n", [dis sumSquared]);
//   printf("Mean   of values in the distribution:%.1f.\n", [dis mean]);
//   printf("StdDev of values in the distribution:%.1f.\n", [dis standardDeviation]);
//   
//                                        // Scores iterator
//   iter = [dis scores];
//   if (iter != nil)
//   {
//     DScore *sc = [iter first];
// 
//     while (sc != nil)
//     {
//       printf("Score:[%.1f-%0.1f] sum:%.1f sq-sum:%.1f perc:%.1f%%.\n",
//              [sc min], [sc max], [sc sum], [sc sumSquared], [sc percentage]);
//       sc = [iter next];
//     }
//   }
//   else
//     printf("No scores in the distribution ?!\n");
// 
//   [iter free];
//   [dis  free];                         // Cleanup
// 
//   return 0;
// }
// 

@interface DDiscreteDistribution : Object
{
@private
  long                _length;     // the number of values in the distribution
  double              _sum;        // the sum of the values in the distribution
  double              _sumSquared; // the sum of the squared values in the distribution
  DList              *_scores;     // the list with the scores
}


//// Constructors

//
// Initialise the object with no range
//
// @return the object
//
- (DDiscreteDistribution *) init;


//// Copy related method

//
// Do a shallow copy of the distribution object (not implemented)
// 
// @return the object
// 
- shallowCopy;


//// Destructor

//
// Free the distribution
// 
// @return the object
// 
- free;


//// Member methods

//
// Return the number of the values in the distribution
// 
// @return the number
// 
- (int) length;

//
// Return the sum of the values in the distribution
// 
// @return the sum
//
- (double) sum;

//
// Return the sum of the squared values in the distribution
// 
// @return the sum of the squared values
// 
- (double) sumSquared;


//// Set methods

//
// Add a score with a range to the distribution: [min,max>
// 
// @param min      the minimum value for the range
// @param max      the maximum value for the range
// 
// @return success
// 
- (BOOL) range :(double) min :(double) max;

//
// Add a score with a (discrete) range to the distribution: [val-0.5,val+0.5>
// 
// @param value    the discrete value for the range
// 
// @return success
// 
- (BOOL) range :(int) value;

//
// Reset the length, sums and scores to zero
// 
// @return the object
// 
- (DDiscreteDistribution *) reset;

//
// Update the distribution with a value
// 
// @param value    the update value
// 
// @return success (is a score updated?)
// 
- (BOOL) update :(double) value;


//// Distribution Calculation methods

//
// Calculate the mean of the values
// 
// @return the mean value
// 
- (double) mean;

//
// Calculate the average of the values
// 
// @return the average value
// 
- (double) average;

//
// Calculate the variance of the values
// 
// @return the variance
// 
- (double) variance;

//
// Calculate the standard deviation of the values
// 
// @return the standard deviation
// 
- (double) standardDeviation;


//// Score methods

//
// Return an iterator on the scores in the distribution
// 
// @return a (new) iterator on the list of the DScore objects
// 
- (DListIterator *) scores;

@end
#endif
