#ifndef _DDOUBLEARRAY_H
#define _DDOUBLEARRAY_H

// This file was automatically generated by objch.

//==============================================================================
//
//  DDoubleArray - the one dimensional array of doubles class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-24 18:02:46 $ $Revision: 1.3 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DComparable.h"
#include "ofc/DSizeable.h"
#include "ofc/DParsable.h"

#include "ofc/DText.h"

//
// The DDoubleArray class implements an one dimensional array of double. It is of 
// course possible to create an array of double with the classes DArray and DDouble, 
// but this is not very efficient and a little bit clumsy. This class also implements some
// statistical methods.
// 
// @example
// #include <stdio.h>
// #include "ofc/DDoubleArray.h"
// 
// int main(int argc, char *argv[])
// {
//   DDoubleArray *dar1  = [DDoubleArray alloc];
//   DDoubleArray *dar2  = [DDoubleArray new  ];
//   DText         *str;
// 
//   double doubles1[] = { 10.0, 20.0, 30.0 };
//   double doubles2[] = { 50.0, 30.0, 10.0 };
// 
//                                        // Init with an array of doubles
//   [dar1 init :doubles1 :(sizeof(doubles1) / sizeof(double))];
// 
//   [dar2 size :50];                     // Resize the array
// 
//   printf("Length of array1:%ld.\n", [dar1 length]);
// 
//   str = [dar1 toText];
//   printf("Array as string:%s.\n", [str cstring]);
//   [str free];
// 
//                                        // Insert and delete values
//   [dar2 push :40.0];
//   [dar2 push :50.0];
//   [dar2 push :60.0];
// 
//   [dar2 insert :1 :70];
//   [dar2 delete :2];
// 
//   str = [dar2 toText];
//   printf("Array (after Insert,delete) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Insert and delete ranges of values
//   [dar2 insert :2 :1 :doubles1 :(sizeof(doubles1) / sizeof(double))];
//   [dar2 delete :3 :4];
//   
//   str = [dar2 toText];
//   printf("Array (after range insert and delete) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Append and prepend
//   [dar2 append  :doubles2 :(sizeof(doubles2)/sizeof(double))];
//   [dar2 prepend :doubles1 :(sizeof(doubles1)/sizeof(double))];
//   
//   str = [dar2 toText];
//   printf("Array (after append and prepend) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Sort the array
//   [dar2 sort :0 :-1];
// 
//   str = [dar2 toText];
//   printf("Array (after sort) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Count a double value
//   printf("Value 70.0 is %ld times present in the array.\n", [dar2 count :70.0 :0 :-1]);
//                                        // Statistical methods
//   printf("Sum of the values in the array:%.1f.\n", [dar2 sum :0 :-1]);
//   printf("Max of the values in the array:%.1f.\n", [dar2 max :0 :-1]);
//   printf("Min of the values in the array:%.1f.\n", [dar2 min :0 :-1]);
//   printf("Avg of the values in the array:%.1f.\n", [dar2 average :0 :-1]);
//   printf("StD of the values in the arary:%.1f.\n", [dar2 standardDeviation :0 :-1]);
// 
//                                        // Compare the arrays
//   if ([dar1 compare :dar2] == 0)
//     printf("Array1 is equal to array2.\n");
//   else
//     if ([dar1 compare :dar2] < 0)
//       printf("Array1 is smaller than array2.\n");
//     else
//       printf("Array1 is greater than array2.\n");
// 
//                                        // Use as fifo
//   [dar2 clear];
// 
//   [dar2 enqueue :3.0];
//   [dar2 enqueue :4.0];
// 
//   printf("Fifo dequeue:%.1f.\n", [dar2 dequeue]);
// 
//                                        // Use as lifo
//   [dar2 clear];
// 
//   [dar2 push :5.0];
//   [dar2 push :7.0];
// 
//   printf("Lifo pop:%.1f.\n", [dar2 pop]);
//   
//   [dar1 free];                         // Cleanup
//   [dar2 free];
// 
//   return 0;
// }
// 

@interface DDoubleArray : Object <DTextable,DComparable,DSizeable,DParsable>
{
@private
  unsigned long  _size;     // the maximum length of the double array
  unsigned       _extra;    // the extra size amount
  unsigned long  _length;   // the current length of the double array
  double        *_array;    // the array of doubles
}


//// Constructors

//
// Initialise to an empty array
//
// @return the object
//
- (DDoubleArray *) init;

//
// Initialise the object with an array of doubles
//
// @param doubles  an array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) init :(const double *) doubles :(unsigned long) length;


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free;


//// Member methods

//
// Check if the array is empty
//
// @return Is it empty ?
//
- (BOOL) isEmpty;

//
// Return the length of the array
//
// @return the length
//
- (unsigned long) length;

//
// Return the pointer to the array of doubles
//
// @return the pointer to the array
//
- (const double *) array;


//// Size methods

//
// Insure the size for the array
//
// @param size     the requested size (in elements) for the array
//
// @return the object
//
- (DDoubleArray *) size :(unsigned long) size;

//
// Set the amount of extra size increase if the size must be extended
//
// @param extra    the amount (def. 0)
//
// @return the object
//
- (DDoubleArray *) extra :(unsigned) extra;


//// DTextable protocol methods

//
// Convert the object to a text string 
// 
// @return a (new) text string object
//
- (DText *) toText;


//// DParsable protocol methods

//
// Parse a string for an array of doubles: <double>,<double>;<double>
// 
// @param cstr     the string to be parsed (moved to the first non-parsable character)
// 
// @return the result (0, ERANGE, ENODATA)
//
- (int) fromString :(char **) cstr;


//// Array setters

//
// Clear the array
// 
// @return the object
// 
- (DDoubleArray *) clear;

//
// Set an array of doubles in the object
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) set :(const double *) doubles :(unsigned long) length;


//// Element methods

//
// Put a double at the given index
//
// @param index     the index (-length..+length)
// @param value     the double
//
// @return the object
//
- (DDoubleArray *) put :(long) index :(double) value  ;

//
// Get the double at the given index
//
// @param index    the index 
//
// @return the double from the array
//
- (double) get :(long) index;

// 
// Insert a double at the given index
// 
// @param index    the index
// @param value    the double
// 
// @return the object
//
- (DDoubleArray *) insert :(long) index :(double) value;

//
// Delete the double at the given index
//
// @param index    the index
//
// @return the object
//
- (DDoubleArray *) delete :(long) index;


//// Slice methods

//
// Insert an array of doubles in a part of the array in the object
//
// @param from     the start index
// @param to       the end index (to >= (from-1))
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) insert :(long) from :(long) to :(const double *) doubles :(unsigned long) length;

//
// Return a sub array
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return a (new) double array object
//
- (DDoubleArray *) get :(long) from :(long) to;

//
// Delete a range in the array
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the object
//
- (DDoubleArray *) delete :(long) from :(long) to;


//// Array manipulation methods

//
// Append an array to the object
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) append :(const double *) doubles :(unsigned long) length;

//
// Prepend an array in front of the object
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) prepend :(const double *) doubles :(unsigned long) length;


//// Stack methods

//
// Push one double at the end of the array
// 
// @param value    the double
//
// @return the object
//
- (DDoubleArray *) push :(double) value;

//
// Pop one double from the end of the array
//
// @return the double that is pulled (or 0.0 for empty stack)
//
- (double) pop;

//
// Return the double at the end of the array
// 
// @return the double (or 0.0 for empty stack)
// 
- (double) tos;


//// Queue methods (FIFO)

//
// Put a double at the start of the array 
//
// @param value    the double 
// 
// @return success
// 
- (BOOL) enqueue :(double) value;

//
// Get the double at the end of the array
// 
// @return the double (or 0.0 for empty queue)
// 
- (double) dequeue;


//// Comparable protocol methods

//
// Compare the double array object with another double array object
//
// @param other    the other object (!= nil)
//
// @return the compare result (-1, 0, 1)
//
- (int) compare :(DDoubleArray *) other;


//// Comparison methods

//
// Compare the object with an array of doubles
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the compare result (-1, 0, 1)
//
- (int) bcompare :(const double *) doubles :(unsigned long) length;


//// Statistical methods

//
// Count the number of occurrences of a double in the object
//
// @param search   the count value
// @param from     the start index 
// @param to       the end index (to >= from)
//
// @return the number of occurrences
//
- (unsigned long) count :(double) search :(long) from :(long) to  ;

//
// Return the first index where the search double is found
//
// @param search   the search value
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the first index or -1 = not found
//
- (long) index :(double) search :(long) from :(long) to ;

//
// Return the last index where the search double is found
//
// @param search   the search value
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the last index or -1 = not found
//
- (long) rindex :(double) search :(long) from :(long) to;

//
// Calculate the sum in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the sum of the doubles
// 
- (double) sum :(long) from :(long) to;

//
// Determine the maximum double in a range of the double array
//
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the maximum double
// 
- (double) max :(long) from :(long) to;

//
// Determine the minimum double in a range of the double array
//
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the minimum double
// 
- (double) min :(long) from :(long) to;

//
// Calculate the average in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the average (or 0.0)
// 
- (double) average :(long) from :(long) to;

// 
// Calculate the variance in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the variance
// 
- (double) variance :(long) from :(long) to;

//
// Calculate the standard deviation in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the standard deviation
// 
- (double) standardDeviation :(long) from :(long) to;

//
// Sort a range in the double array
//
// @param from     the start index
// @param to       the end index (to > from)
// 
// @return the object
// 
- (DDoubleArray *) sort :(long) from :(long) to;

//
// Invert (mirror) a range in the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the object
// 
- (DDoubleArray *) invert :(long) from :(long) to;

@end
#endif
