//==============================================================================
//
//  DDoubleArray - the one dimensional array of doubles class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-24 18:02:46 $ $Revision: 1.3 $
//
//==============================================================================

#include "ofc/DDoubleArray.h"

#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <ctype.h>


#if _INTERFACE_

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DComparable.h"
#include "ofc/DSizeable.h"
#include "ofc/DParsable.h"

#include "ofc/DText.h"

//
// The DDoubleArray class implements an one dimensional array of double. It is of 
// course possible to create an array of double with the classes DArray and DDouble, 
// but this is not very efficient and a little bit clumsy. This class also implements some
// statistical methods.
// 
// @example
// #include <stdio.h>
// #include "ofc/DDoubleArray.h"
// 
// int main(int argc, char *argv[])
// {
//   DDoubleArray *dar1  = [DDoubleArray alloc];
//   DDoubleArray *dar2  = [DDoubleArray new  ];
//   DText         *str;
// 
//   double doubles1[] = { 10.0, 20.0, 30.0 };
//   double doubles2[] = { 50.0, 30.0, 10.0 };
// 
//                                        // Init with an array of doubles
//   [dar1 init :doubles1 :(sizeof(doubles1) / sizeof(double))];
// 
//   [dar2 size :50];                     // Resize the array
// 
//   printf("Length of array1:%ld.\n", [dar1 length]);
// 
//   str = [dar1 toText];
//   printf("Array as string:%s.\n", [str cstring]);
//   [str free];
// 
//                                        // Insert and delete values
//   [dar2 push :40.0];
//   [dar2 push :50.0];
//   [dar2 push :60.0];
// 
//   [dar2 insert :1 :70];
//   [dar2 delete :2];
// 
//   str = [dar2 toText];
//   printf("Array (after Insert,delete) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Insert and delete ranges of values
//   [dar2 insert :2 :1 :doubles1 :(sizeof(doubles1) / sizeof(double))];
//   [dar2 delete :3 :4];
//   
//   str = [dar2 toText];
//   printf("Array (after range insert and delete) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Append and prepend
//   [dar2 append  :doubles2 :(sizeof(doubles2)/sizeof(double))];
//   [dar2 prepend :doubles1 :(sizeof(doubles1)/sizeof(double))];
//   
//   str = [dar2 toText];
//   printf("Array (after append and prepend) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Sort the array
//   [dar2 sort :0 :-1];
// 
//   str = [dar2 toText];
//   printf("Array (after sort) as string:%s.\n", [str cstring]);
//   [str free];
//                                        // Count a double value
//   printf("Value 70.0 is %ld times present in the array.\n", [dar2 count :70.0 :0 :-1]);
//                                        // Statistical methods
//   printf("Sum of the values in the array:%.1f.\n", [dar2 sum :0 :-1]);
//   printf("Max of the values in the array:%.1f.\n", [dar2 max :0 :-1]);
//   printf("Min of the values in the array:%.1f.\n", [dar2 min :0 :-1]);
//   printf("Avg of the values in the array:%.1f.\n", [dar2 average :0 :-1]);
//   printf("StD of the values in the arary:%.1f.\n", [dar2 standardDeviation :0 :-1]);
// 
//                                        // Compare the arrays
//   if ([dar1 compare :dar2] == 0)
//     printf("Array1 is equal to array2.\n");
//   else
//     if ([dar1 compare :dar2] < 0)
//       printf("Array1 is smaller than array2.\n");
//     else
//       printf("Array1 is greater than array2.\n");
// 
//                                        // Use as fifo
//   [dar2 clear];
// 
//   [dar2 enqueue :3.0];
//   [dar2 enqueue :4.0];
// 
//   printf("Fifo dequeue:%.1f.\n", [dar2 dequeue]);
// 
//                                        // Use as lifo
//   [dar2 clear];
// 
//   [dar2 push :5.0];
//   [dar2 push :7.0];
// 
//   printf("Lifo pop:%.1f.\n", [dar2 pop]);
//   
//   [dar1 free];                         // Cleanup
//   [dar2 free];
// 
//   return 0;
// }
// 

@interface DDoubleArray : Object <DTextable,DComparable,DSizeable,DParsable>
{
@private
  unsigned long  _size;     // the maximum length of the double array
  unsigned       _extra;    // the extra size amount
  unsigned long  _length;   // the current length of the double array
  double        *_array;    // the array of doubles
}

#endif



@implementation DDoubleArray


#if _PRIVATE_

//
// Check if two doubles are equal
// 
// @param d1    the first double
// @param d2    the second double
// 
// @return are the doubles equal ?
// 
#define _equal(d1,d2)  !(((d1) < (d2)) || ((d1) > (d2)))
  
//
// Convert the index to an array offset
// 
// @param array    the array with the doubles
// @param index    the index in the array (-_length .. _length)
// 
// @return the offset in the array (0.._length)
// 
static long index2offset(DDoubleArray *array, int index)
{
  if (index < 0)
    index = array->_length + index;

  if (index < 0)
  {
    index  = 0;

    WARNING(DW_INVALID_ARG, "index");
  }
  if (index > array->_length)
  {  
    index = array->_length-1;

    WARNING(DW_INVALID_ARG, "index");
  }

  return index;
}
#endif


//// Constructors

//
// Initialise to an empty array
//
// @return the object
//
- (DDoubleArray *) init
{
  [super init];
  
  _length = 0;
  _size   = 0;
  _extra  = 0;
  _array  = NULL;
    
  return self;
}

//
// Initialise the object with an array of doubles
//
// @param doubles  an array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) init :(const double *) doubles :(unsigned long) length
{
  [self init];

  [self set :doubles :length];
  
  return self;
}


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy
{
  DDoubleArray *copy = [super shallowCopy];
  
  copy->_array = (double *) objc_malloc(sizeof(double) * _size);
  
  memcpy(copy->_array, _array, sizeof(double) * _length);
  
  return copy;
}


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free
{
  if (_array != NULL)
  {
    objc_free(_array);
  }

  return [super free];
}


//// Member methods

//
// Check if the array is empty
//
// @return Is it empty ?
//
- (BOOL) isEmpty
{
  return (_length == 0);
}

//
// Return the length of the array
//
// @return the length
//
- (unsigned long) length
{
  return _length;
}

//
// Return the pointer to the array of doubles
//
// @return the pointer to the array
//
- (const double *) array
{
  return _array;
}


//// Size methods

//
// Insure the size for the array
//
// @param size     the requested size (in elements) for the array
//
// @return the object
//
- (DDoubleArray *) size :(unsigned long) size
{
  if (size  > _size)
  {
    _size = size + _extra;
    
    if (_array == NULL)
    {
      _array = (double *) objc_malloc(sizeof(double) * _size);
    }
    else
    {
      _array = (double *) objc_realloc(_array, sizeof(double) * _size);
    }
  }

  return self;
}

//
// Set the amount of extra size increase if the size must be extended
//
// @param extra    the amount (def. 0)
//
// @return the object
//
- (DDoubleArray *) extra :(unsigned) extra
{
  _extra = extra;

  return self;
}


//// DTextable protocol methods

//
// Convert the object to a text string 
// 
// @return a (new) text string object
//
- (DText *) toText
{
  DText *text = [DText new];
  
  unsigned long index = 0;
  
  if (_length > 0)
  {
    DText *temp = [DText new];
    
    while (index < (_length-1))
    {
      [temp format :"%g",_array[index++]];
    
      [text append :[temp cstring]];
      [text push   :','];
    }
    
    [temp format :"%g", _array[index]];
    [text append :[temp cstring]];
    
    [temp free];
  }
  
  return text;
}


//// DParsable protocol methods

//
// Parse a string for an array of doubles: <double>,<double>;<double>
// 
// @param cstr     the string to be parsed (moved to the first non-parsable character)
// 
// @return the result (0, ERANGE, ENODATA)
//
- (int) fromString :(char **) cstr
{
  int     result = 0;
  char   *src    = *cstr;
  char   *end    = src;
  char   *last   = src;
  double  number;
  BOOL    done   = NO;
  
  [self clear];
  
  while (!done)
  {
    last = src;
    
    number = strtod(src, &src);
    
    if (last == src)
    {
      done = YES;
    }
    else if (errno == ERANGE)
    {
      result = ERANGE;
      done   = YES;
    }
    else
    {
      [self size :_length+1];
      
      _array[_length++] = number;
      
      while (isspace(*src))
      {
        src++;
      }
      end = src;

      if ((*src == ',') || (*src == ';'))
      {
        src++;
      }
      else
      {
        done = YES;
      }
    }
  }
  
  if (end == *cstr)
  {
    result = ENODATA;
  }
  else
  {
    *cstr = end;
  }
  
  return result;
}


//// Array setters

//
// Clear the array
// 
// @return the object
// 
- (DDoubleArray *) clear
{
  _length = 0;
  
  return self;
}

//
// Set an array of doubles in the object
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) set :(const double *) doubles :(unsigned long) length
{
  _length = 0;

  if (doubles != NULL)
  {
    _length = length;
  }

  [self size :_length];

  memcpy(_array, doubles, sizeof(double) * _length);
  
  return self;
}


//// Element methods

//
// Put a double at the given index
//
// @param index     the index (-length..+length)
// @param value     the double
//
// @return the object
//
- (DDoubleArray *) put :(long) index :(double) value  
{
  if (_length > 0)
  {
    index = index2offset(self, index);

    _array[index] = value;
  }
  
  return self;
}

//
// Get the double at the given index
//
// @param index    the index 
//
// @return the double from the array
//
- (double) get :(long) index
{
  if (_length > 0)
  {
    index = index2offset(self, index);
  
    return _array[index];
  }
  else
    return 0.0;
}

// 
// Insert a double at the given index
// 
// @param index    the index
// @param value    the double
// 
// @return the object
//
- (DDoubleArray *) insert :(long) index :(double) value
{
  [self size :_length+1];
  
  index = index2offset(self, index);
  
  memmove(_array+index+1, _array+index, sizeof(double) * (_length-index));
  
  _array[index] = value;
  
  _length++;
  
  return self;
}
    
//
// Delete the double at the given index
//
// @param index    the index
//
// @return the object
//
- (DDoubleArray *) delete :(long) index
{
  index = index2offset(self, index);
  
  if (_length > 0)
  {
    memmove(_array+index, _array+index+1, sizeof(double) * (_length-index-1));

    _length--;
  }
  
  return self;
}


//// Slice methods

//
// Insert an array of doubles in a part of the array in the object
//
// @param from     the start index
// @param to       the end index (to >= (from-1))
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) insert :(long) from :(long) to :(const double *) doubles :(unsigned long) length
{
  from = index2offset(self, from);
  to   = index2offset(self, to  );

  if (to >= (from-1))
  {
    length = (doubles != NULL) ? length : 0;

    [self size :(_length+length-(to-from+1))];
    
    memmove(_array+from+length, _array+to+1, sizeof(double) * (_length-to-1));
    
    memmove(_array+from, doubles, sizeof(double) * length);

    _length += (length-(to-from+1));
  }
    
  return self;
}

//
// Return a sub array
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return a (new) double array object
//
- (DDoubleArray *) get :(long) from :(long) to
{
  DDoubleArray *sub = [DDoubleArray new];
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );

  if (to >= from)
  {
    [sub set :_array+from :(to - from + 1)];
  }
  
  return sub;
}

//
// Delete a range in the array
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the object
//
- (DDoubleArray *) delete :(long) from :(long) to
{
  if (_length > 0)
  {
    from = index2offset(self, from);
    to   = index2offset(self, to  );

    if (to > from)
    {
      memmove(_array+from, _array+to+1, sizeof(double) * (_length-to-1));

      _length -= (to-from+1);
    }
  }
  
  return self;
}


//// Array manipulation methods

//
// Append an array to the object
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) append :(const double *) doubles :(unsigned long) length
{
  length = (doubles != NULL) ? length : 0;

  if (length > 0)
  {
    [self size :(_length + length)];

    memcpy(_array + _length, doubles, sizeof(double) * length);
    
    _length += length;
  }
  
  return self;
}

//
// Prepend an array in front of the object
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the object
//
- (DDoubleArray *) prepend :(const double *) doubles :(unsigned long) length
{
  length = (doubles != NULL) ? length : 0;

  if (length > 0)
  {
    [self size :(_length + length)];

    memmove(_array + length, _array, sizeof(double) * _length);
    
    memcpy(_array, doubles, sizeof(double) * length);

    _length += length;
  }

  return self;
}


//// Stack methods

//
// Push one double at the end of the array
// 
// @param value    the double
//
// @return the object
//
- (DDoubleArray *) push :(double) value
{
  [self size :(_length + 1)];

  _array[_length++] = value;

  return self;
}

//
// Pop one double from the end of the array
//
// @return the double that is pulled (or 0.0 for empty stack)
//
- (double) pop
{
  double value = 0.0;
  
  if (_length > 0)
  {
    value = _array[--_length];
  }

  return value;
}

//
// Return the double at the end of the array
// 
// @return the double (or 0.0 for empty stack)
// 
- (double) tos
{
  if (_length > 0)
  {
    return _array[_length-1];
  }
  else
  {
    return 0.0;
  }
}


//// Queue methods (FIFO)

//
// Put a double at the start of the array 
//
// @param value    the double 
// 
// @return success
// 
- (BOOL) enqueue :(double) value
{
  [self size :(_length + 1)];

  memmove(_array + 1, _array, sizeof(double) * _length);
  
  _array[0] = value;
  
  _length++;
  
  return YES;
}

//
// Get the double at the end of the array
// 
// @return the double (or 0.0 for empty queue)
// 
- (double) dequeue
{
  double value = 0.0;
  
  if (_length > 0)
  {
    value = _array[--_length];
  }

  return value;
}


//// Comparable protocol methods

//
// Compare the double array object with another double array object
//
// @param other    the other object (!= nil)
//
// @return the compare result (-1, 0, 1)
//
- (int) compare :(DDoubleArray *) other
{
  if (other != nil)
  {
    double *d1                 = _array;
    const double *d2           = [other array];
    long l1                 = _length;
    long l2                 = [other length];
  
    while ((l1 > 0) && (l2 > 0) && (_equal(*d1,*d2)))
    {
      d1++; d2++; l1--; l2--;
    }
  
    if ((l1 > 0) && (l2 > 0))
    {
      return (*d1 > *d2) ? 1 : -1;
    }
    else if (l1 > 0)
    {
      return 1;
    }
    else if (l2 > 0)
    {
      return -1;
    }
  }
  else
  {
    WARNING(DW_NIL_NOT_ALLOWED, "other");
  }
  
  return 0;
}


//// Comparison methods

//
// Compare the object with an array of doubles
//
// @param doubles  the array with the doubles
// @param length   the length of the array
//
// @return the compare result (-1, 0, 1)
//
- (int) bcompare :(const double *) doubles :(unsigned long) length
{
  const double *d1        = _array;
  long l1                 = _length;
  const double *d2        = doubles;
  long l2                 = (doubles != NULL) ? length : 0;
  
  while ((l1 > 0) && (l2 > 0) && (_equal(*d1,*d2)))
  {
    d1++; d2++; l1--; l2--;
  }
  
  if ((l1 > 0) && (l2 > 0))
  {
    return (*d1 > *d2) ? 1 : -1;
  }
  else if (l1 > 0)
  {
    return 1;
  }
  else if (l2 > 0)
  {
    return -1;
  }
  
  return 0;
}


//// Statistical methods

//
// Count the number of occurrences of a double in the object
//
// @param search   the count value
// @param from     the start index 
// @param to       the end index (to >= from)
//
// @return the number of occurrences
//
- (unsigned long) count :(double) search :(long) from :(long) to  
{
  unsigned long  cnt  = 0;
  double        *pntr;
  
  from = index2offset(self, from);
  to   = index2offset(self ,to);

  pntr = _array + from;
  
  while (from <= to)
  {
    if (_equal(search,*pntr))
    {
      cnt++;
    }
    
    pntr++;
    from++;
  }
      
  return cnt;
}

//
// Return the first index where the search double is found
//
// @param search   the search value
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the first index or -1 = not found
//
- (long) index :(double) search :(long) from :(long) to 
{
  double *pntr;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );
  
  pntr = _array + from;
    
  while (from <= to)
  {
    if (_equal(search,*pntr))
    {
      return from;
    }
    from++;
    pntr++;
  }

  return -1;
}

//
// Return the last index where the search double is found
//
// @param search   the search value
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the last index or -1 = not found
//
- (long) rindex :(double) search :(long) from :(long) to
{
  double *pntr;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );

  pntr = _array + to;
  
  while (to >= from)
  {
    if (_equal(search,*pntr))
    {
      return to;
    }
    to--;
    pntr--;
  }
  
  return -1;
}

//
// Calculate the sum in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the sum of the doubles
// 
- (double) sum :(long) from :(long) to
{
  double  *pntr;
  double   sum = 0.0;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );
  
  pntr = _array + from;
    
  while (from <= to)
  {
    sum += *pntr;
    
    from++;
    pntr++;
  }

  return sum;
}

//
// Determine the maximum double in a range of the double array
//
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the maximum double
// 
- (double) max :(long) from :(long) to
{
  double *pntr;
  double  max = DBL_MIN;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );
  
  pntr = _array + from;
    
  while (from <= to)
  {
    if (*pntr > max)
    {
      max = *pntr;
    }
    
    from++;
    pntr++;
  }

  return max;
}

//
// Determine the minimum double in a range of the double array
//
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the minimum double
// 
- (double) min :(long) from :(long) to
{
  double *pntr;
  double  min = DBL_MAX;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );
  
  pntr = _array + from;
    
  while (from <= to)
  {
    if (*pntr < min)
    {
      min = *pntr;
    }
    
    from++;
    pntr++;
  }

  return min;
}

//
// Calculate the average in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the average (or 0.0)
// 
- (double) average :(long) from :(long) to
{
  double *pntr;
  double  avg    = 0.0;
  long    length = 0;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );
  
  pntr = _array + from;
    
  while (from <= to)
  {
    avg += (double) *pntr;
    
    length++;
    from++;
    pntr++;
  }
  
  if (length > 0)
  {
    avg /= (double) length;
  }
  
  return avg;
}

// 
// Calculate the variance in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the variance
// 
- (double) variance :(long) from :(long) to
{
  double  variance = 0.0;
  double  avg      = 0.0;
  long    length   = 0;
  double *pntr;
  long    walk;
  
  from = index2offset(self, from);
  to   = index2offset(self, to  );
  
  pntr = _array + from;
  
  walk = from;
  while (walk <= to)
  {
    avg += (double) *pntr;
    
    length++;
    walk++;
    pntr++;
  }
  
  if (length > 0)
  {
    avg /= (double) length;
    
    pntr = _array + from;
    
    walk = from;
    while (walk <= to)
    {
      double diff = ((double) *pntr - avg);
      
      variance += (diff * diff);
      
      walk++;
      pntr++;
    }
    
    variance /= (double) length;
  }
  
  return variance;
}

//
// Calculate the standard deviation in a range of the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the standard deviation
// 
- (double) standardDeviation :(long) from :(long) to
{
  double result = [self variance :from :to];
  
  if (result > 0.0)
  {
    result = sqrt(result);
  }
  
  return result;
}


#if _PRIVATE_

//
// Compare function for sort method
// 
// @param first    the pointer to the first element
// @param second   the pointer to the second element
// 
// @return the compare result -1,0,1)
// 
static int _compare(const void *c1, const void *c2)
{
  double d1 = * ((double *) c1);
  double d2 = * ((double *) c2);
  
  if (d1 > d2)
  {
    return 1;
  }
  else if (d1 < d2)
  {
    return -1;
  }
  else
  {
    return 0;
  }
}
#endif

//
// Sort a range in the double array
//
// @param from     the start index
// @param to       the end index (to > from)
// 
// @return the object
// 
- (DDoubleArray *) sort :(long) from :(long) to
{
  from = index2offset(self, from);
  to   = index2offset(self, to  );

  if (to > from)
  {
    qsort(_array + from, (to - from + 1), sizeof(double), _compare);
  }
  
  return self;
}

//
// Invert (mirror) a range in the double array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the object
// 
- (DDoubleArray *) invert :(long) from :(long) to
{
  from  = index2offset(self, from);
  to    = index2offset(self, to  );
  
  while (to > from)
  {
    double tmp = _array[to];
    
    _array[to] = _array[from];
    
    _array[from] = tmp;
    
    to--;
    from++;
  }
  
  return self;
}

@end

/*===========================================================================*/

