#ifndef _DFSM_H
#define _DFSM_H

// This file was automatically generated by objch.

//==============================================================================
//
//        DFSM - the finite state machine class in the ofc-library
//
//            Copyright (C) 2003-2006  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-25 16:35:58 $ $Revision: 1.13 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DBitArray.h"
#include "ofc/DList.h"


@class DFSMState;

//
// The DFSMTrans class implements a transition in the finite state machine. 
// This class is privately used by DFSM. The class stores the trigger and 
// the destination label.
//

@interface DFSMTrans : Object
{
@private
  DBitArray *_trigger;       // the range of events that trigger the transition
  DFSMState *_destination;   // the destination of the transition (reference !)
}


//// Constructors

//
// Initialise to a default transition (not allowed)
//
// @return the object
//
- (DFSMTrans *) init;

//
// Initialise a finite state machine transition (used by DFSM)
//
// @param trigger     the range of events that trigger the transition
// @param destination the destination of the transition
//
// @return the object
//
- (DFSMTrans *) init :(DBitArray *) trigger :(DFSMState *) destination;


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free;


//// Main method

//
// Feed the transition an event (used by DFSM)
//
// @param event    the event
//
// @return the state (or nil if the transition didn't trigger)
//
- (DFSMState *) feed :(int) event;

@end

//
// The DFSMState class implements a state of the finite state machine. The
// class stores the transitions and the any-transition. Special purpose
// states should inherit from FSMState.
//


@interface DFSMState : Object
{
@private
  DList     *_transitions;   // all the transitions for the state
  DFSMState *_any;           // the any-transition-state (reference)
}


//// Constructors

//
// Initialise a finite state machine state
//
// @return the object
//
- (DFSMState *) init;


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object (including the transitions)
//
// @return the object
//
- free;


//// State manipulation

//
// Add a transition to the state (used by DFSM)
//
// @param trigger     the range of events that trigger the transition (or nil)
// @param destination the next state if the transition is triggered
//
// @return the object
//
- (DFSMState *) transition :(DBitArray *) trigger :(DFSMState *) destination;

//
// Feed the state with an event (used by DFSM)
//
// @param event    the event for the state
//
// @return the next state (or nil if no transition was found for the event)
//
- (DFSMState *) feed :(int) event;

@end
  
//
// The DFSM class implements a (simple) finite state machine. The state machine is
// built by creating a number of transitions for every state. After the state
// machine is built, the processing is started by using the 'start' method.
// From that moment on, events can be fed to the state machine by calling
// 'feed'. This method returns the current state. With the method 'try' an event
// can be tested for the state machine, without changing the internal state of the
// machine. There is one assumption: events are ints.
//
// @example
// #include <stdio.h>
// #include "ofc/DFSM.h"
// 
// enum
// {
//   INSERT_CARD,
//   ENTER_PIN,
//   REFUSE_PIN,
//   ENTER_AMOUNT,
//   ACCEPT_MONEY,
//   REMOVE_CARD
// };
// 
// int main(int argc, char *argv[])
// {
//   DFSM *fsm = [DFSM new];
// 
//   DFSMState *waitCard    = [DFSMState new]; // State: wait for card
//   DFSMState *waitPin     = [DFSMState new]; // State: wait for pin
//   DFSMState *waitAmount  = [DFSMState new]; // State: wait for choice amount money
//   DFSMState *waitAccept  = [DFSMState new]; // State: wait for acceptance money
//   DFSMState *waitRemoval = [DFSMState new]; // State: wait for removal card
// 
//   DBitArray *events      = [DBitArray alloc]; // the events bit array
// 
//   [events init :INSERT_CARD :REMOVE_CARD]; // Init the events bit array 
// 
//                                        // Setup the state machine
//   [events reset]; [events set :INSERT_CARD];
//   [fsm transition :waitCard :[events copy] :waitPin];
//   
//   [events reset]; [events set :ENTER_PIN];
//   [fsm transition :waitPin :[events copy] :waitAmount];
// 
//   [events reset]; [events set :REFUSE_PIN];
//   [fsm transition :waitPin :[events copy] :waitRemoval];
// 
//   [events reset]; [events set :ENTER_AMOUNT];
//   [fsm transition :waitAmount :[events copy] :waitAccept];
// 
//   [events reset]; [events set :ACCEPT_MONEY];
//   [fsm transition :waitAccept :[events copy] :waitRemoval];
// 
//   [events reset]; [events set :REMOVE_CARD];
//   [fsm transition :waitRemoval :[events copy] :waitCard];
// 
//                                        // Start using the state machine
//   [fsm start :waitCard];
//                                        // Feed the state machine with events
//   if ([fsm feed :INSERT_CARD] != nil)
//     printf("Event INSERT_CARD accepted.\n");
//   else 
//     printf("Event INSERT_CARD was not allowed in the current state.\n");
// 
//                                        // Try an event in the current state
//   printf("Event: ENTER_PIN is %s accaptable in the current state.\n", ([fsm try :ENTER_PIN] != nil ? "" : "not"));
// 
//   if ([fsm feed :REFUSE_PIN] == waitRemoval)
//     printf("Event REFUSE_PIN accepted and fsm is moved to waitRemoval.\n");
//   else
//     printf("Event REFUSE_PIN not accepted.\n");
// 
//   [fsm free];                          // Cleanup fsm, states and transations
//   [events free];
// 
//   return 0;
// }
// 

@interface DFSM : Object
{
@private
  DFSMState  *_current;   // the current state
  DFSMState  *_previous;  // the previous state
  DList      *_states;    // all known states
}


//// Constructors

//
// Initialise an empty finite state machine
//
// @return the object
//
- (DFSM *) init;


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;

//
// Deepen the copy of the object
//
// @return the object
//
- deepen;


//// Deconstructor

//
// Free the FSM (but *NOT* the states)
//
// @return the object
//
- shallowFree;

//
// (Deep) free the FSM (including the states)
//
// @return the object
//
- free;


//// Member methods

//
// Get the current state
//
// @return the current state (or nil for no state)
//
- (DFSMState *) current;

//
// Get the previous state
//
// @return the previous state (or nil for no state)
//
- (DFSMState *) previous;

//
// Check if the state is changed during the last event
//
// @return is it ?
//
- (BOOL) isChanged;


//// Create State machine network

//
// Add a transition to the state machine network. Special case: if trigger is
// nil then this transition is activated if all other transitions fail for an
// event (any-other-transition). Only one such transition is allowed. Note: trigger
// will be freed when FSMState object is freed.
//
// @param origin      the state that has this transition
// @param trigger     the range of events that trigger the transition (or nil)
// @param destination the next state if the transition is triggered
//
// @return the object
//
- (DFSM *) transition :(DFSMState *) origin :(DBitArray *) trigger :(DFSMState *) destination;


//// Using the state machine

//
// Start the state machine
//
// @param state    the start state
//
// @return the current state (or nil if start event not found)
//
- (DFSMState *) start :(DFSMState *) state;

//
// Feed an event to the state machine
//
// @param event    the event for the state machine
//
// @return the current state (or nil for unexpected event)
//
- (DFSMState *) feed :(int) event;

//
// Try an event in the state machine; the state of the state machine is 
// NOT changed; only the state that will be reached if the event is fed, is
// returned
//
// @param event    the event for the state machine
//
// @return the state if the event is fed (or nil for unexpected event)
//
- (DFSMState *) try :(int) event;

@end
#endif
