#ifndef _DFTPCLIENT_H
#define _DFTPCLIENT_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DFTPClient - the FTP client class in the ofc-library
//
//               Copyright (C) 2006  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-16 12:38:03 $ $Revision: 1.13 $
//==============================================================================

#include "ofc/config.h"

#include "ofc/DSocketAddress.h"
#include "ofc/DSocket.h"
#include "ofc/DList.h"
#include "ofc/DTextReadable.h"
#include "ofc/DTextWritable.h"
#include "ofc/DDataReadable.h"
#include "ofc/DDataWritable.h"


#define DFPC_PORT       (21)

#define DFPC_PASSIVE    (0)
#define DFPC_ACTIVE     (1)

//
// The DFTPClient class implements a number of methods for building a FTP (File
// Transfer Protocol) client.
//
// @example
// #include <stdio.h>
// #include "ofc/DFTPClient.h"
// 
// int main(int argc, char *argv[])
// {
//   DFTPClient         *ftp   = [DFTPClient new];
//   DInetSocketAddress *addr  = [DInetSocketAddress new];
//   DList              *files = [DList new];
//   DListIterator      *iter  = [DListIterator new];
//   DText              *str;
// 
//   [addr host :"ftp.gnu.org" :DFPC_PORT]; // Set address to the host and default FTP port
// 
//   if ([ftp open :addr])                // Open the ftp client (passive)
//   {
//     if ([ftp waitForGreetings])        // Wait for server greetings
//     {
//       if ([ftp login :NULL :NULL :NULL]) // Login
//       {
//         if (![ftp getWorkingDirectory])
//           printf("Could not do a get working directory:%d\n", [ftp responseCode]);
// 
//         if ([ftp list :NULL :files]) // Ask for a directory listing
//         {
//           printf("Directory with %ld files.\n", [files length]);
// 
//           [iter list :files];
// 
//           str = [iter first];
//           while (str != nil)
//           {
//             printf("File: %s\n", [str cstring]); // Print the files in the listing
// 
//             str = [iter next];
//           }
//         }
//         else
//           printf("Could not do a directory listing:%d\n", [ftp responseCode]);
//       
//       }
//       else
//         printf("Could not login:%d\n", [ftp responseCode]);
//     }
//     else
//       printf("Service not available:%d\n", [ftp responseCode]);
// 
//     [ftp quit];                       // Quit the connection
//   }
//   else
//     printf("Could not connect to \"ftp.gnu.org\".\n");
//   
//   return 0;
// }
// 

@interface DFTPClient : Object
{
@private
  DSocket                *_client;       // the client socket
  DText                  *_request;      // the request for the server
  DText                  *_response;     // the response from the server
  DText                  *_directory;    // the directory info received from the server
  id <DSocketAddressable> _server;       // the address of the server
  id <DSocketAddressable> _address;      // the address received from the server
  int                     _bytes;        // the number of bytes received from the server
  BOOL                    _connected;    // is there a connection to the server ?
  int                     _port;         // the active port (or 0 for passive mode)
  int                     _code;         // the response code
  int                     _bufferSize;   // the buffer size during binary transfer (def. 8192)
}


//// Constructors

//
// Initialise the FTP Client
//
// @return the object
//
- (DFTPClient *) init;


//// Deconstructor

//
// Free the FTP client object
//
// @return the object
//
- free;


//// Copy related methods

//
// Deepen the object (not implemented)
//
//
- deepen;


//// Member methods

//
// Return the local socket
// 
// @return a reference to the socket object
//
- (DSocket *) socket;

//
// Check if there is a connection to the server
// 
// @return is there a connection ?
// 
- (BOOL) isConnected;

//
// Return the mode of the ftp client
// 
// @return the mode (DFPC_PASSIVE or DFPC_ACTIVE)
// 
- (BOOL) mode;

//
// Set the ftp client in passive mode
// 
// @return the object
// 
- (DFTPClient *) passive;

//
// Set the ftp client in active mode
// 
// @param port     the port to be used during transfer of data with the server
// 
// @return the object
// 
- (DFTPClient *) active :(int) port;

//
// Get the last response code
// 
// @return the last response code (1xx-5xx, -1)
// 
- (int) responseCode;

//
// Get the directory name from the last response with a directory name
// 
// @return the directory name (or EOS)
// 
- (const char *) directory;

//
// Get the number of bytes received from the server (response 150)
// 
// @return the number of bytes (or -1)
// 
- (int) bytes;

//
// Set the buffer size during binary transfer
// 
// @param size     the buffer size 
// 
// @return the object
// 
- (DFTPClient *) bufferSize :(int) size;

//
// Get the buffer size during binary transfer
// 
// @return the size
// 
- (int) bufferSize;


//// Connection methods

//
// Open the FTP client
// 
// @param server        the address of the server
// 
// @return success
// 
- (BOOL) open :(id <DSocketAddressable>) server;

//
// Close the connection to the ftp server
// 
// @return success
// 
- (BOOL) close;


//// Sending and receiving messages

//
// Send a command to the server
// 
// @param command       the command to be sent
// @param parameters    the parameters of the command (or NULL)
// 
// @return success
// 
- (BOOL) sendCommand :(const char *) command :(const char *) parameters;

//
// Receive a response from the server
// 
// @return the first digit of the last response code (or -1)
// 
- (int) receiveResponse;


//// FTP methods

//
// Wait for server greetings
//
// @return success
//
- (BOOL) waitForGreetings;

//
// Do a login on the ftp server
// 
// @param user     the user name (or NULL)
// @param password the password for the user (or NULL)
// @param account  the account of the user (or NULL)
// 
// @return success
// 
- (BOOL) login :(const char *) user :(const char *) password :(const char *) account;

//
// Use ASCII during data transfer
// 
// @return success
// 
- (BOOL) transferASCII;

//
// Use binary during data transfer
// 
// @return success
// 
- (BOOL) transferBinary;

//
// Send the PORT/EPTR commando to the ftp server
// 
// @return a reference to the socket address received from the server (or nil)
// 
- (id <DSocketAddressable>) sendPort;

//
// Change the current directory
// 
// @param directory     the directory
// 
// @return success
// 
- (BOOL) changeCurrentDirectory :(const char *) directory;

//
// Change the current directory to the parent directory
// 
// @return success
// 
- (BOOL) changeCurrentToParent;

//
// Mount a different file system
// 
// @param directory     the directory of the file system
// 
// @return success
// 
- (BOOL) mount :(const char *) directory;

//
// Reinitialize the connection (note: continu with login)
// 
// @return success
// 
- (BOOL) reinitialize;

//
// Store ASCII on the ftp server. This method supports two ways of data 
// feeding. If the file parameter is not nil, the ASCII data is read from
// the TextReadable; if it is nil, the data is read by calling the method 
// getASCIIData, which should be overridden.
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (if nil then use getAsciiData)
//
// @return success
// 
- (BOOL) storeASCII :(const char *) command :(const char *) filename :(id <DTextReadable>) file;

//
// Store ASCII on the ftp server, replacing an already existing file
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (or nil)
//
// @return success
// 
- (BOOL) overwriteASCII :(const char *) filename :(id <DTextReadable>) file;

//
// Store ASCII on the ftp server, appending an already existing file
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (or nil)
//
// @return success
// 
- (BOOL) appendASCII :(const char *) filename :(id <DTextReadable>) file;

//
// Store ASCII file with a filename that is unique on the server
// 
// @param file          the (opened) file (or nil)
//
// @return success
// 
- (BOOL) storeASCIIUnique :(id <DTextReadable>) file;

//
// Store Binary data on the ftp server. This method supports two ways of data 
// feeding. If the file parameter is not nil, the binary data is read from
// the DataReadable; if it is nil, the data is read by calling the method 
// getBinaryData, which should be overridden.
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (if nil then use getBinaryData)
//
// @return success
// 
- (BOOL) storeBinary :(const char *) command :(const char *) filename :(id <DDataReadable>) file;

//
// Store a binary file on the ftp server, replacing an already existing file
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (or nil)
//
// @return success
// 
- (BOOL) overwriteBinary :(const char *) filename :(id <DDataReadable>) file;

//
// Store a binary file on the ftp server, appending an already existing file
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (or nil)
//
// @return success
// 
- (BOOL) appendBinary :(const char *) filename :(id <DDataReadable>) file;

//
// Store a binary file with a filename that is unique on the server
// 
// @param file          the (opened) file (or nil)
//
// @return success
// 
- (BOOL) storeBinaryUnique :(id <DDataReadable>) file;

//
// Retrieve ASCII from the ftp server. This method supports two ways of data 
// receiving. If the file parameter is not nil, the ASCII data is written to
// the TextWritable; if the file parameter is nil, the data is written by 
// calling the method setASCIIData, which should be overridden.
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (if nil use setAsciiData)
//
// @return success
// 
- (BOOL) retrieveASCII :(const char *) filename :(id <DTextWritable>) file;

//
// Retrieve binary data from the ftp server. This method supports two ways of 
// data receiving. If the file parameter is not nil, the binary data is written
//  to the DataWritable; if the file parameter is nil, the data is written by 
// calling the method setBinaryData, which should be overridden.
// 
// @param filename      the (path and) name of the file on the server
// @param file          the (opened) file (if nil use setAsciiData)
//
// @return success
// 
- (BOOL) retrieveBinary :(const char *) filename :(id <DDataWritable>) file;

//
// Send the rename command
// 
// @param oldName       the old filename
// @param newName       the new filename
//
// @return success
// 
- (BOOL) rename :(const char *) oldName :(const char *) newName;

//
// Send the delete file to the ftp server
// 
// @param filename      the name of the file
// 
// @return success
// 
- (BOOL) delete :(const char *) filename;

//
// Send the remove directory command to the ftp server
// 
// @param directory     the directory name
// 
// @return success
// 
- (BOOL) removeDirectory :(const char *) directory;

//
// Send the make directory command to the ftp server.
// Use 'directory' method for the name of the directory in the response.
// 
// @param directory     the directory name
// 
// @return success
// 
- (BOOL) makeDirectory :(const char *) directory;

//
// Send the request for the current working directory to the ftp server.
// Use 'directory' method for the directory name in the response.
// 
// @return success
// 
- (BOOL) getWorkingDirectory;

//
// Send a list commando (LIST or NLST) (private)
// 
// @param command       the command to be sent
// @param parameter     the parameter (or NULL)
// @param list          the list to store the information (or nil)
// 
// @return success
// 
- (BOOL) sendListCommand :(const char *) command :(const char *) parameter :(DList *) list;

//
// Send the request for a list to the ftp server. If the list is not
// nil, the received names will be stored in the list. If it is nil, the
// method processListLine will be called for every received list line.
// 
// @param name          the directory or filename (or NULL)
// @param list          the list to store the received info (or nil)
// 
// @return success
// 
- (BOOL) list :(const char *) name :(DList *) list;

//
// Send the request for site parameters to the ftp server
// 
// @param parameter     the parameter
// 
// @return success
// 
- (BOOL) siteParameter :(const char *) parameter;

//
// Send the request for system information to the ftp server
// 
// @return success
// 
- (BOOL) systemInfo;

//
// Send the request for status to the ftp server
// 
// @return success
// 
- (BOOL) state;

//
// Send the request for help to the ftp server
// 
// @param key           the key for the help (or NULL)
// 
// @return success
// 
- (BOOL) help :(const char *) key;

//
// Send the request for noop to the ftp server
// 
// @return success
// 
- (BOOL) noop;

//
// Send the request for a name list to the ftp server. If the list is not
// nil, the received names will be stored in the list. If it is nil, the
// method processListLine will be called for every received list line.
// 
// @param directory     the directory name (or NULL)
// @param list          the list to store the received names (or nil)
// 
// @return success
// 
- (BOOL) nameList :(const char *) directory :(DList *) list;

//
// Send the QUIT command to the ftp server (and close the connnection)
// 
// @return success
// 
- (BOOL) quit;


//// Data feeders

//
// Get ASCII data during an ASCII transfer (to be overridden)
// 
// @return a (new) text string with ASCII data to be stored (without CR and/or LF) (or nil for done)
// 
- (DText *) getASCIIData;

//
// Get binary data during an binary transfer (to be overridden)
// 
// @param bufferSize     the size of the buffer
// 
// @return a (new) data string with data to be stored (or nil for done)
// 
- (DData *) getBinaryData :(int) bufferSize;


//// Data savers

//
// Set ASCII data during an ASCII retrieve (to be overridden)
// 
// @param cstring        the string with the ASCII data (no end of line)
// 
- (void) setASCIIData :(const char *) cstring;

//
// Set binary data during an binary retrieve (to be overridden)
// 
// @param data           the data
// @param length         the length of the data
// 
- (void) setBinaryData :(const unsigned char *) data :(unsigned long) length;


//// Process responses

//
// Process the receiving of a response (to be overridden)
// 
// @param code          the response code 1xx - 5xx
// @param response      the response text
// 
// @return none
// 
- (void) processResponse :(int) code :(const char *) response;

//
// Process the receiving of list information (to be overridden)
// 
// @param line          one line of the list information
// 
// @return none
// 
- (void) processListLine :(const char *) line;

//
// Process the 257 response (the directory name)
// 
// @param response      the response text
// 
// @return none
// 
- (void) processResponse257 :(DText *) response;

//
// Process the 227 response (the inet address)
// 
// @param response      the response text
// 
// @return none
// 
- (void) processResponse227 :(DText *) response;

//
// Process the 150 response (the byte size)
// 
// @param response      the response text
// 
// @return none
// 
- (void) processResponse150 :(DText *) response;

//
// Process the 229 response (the inet address for IPv6)
// 
// @param response      the response text
// 
// @return none
// 
- (void) processResponse229 :(DText *) response;

@end
#endif
