#ifndef _DFRACTION_H
#define _DFRACTION_H

// This file was automatically generated by objch.

//==============================================================================
//
//         DFraction - the fraction number class in the ofc-library
//
//               Copyright (C) 2003-2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-25 16:35:58 $ $Revision: 1.6 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DParsable.h"
#include "ofc/DSimpleArithmetic.h"
#include "ofc/DComparable.h"

#include "ofc/DText.h"

//
// The fraction class implements a number of methods for working with fractions.
//
// @example: 
// #include <stdio.h>
// #include "ofc/DFraction.h"
// 
// int main(int argc, char *argv[])
// {
//   DFraction *fr1  = [DFraction alloc];
//   DFraction *fr2  = [DFraction new  ];
//   DFraction *fr3  = [DFraction new  ];
//   DText     *str;
// 
//   [fr1 init :3 :4];                    // Init with fraction 3/4
// 
//   str = [fr1 toText];
//   printf("Fraction1 %0d/%0d or \"%s\".\n", [fr1 numerator], [fr1 denominator], [str cstring]);
//   [str free];
// 
//   [fr2 set :4 :8];                     // Set to 1/2
// 
//   str = [fr2 toText];
//   printf("Fraction2 %0d/%0d or \"%s\".\n", [fr2 numerator], [fr2 denominator], [str cstring]);
//   [str free];
// 
//   [fr2 mul :fr1];                      // Fraction2 = Fraction2 * Fraction1
// 
//   [fr3 sub :fr2 :fr1];                 // Fraction3 = Fraction2 - Fraction1
// 
//   str = [fr3 toText];
//   printf("Fraction3 as double %f or as string \"%s\".\n", [fr3 toDouble], [str cstring]);
//   [str free];
// 
//   if ([fr1 compare :fr3] == 0)
//     printf("Fraction1 is equal to fraction3.\n");
//   else if ([fr1 compare :fr3] < 0)
//     printf("Fraction1 is smaller than fraction3.\n");
//   else
//     printf("Fraction1 is greater than fraction3.\n");
// 
//   [fr1 free];                          // Cleanup
//   [fr2 free];
//   [fr3 free];
// 
//   return 0;
// }
// 


@interface DFraction : Object <DTextable,DSimpleArithmetic,DComparable,DParsable>
{
@private
  int          _num;   // the numerator
  int          _denom; // the denominator
}


//// Constructors

//
// Initialise a zero fraction
//
// @return the object
//
- (DFraction *) init;

//
// Initialise a fraction
//
// @param num      the numerator
// @param denom    the denominator
//
// @return the object
//
- (DFraction *) init :(int) num :(int) denom;


//// Member methods

//
// Get the numerator
//
// @return the numerator
//
- (int) numerator;

//
// Set the numerator
//
// @param num      the numerator of the fraction
//
// @return the object
//
- (DFraction *) numerator :(int) num;

//
// Get the denominator
//
// @return the denominator
//
- (int) denominator;

//
// Set the denominator of the fraction
//
// @param denom    the denominator
//
// @return the object
//
- (DFraction *) denominator :(int) denom;

//
// Set the numerator and denominator of the fraction
//
// @param num      the numerator
// @param denom    the denominator
//
// @return the object
//
- (DFraction *) set :(int) num :(int) denom;

//
//
// Move an other object in this object
//
// @param other    the other object
//
// @return this object
//
- (DFraction *) move :(DFraction *) other;


//// Methods on fractions

//
// Add with an other fraction
//
// @param other    the fraction
//
// @return the object
//
- (DFraction *) add :(DFraction *) other;

//
// Add two fractions and store the result in the object
//
// @param fr1     the first fraction
// @param fr2     the second fraction
//
// @return the object
//
- (DFraction *) add :(DFraction *) fr1 :(DFraction *) fr2;

//
// Subract with an other fraction
//
// @param other     the fraction
//
// @return the object
//
- (DFraction *) sub :(DFraction *) other;

//
// Subtract two fractions and store the result in the object
//
// @param fr1     the first fraction
// @param fr2     the second fraction
//
// @return the object
//
- (DFraction *) sub :(DFraction *) fr1 :(DFraction *) fr2;

//
// Multiply with an other fraction
//
// @param other    the other fraction
//
// @return the object
//
- (DFraction *) mul :(DFraction *) other;

//
// Multiply two fractions and store the result in the object
//
// @param fr1     the first fraction
// @param fr2     the second fraction
//
// @return the object
//
- (DFraction *) mul :(DFraction *) fr1 :(DFraction *) fr2;

//
// Divide with an other fraction
//
// @param other    the divider
//
// @return the object
//
- (DFraction *) div :(DFraction *) other;

//
// Divide two fractions and store the result in the object
//
// @param fraction the fraction
// @param divider  the divider
//
// @return the object
//
- (DFraction *) div :(DFraction *) fraction :(DFraction *) divider;

//
// Normalize the fraction
//
// @return the object
//
- (DFraction *) norm;

//
// Convert to a double
//
// @return the double
//
- (double) toDouble;

//
// Invert the fraction
//
// @return the object
//
- (DFraction *) invert;


//// Class methods

//
// Calculate the Greatest Common Divider
// 
// @param a        the first number
// @param b        the second number
// 
// @return the greatest common divider
// 
+ (int) gcd :(int) a :(int) b;

//
// Calculate the Least Common Multiplier
// 
// @param a        the first number
// @param b        the second number
// 
// @return the least common multiplier
// 
+ (int) lcm :(int) a :(int) b;


//// Textable protocol methods

//
// Convert to a text string 
//
// @return a (new) text string with the complex number
//
- (DText *) toText;


//// Comparable protocol method

//
// Compare with another fraction (using doubles)
// 
// @param other   the other fraction
// 
// @return the compare result (-1,0,1)
// 
- (int) compare :(DFraction *) other;


//// DParsable protocol implementation

//
// Parse a string for a big fraction. Format: [space]*[+-]number/number
// and number is one of:
// decimal:     [1..9][0..9]*
// hexadecimal: 0x[0..9a..f]*
// octal:       0[0..7]*
//
// @param cstr     the string to be parsed (moved to the first non-parsable char)
// 
// @return the result (0,ENODATA)
//
- (int) fromString :(char **) cstr;

@end
#endif
