//==============================================================================
//
//            DGraph - the graph collection in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-11 16:59:46 $ $Revision: 1.3 $
//
//==============================================================================

#include "ofc/DGraph.h"

#include <float.h>


#if _INTERFACE_

#include "ofc/config.h"

#include "ofc/DList.h"

@class DGraphEdge;  // forward definition

//
// The graph node object implements a node in a graph. This node can store
// a different object so that it can be used as a collection. It can also
// store a label and attributes for graphviz visualation.
//

@interface DGraphNode : Object
{
@private
  DList         *_ingoing;    // the ingoing edges
  DList         *_outgoing;   // the outgoing edges
  DText         *_name;       // the name (set by DGraph)
  DText         *_label;      // the (graphviz) label
  DText         *_attributes; // the (graphviz) attributes
  id             _object;     // the stored object
  DGraphNode    *_prev;       // the previous node (during shortest path)
  double         _sum;        // the weight sum (during shortest path)
}


#endif



@implementation DGraphNode


//// Constructors

//
// Initialise an empty graph node
//
// @return the object
//
- (DGraphNode *) init
{
  [super init];
  
  _ingoing    = [DList new];
  _outgoing   = [DList new];

  _name       = [DText new];

  _label      = nil;
  _attributes = nil;
  
  _object     = nil;
  
  _prev       = nil;
  _sum        = DBL_MAX;
  
  return self;
}

//
// Initialise a graph node with a label and object
//
// @param label    the label for the node
// @param object   the object
//
// @return the object
//
- (DGraphNode *) init :(const char *) label :(id) object
{
  [self init];
  
  [self label  :label ];
  [self object :object];
  
  return self;
}

//
// Initialise a graph node with a label, attributes and object
//
// @param label        the label for the node
// @param attributes   the (graphviz) attributes for the node
// @param object       the object
//
// @return the object
//
- (DGraphNode *) init :(const char *) label :(const char *) attributes :(id) object
{
  [self init];
  
  [self label      :label ];
  [self attributes :attributes];
  [self object     :object];
  
  return self;
}


//// Copy related methods

//
// Do a shallow copy of the graph node (not implemented)
//
// @return the object
//
- shallowCopy
{
  DGraphNode *copy = [super shallowCopy];

  copy->_ingoing   = [_ingoing shallowCopy]; // references to edges
  copy->_outgoing  = [_outgoing shallowCopy]; // references to edges

  copy->_name      = [_name copy];
  
  if (_label != nil)
  {
    copy->_label   = [_label copy];
  }

  if (_attributes != nil)
  {
    copy->_attributes = [_attributes copy];
  }
  
  copy->_prev = nil;
  
  return copy;
}

//
// Deepen the copied object
//
// @return the object
//
- deepen
{
  [super deepen];
  
  _object = [_object copy];
  
  return self;
}


//// Destructor

//
// (Shallow) Free the graph node (the stored object is NOT freed)
//
// @return the object
//
- shallowFree
{
  [_ingoing  shallowFree]; _ingoing  = nil;
  [_outgoing shallowFree]; _outgoing = nil;
  

  [_name free]; _name = nil;
  
  if (_label != nil)
  {
    [_label free]; _label = nil;
  }
  
  if (_attributes != nil)
  {
    [_attributes free]; _attributes = nil;
  }
  
  _object     = nil;
  _prev       = nil;  
  
  return [super free];
}

//
// Free the graph node (the stored object is freed)
//
// @return the object
//
- free
{
  if (_object != nil)
  {
    [_object free]; _object = nil;
  }
  
  return [self shallowFree];
}


//// Member methods

//
// Return the name of the graph node (used by DGraph)
// 
// @return the name of the node
// 
- (const char *) name
{
  return [_name cstring];
}

//
// Set the name of the graph node (used by DGraph)
// 
// @param name    the name
// 
// @return the node object
// 
- (DGraphNode *) name :(const char *) name
{
  if ((name == NULL) || (*name == EOS))
  {
    WARNING(DW_INVALID_ARG, "name");
  }
  else
  {
    [_name set :name];
  }
  
  return self;
}

//
// Return the label of the graph node
//
// @return the label (or NULL)
//
- (const char *) label
{
  return (_label != nil) ? [_label cstring] : NULL;
}

//
// Set the label of the graph node
//
// @param label   the label (or NULL)
// 
// Return the node object
// 
- (DGraphNode *) label :(const char *) label
{
  if ((label == NULL) || (*label == EOS))
  {
    if (_label != nil)
    {
      [_label free]; _label = nil;
    }
  }
  else
  {
    if (_label == nil)
    {
      _label = [DText new];
    }
    
    [_label set :label];
  }
  
  return self;
}

//
// Return the (graphviz) attributes of the node
// 
// @return the attributes (or NULL)
// 
- (const char *) attributes 
{
  return (_attributes != nil) ? [_attributes cstring] : NULL;
}

//
// Set the (graphviz) attributes for the node
// 
// @param attributes    the attributes (or NULL)
// 
// @return the node object
// 
- (DGraphNode *) attributes :(const char *) attributes
{
  if ((attributes == NULL) || (*attributes == EOS))
  {
    if (_attributes != nil)
    {
      [_attributes free]; _attributes = nil;
    }
  }
  else
  {
    if (_attributes == nil)
    {
      _attributes = [DText new];
    }
    
    [_attributes set :attributes];
  }
  
  return self;
}

//
// Return the stored object in the node
// 
// @return the (stored) object (or nil)
// 
- (id) object
{
  return _object;
}

//
// Set the object for the node
// 
// @param object    the object (or nil)
// 
// @return the previous stored object
// 
- (id) object :(id) object
{
  id tmp = _object;
  
  _object = object;
  
  return tmp;
}


//// Path methods (used during shortest path determination)

//
// Reset the path
// 
// @return the node object
// 
- (DGraphNode *) path
{
  _prev = nil;
  _sum  = DBL_MAX;
  
  return self;
}

//
// Set the path 
// 
// @param prev    the previous node in the path
// @param sum     the sum of weights in the path
// 
// @return the node object
// 
- (DGraphNode *) path :(DGraphNode *) prev :(double) sum
{
  _prev = prev;
  _sum  = sum;
  
  return self;
}

//
// Return the previous node in the path
// 
// @return the previous node (or nil)
// 
- (DGraphNode *) prev
{
  return _prev;
}

//
// Return the sum of weights in the path
// 
// @return the sum
// 
- (double) sum
{
  return _sum;
}


//// Edge methods

//
// Return the ingoing degree of the node
// 
// @return the number of ingoing edges
// 
- (unsigned long) ingoingDegree
{
  return [_ingoing length];
}

//
// Return the outgoing degree of the node
// 
// @return the number of output edges
// 
- (unsigned long) outgoingDegree
{
  return [_outgoing length];
}

//
// Return the degree of the node
// 
// @return the number of ingoing and outgoing edges
// 
- (unsigned long) degree
{
  return [_ingoing length] + [_outgoing length];
}

//
// Add an ingoing edge
// 
// @param edge    the edge
// 
// @return the node object
// 
- (DGraphNode *) addIngoingEdge :(DGraphEdge *) edge
{
  if (edge == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "edge");
  }
  else if ([_ingoing has :edge])
  {
    WARNING(DW_UNKNOWN_WARNING, "edge already ingoing");
  }
  else
  {
    [_ingoing append :edge];
  }
  
  return self;
}

//
// Remove an ingoing edge
// 
// @param edge    the edge
// 
// @return success
// 
- (BOOL) removeIngoingEdge :(DGraphEdge *) edge
{
  BOOL ok = NO;
  
  if (edge == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "edge");
  }
  else
  {
    ok = [_ingoing remove :edge];
  }

  return ok;
}

//
// Add an outgoing edge
// 
// @param edge    the edge
// 
// @return the node object
// 
- (DGraphNode *) addOutgoingEdge :(DGraphEdge *) edge
{
  if ([_outgoing has :edge])
  {
    WARNING(DW_UNKNOWN_WARNING, "edge already outgoing");
  }
  else
  {
    [_outgoing append :edge];
  }
  
  return self;
}

//
// Remove an outgoing edge
// 
// @param edge    the edge
// 
// @return success
// 
- (BOOL) removeOutgoingEdge :(DGraphEdge *) edge
{
  BOOL ok = NO;
  
  if (edge == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "edge");
  }
  else
  {
    ok = [_outgoing remove :edge];
  }

  return ok;
}

//
// Replace edge with the copied edge (used during coping of graph)
// 
// @param orig     the original edge
// @param copy     the copied edge
// 
// @return the object
// 
- (DGraphNode *) replace :(DGraphEdge *) orig :(DGraphEdge *) copy
{
  [_ingoing  replace :orig :copy :-1];
  [_outgoing replace :orig :copy :-1];
  
  return self;
}

//// Edge iterators

//
// Return the ingoing edge iterator
// 
// @return a (new) iterator on the ingoing edges
//
- (DListIterator *) ingoingEdges
{
  DListIterator *iter = [DListIterator alloc];
  
  [iter init :_ingoing];
  
  return iter;
}

//
// Return the outgoing edge iterator
// 
// @return a (new) list iterator on the outgoing edges
// 
- (DListIterator *) outgoingEdges
{
  DListIterator *iter = [DListIterator alloc];
  
  [iter init :_outgoing];
  
  return iter;
}

@end


  
  
#if _INTERFACE_

//
// The graph edge object implements an edge in a graph. This edge can store
// another object, so that it can be used as a collection. It can also
// store a label and attributes for graphviz visualation.
//

@interface DGraphEdge : Object
{
@private
  DGraphNode    *_source;     // the source node
  DGraphNode    *_target;     // the target node
  DText         *_name;       // the name (set by DGraph)
  DText         *_label;      // the (graphviz) label 
  double         _weight;     // the weight of the edge (> 0.0)
  DText         *_attributes; // the (graphviz) attributes
  id             _object;     // the stored object
}
#endif


@implementation DGraphEdge


//// Constructors

//
// Initialise an empty graph edge
//
// @return the object
//
- (DGraphEdge *) init
{
  [super init];

  _name       = [DText new];
  
  _label      = nil;
  _attributes = nil;

  _weight     = 1.0;
  
  _source     = nil;
  _target     = nil;
  
  _object     = nil;
  
  return self;
}

//
// Initialise a graph edge with a label, weight and object
//
// @param label    the label
// @param weight   the weight (> 0.0)
// @param object   the object
//
// @return the object
//
- (DGraphEdge *) init :(const char *) label :(double) weight :(id) object
{
  [self init];
  
  [self label  :label ];
  [self weight :weight];
  [self object :object];
  
  return self;
}

//
// Initialise a graph edge with a label, weight, attributes and object
//
// @param label        the label 
// @param attributes   the (graphviz) attributes
// @param weight       the weight (> 0.0)
// @param object       the object
//
// @return the object
//
- (DGraphEdge *) init :(const char *) label :(const char *) attributes :(double) weight :(id) object
{
  [self init];
  
  [self label      :label ];
  [self weight     :weight];
  [self attributes :attributes];
  [self object     :object];
  
  return self;
}


//// Copy related methods

//
// Do a shallow copy of the graph edge
//
// @return the object
//
- shallowCopy
{
  DGraphEdge *copy = [super shallowCopy];
  
  copy->_name = [_name copy];
  
  if (_label != nil)
  {
    copy->_label = [_label copy];
  }
  
  if (_attributes != nil)
  {
    copy->_attributes = [_attributes copy];
  }

  return copy;
}

//
// Deepen a copied edge
//
// @return the object
//
- deepen
{
  [super deepen];
  
  _object = [_object copy];
  
  return self;
}


//// Destructor


//
// Free the graph edge (the stored object is NOT freed)
//
// @return the object
//
- shallowFree
{
  _source = nil;
  _target = nil;

  [_name free]; _name = nil;
  
  if (_label != nil)
  {
    [_label free]; _label = nil;
  }
 
  if (_attributes != nil)
  {
    [_attributes free]; _attributes = nil;
  }
  
  _object     = nil;
  
  return [super free];
}

//
// Free the graph edge (the stored object is freed)
//
// @return the object
//
- free
{
  if (_object != nil)
  {
    [_object free]; _object = nil;
  }
  
  return [self shallowFree];
}


//// Member methods

//
// Return the name of the edge (used by DGraph)
// 
// @return the name of the node
// 
- (const char *) name
{
  return [_name cstring];
}

//
// Set the name of the edge (used by DGraph)
// 
// @param name    the name
// 
// @return the edge object
// 
- (DGraphEdge *) name :(const char *) name
{
  if ((name == NULL) || (*name == EOS))
  {
    WARNING(DW_INVALID_ARG, "name");
  }
  else
  {
    [_name set :name];
  }
    
  return self;
}

//
// Return the (graphviz) label of the graph edge
//
// @return the label (or NULL)
//
- (const char *) label
{
  return (_label != nil) ? [_label cstring] : NULL;
}

//
// Set the (graphviz) label of the graph edge
//
// @param label   the label
// 
// Return the edge object
// 
- (DGraphEdge *) label :(const char *) label
{
  if ((label == NULL) || (*label == EOS))
  {
    if (_label != nil)
    {
      [_label free]; _label = nil;
    }
  }
  else
  {
    if (_label == nil)
    {
      _label = [DText new];
    }
    
    [_label set :label];
  }
  
  return self;
}

//
// Return the weight of the graph edge
//
// @return the weight
//
- (double) weight
{
  return _weight;
}

//
// Set the weight of the graph edge
//
// @param weight  the weight (> 0.0)
// 
// Return the edge object
// 
- (DGraphEdge *) weight :(double) weight
{
  if (weight <= 0.0)
  {
    WARNING(DW_INVALID_ARG, "double");
  }
  else
  {
    _weight = weight;
  }
  
  return self;
}

//
// Return the (graphviz) attributes of the edge
// 
// @return the attributes (or NULL)
// 
- (const char *) attributes 
{
  return (_attributes != nil) ? [_attributes cstring] : NULL;
}

//
// Set the (graphviz) attributes for the edge
// 
// @param attributes    the attributes
// 
// @return the edge object
// 
- (DGraphEdge *) attributes :(const char *) attributes
{
  if ((attributes == NULL) || (*attributes == EOS))
  {
    if (_attributes != nil)
    {
      [_attributes free]; _attributes = nil;
    }
  }
  else
  {
    if (_attributes == nil)
    {
      _attributes = [DText new];
    }
    
    [_attributes set :attributes];
  }
  
  return self;
}

//
// Return the stored object in the edge
// 
// @return the (stored) object (or nil)
// 
- (id) object
{
  return _object;
}

//
// Set the object for the edge
// 
// @param object    the object
// 
// @return the previous stored object
// 
- (id) object :(id) object
{
  id tmp = _object;
  
  _object = object;
  
  return tmp;
}


//// main methods

//
// Connect the source and target nodes by the edge
// 
// @param source    the source node (!= nil)
// @param target    the target node (!= nil)
// 
// @return success
// 
- (BOOL) connect :(DGraphNode *) source :(DGraphNode *) target
{
  BOOL ok = NO;
  
  if ((source == nil) || (target == nil))
  {
    WARNING(DW_NIL_NOT_ALLOWED, "source/target");
  }
  else
  {
    [self disconnect];
    
    
    _source = source;
    
    [_source addOutgoingEdge :self];


    _target = target;
    
    [_target addIngoingEdge :self];
    
    
    ok = YES;
  }
  
  return ok;
}

//
// Disconnect the edge from the source and target node
// 
// @return the edge object
// 
- (DGraphEdge *) disconnect 
{
  if (_source != nil)
  {
    [_source removeOutgoingEdge :self]; _source = nil;
  }
  
  if (_target != nil)
  {
    [_target removeIngoingEdge :self]; _target = nil;
  }

  return self;
}

//
// Return the source node
// 
// @return the source node
// 
- (DGraphNode *) source
{
  return _source;
}

//
// Return the target node
// 
// @return the target node (or nil)
// 
- (DGraphNode *) target
{
  return _target;
}

//
// Reverse the edge
// 
// @return the edge object
// 
- (DGraphEdge *) reverse
{
  DGraphNode *target = _source;
  DGraphNode *source = _target;

  [self connect :source :target];
  
  return self;
}

//
// Replace node with the copied node (used during coping of graph)
// 
// @param orig     the original node
// @param copy     the copied node
// 
// @return the object
// 
- (DGraphEdge *) replace :(DGraphNode *) orig :(DGraphNode *) copy
{
  if (_source == orig)
  {
    _source = copy;
  }
  if (_target == orig)
  {
    _target = copy;
  }
  
  return self;
}

@end
  


#if _INTERFACE_

  
#include "ofc/DTextWritable.h"
  
//
// The graph collection stores objects in directed graph with nodes and edges. 
// Both the nodes as the edges can store objects. The graph collection can also 
// be used as graph data structure. Using the toDot method, the graph structure 
// can be exported to a dot-file which can be visualized with the software of 
// @link(www.graphviz.org). IMPORTANT: as long as a node or edge object are part 
// of a graph, the node or edge object is owned by the graph object. Do not free 
// them outside the scope of the graph object.
// 
// @example
// #include <stdio.h>
// #include "ofc/DGraph.h"
// #include "ofc/DInt.h"
// #include "ofc/DFile.h"
// 
// int main(int argc, char *argv[])
// {
//   DGraph     *graph = [DGraph new];
//   DFile      *file  = [DFile  new];
// 
//   DList      *path;
// 
//   DGraphNode *node1,*node2,*node3;
//   DGraphEdge *edge1,*edge2,*edge3;
// 
//   DInt       *nr;
//   double      weight;
// 
//   [graph attributes :"rotate=90"];     // Set the graphviz attributes for the graph itself
// 
//   printf("Add the nodes to the graph.\n");
// 
//   nr = [DInt new]; [nr set :7];        // Create and add node to graph with label:start, attributes and object:7
//   node1 = [graph addNode :"start" :"shape=diamond" :nr];
// 
//   nr = [DInt new]; [nr set :3];        // Create and add node to graph with label:between, attributes and object:3
//   node2 = [graph addNode :"between" :"color=green" :nr];
// 
//   nr = [DInt new]; [nr set :9];        // Create and add node to graph with label:start, attributes and object:7
//   node3 = [graph addNode :"end" :"shape=circle" :nr];
// 
//   printf("Add the edges to the graph.\n");
// 
//   nr = [DInt new]; [nr set :-4];       // Create and add edge to graph from node1 to node2
//   edge1 = [graph addEdge :"edge1" :"color=blue" :1.0 :nr :node1 :node2];
//   
//   nr = [DInt new]; [nr set :-5];       // Create and add edge to graph from node2 to node3
//   edge2 = [graph addEdge :"edge2" :"style=dashed" :2.0 :nr :node2 :node3];
//   
//   nr = [DInt new]; [nr set :-3];       // Create and add edge to graph from node1 to node3
//   edge3 = [graph addEdge :"edge3" :"arrowhead=crow" :3.5 :nr :node1 :node3];
//  
//   path = [graph shortestPath :&weight :node1 :node3];
//   printf("Number of nodes in shortestpath:%ld with weight:%.1f\n", [path length], weight);
//   [path shallowFree];
// 
//   printf("Ingoing degree of node3:%lu\n",  [node3 ingoingDegree]);
//   printf("Outgoing degree of node2:%lu\n", [node2 outgoingDegree]);
//  
//   nr = [node2 object];
//   printf("Object of node2:%d\n", [nr get]);
// 
//   if ([file open :"example.dot" :"w"])
//   {
//     if ([graph toDot :file])
//       printf("Graph is succesfully exported to \"example.dot\".\n");
//     else
//       printf("Graph could not be exported to \"example.dot\".\n");
// 
//     [file close];
//   }
//   else
//     printf("Could not open the file \"example.dot\".\n");
// 
//   [graph free];                        // Cleanup
//   [file  free];
// 
//   return 0;
// }
//

@interface DGraph : Object
{
@private
  DList         *_nodes;       // the list with the graph nodes
  DList         *_edges;       // the list with the edges
  
  DText         *_attributes;  // the optional (graphviz) attributes
  
  long           _nodeCounter; // the node counter (for the node name)
  long           _edgeCounter; // the edge counter (for the edge name)
}
#endif


@implementation DGraph


//// Constructors

//
// Initialise an empty graph 
//
// @return the object
//
- (DGraph *) init
{
  [super init];
  
  _nodes       = [DList new];
  _edges       = [DList new];
  
  _attributes  = nil;

  _nodeCounter = 1;
  _edgeCounter = 1;
  
  return self;
}

//
// Initialise a graph with attributes 
//
// @param attributes   the (graphviz) attributes
//
// @return the object
//
- (DGraph *) init :(const char *) attributes
{
  [self init];
  
  [self attributes :attributes];
  
  return self;
}


//// Copy related methods

//
// Do a shallow copy of the graph 
//
// @return the object
//
- shallowCopy
{
  DGraph *copy = [super shallowCopy];

  copy->_attributes = [_attributes copy];
  
  // shallow copy the lists
  copy->_nodes = [_nodes shallowCopy];
  copy->_edges = [_edges shallowCopy];

  // shallow copy the stored nodes
  {
    DListIterator *iter = [DListIterator alloc];
    DGraphNode    *node = nil;
    
    [iter init :copy->_nodes];
    
    node = [iter first];
    while (node != nil)
    {
      [iter object :[node shallowCopy]];
        
      node = [iter next];
    }
      
    [iter free];
  }
  
  // shallow copy the stored edges
  {
    DListIterator *iter = [DListIterator alloc];
    DGraphEdge    *edge  = nil;
    
    [iter init :copy->_edges];
    
    edge = [iter first];
    while (edge != nil)
    {
      [iter object :[edge shallowCopy]];
        
      edge = [iter next];
    }
      
    [iter free];
  }
  
  // replace the node references in the edges
  {
    DListIterator *cIter = [DListIterator alloc];
    DListIterator *oIter = [DListIterator alloc];
    DListIterator *eIter = [DListIterator alloc];
    
    DGraphNode    *cNode = nil;
    DGraphNode    *oNode = nil;
    DGraphEdge    *edge  = nil;
    
    [cIter init :copy->_nodes];
    [oIter init :_nodes];
    [eIter init :copy->_edges];
    
    cNode = [cIter first];
    oNode = [oIter first];
    
    while ((cNode != nil) && (oNode != nil))
    {
      edge = [eIter first];
      
      while (edge != nil)
      {
        [edge replace :oNode :cNode];
        
        edge = [eIter next];
      }
      
      cNode = [cIter next];
      oNode = [oIter next];
    }
    
    [cIter free];
    [oIter free];
    [eIter free];
  }
      
  // replace the edge references in the nodes
  {
    DListIterator *cIter = [DListIterator alloc];
    DListIterator *oIter = [DListIterator alloc];
    DListIterator *nIter = [DListIterator alloc];
    
    DGraphEdge    *cEdge = nil;
    DGraphEdge    *oEdge = nil;
    DGraphNode    *node  = nil;
    
    [cIter init :copy->_edges];
    [oIter init :_edges];
    [nIter init :copy->_nodes];
    
    cEdge = [cIter first];
    oEdge = [oIter first];
    
    while ((cEdge != nil) && (oEdge != nil))
    {
      node = [nIter first];
      
      while (node != nil)
      {
        [node replace :oEdge :cEdge];
        
        node = [nIter next];
      }
      
      cEdge = [cIter next];
      oEdge = [oIter next];
    }
    
    [cIter free];
    [oIter free];
    [nIter free];
  }
      
  return copy;
}

//
// Deepen the copied graph
// 
// @return the object
// 
- deepen
{
  [super deepen];
  
  [_nodes each :@selector(deepen)];
  [_edges each :@selector(deepen)];

  return self;
}


//// Destructor

//
// Free the graph (the stored objects are NOT freed)
//
// @return the object
//
- shallowFree
{
  [_nodes each :@selector(shallowFree)];
  [_nodes shallowFree]; _nodes = nil;

  [_edges each :@selector(shallowFree)];
  [_edges shallowFree]; _edges = nil;
  
  if (_attributes != nil)
  {
    [_attributes free]; _attributes = nil;
  }
  
  return [super free];
}

//
// Free the graph  (the stored objects are freed)
//
// @return the object
//
- free
{
  [_nodes each :@selector(free)];
  [_nodes shallowFree]; _nodes = nil;

  [_edges each :@selector(free)];
  [_edges shallowFree]; _edges = nil;
  
  if (_attributes != nil)
  {
    [_attributes free]; _attributes = nil;
  }
  
  return [super free];
}


//// Member methods

//
// Return the (graphviz) attributes of the graph
// 
// @return the attributes (or NULL)
// 
- (const char *) attributes 
{
  return (_attributes != nil) ? [_attributes cstring] : NULL;
}

//
// Set the (graphviz) attributes for the graph
// 
// @param attributes    the attributes (or NULL)
// 
// @return the graph object
// 
- (DGraph *) attributes :(const char *) attributes
{
  if ((attributes == NULL) || (*attributes == EOS))
  {
    if (_attributes != nil)
    {
      [_attributes free]; _attributes = nil;
    }
  }
  else
  {
    if (_attributes == nil)
    {
      _attributes = [DText new];
    }
    
    [_attributes set :attributes];
  }
  
  return self;
}


//// Main methods

//
// Add a (new) node to the graph (graph is the owner of the node)
// 
// @param label       the label for the node (or NULL)
// @param attributes  the attributes for the node (or NULL)
// @param object      the object for the node (or nil)
// 
// @return the (new) node object
// 
- (DGraphNode *) addNode :(const char *) label :(const char *) attributes :(id) object
{
  DGraphNode *node = [DGraphNode alloc];

  [node init :label :attributes :object];

  [self addNode :node];
  
  return node;
}

//
// Add a (new) node to the graph (graph is the owner of the node)
// 
// @param node        the graph node
// 
// @return success
// 
- (BOOL) addNode :(DGraphNode *) node
{
  BOOL ok = NO;
  
  if (node == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "node");
  }
  else if ([_nodes has :node])
  {
    WARNING(DW_UNKNOWN_WARNING, "node already in graph");
  }
  else
  {
    DText *tmp = [DText new];
    
    [tmp format :"n%ld", _nodeCounter++];
    
    [node name :[tmp cstring]];
      
    [_nodes append :node];
    
    [tmp free];
    
    ok = YES;
  }
  
  return ok;
}

//
// Add a (new) edge to the graph (graph is the owner of the edge object)
// 
// @param label       the label for the edge (or NULL)
// @param attributes  the attributes for the edge (or NULL)
// @param weight      the weight for the edge
// @param object      the object for the edge (or nil)
// @param source      the source node
// @param target      the target node
// 
// @return the (new) edge object (or nil)
// 
- (DGraphEdge *) addEdge :(const char *) label :(const char *) attributes :(double) weight :(id) object :(DGraphNode *) source :(DGraphNode *) target
{
  DGraphEdge *edge = [DGraphEdge alloc];

  [edge init :label :attributes :weight :object];

  if (![self addEdge :edge :source :target])
  {
    // problems..
    [edge shallowFree]; edge = nil;
  }
  
  return edge;
}

//
// Add a (new) edge to the graph (graph is the owner of the edge object)
// 
// @param edge        the graph edge
// @param source      the source node 
// @param target      the target node
// 
// @return success
// 
- (BOOL) addEdge :(DGraphEdge *) edge :(DGraphNode *) source :(DGraphNode *) target
{
  BOOL ok = NO;
  
  if (edge == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "edge");
  }
  else if ([_edges has :edge])
  {
    WARNING(DW_UNKNOWN_WARNING, "edge already in graph");
  }
  else if ((source == nil) || (target == nil))
  {
    WARNING(DW_NIL_NOT_ALLOWED, "source/target");
  }
  else if (![_nodes has :source])
  {
    WARNING(DW_UNKNOWN_WARNING, "source not in graph");
  }
  else if (![_nodes has :target])
  {
    WARNING(DW_UNKNOWN_WARNING, "target not in graph");
  }
  else
  {
    {
      DText *tmp = [DText new];
    
      [tmp format :"e%ld", _edgeCounter++];
    
      [edge name :[tmp cstring]];
      
      [tmp free];
    }
  
    ok = [edge connect :source :target];

    if (ok)
    {
      [_edges append :edge];
    }
  }
  
  return ok;
}

//
// Reroute an existing edge in the graph
//  
// @param edge        the edge
// @param source      the new source for the edge
// @param target      the new target for the edge
// 
// @return success
// 
- (BOOL) reroute :(DGraphEdge *) edge :(DGraphNode *) source :(DGraphNode *) target
{
  BOOL ok = NO;
  
  if (edge == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "edge");
  }
  else if (![_edges has :edge])
  {
    WARNING(DW_UNKNOWN_WARNING, "edge not in graph");
  }
  else if ((source == nil) || (target == nil))
  {
    WARNING(DW_NIL_NOT_ALLOWED, "source/target");
  }
  else if (![_nodes has :source])
  {
    WARNING(DW_UNKNOWN_WARNING, "source not in graph");
  }
  else if (![_nodes has :target])
  {
    WARNING(DW_UNKNOWN_WARNING, "target not in graph");
  }
  else
  {
    ok = [edge connect :source :target];
  }
  
  return ok;
}

//
// Determine the shortest path between two nodes; a (new) list with REFERENCES
// to the nodes in the shortest path are returned; IMPORTANT: free this list 
// with shallowFree
// 
// @param from      the start node
// @param to        the end node
// @param weight    the destination for the path weight (or NULL)
// 
// @return a (new) list with REFERENCES to the nodes in the shortes path (or nil)
//
- (DList *) shortestPath :(double *) weight :(DGraphNode *) from :(DGraphNode *) to
{
  DList *path = nil;
  
  if ((from == nil) || (![_nodes has :from]))
  {
    WARNING(DW_INVALID_ARG, "from");
  }
  else if ((to == nil) || (![_nodes has :to]))
  {
    WARNING(DW_INVALID_ARG, "to");
  }
  else
  {
    DGraphNode *node  = from;
    DList      *todo  = [DList new];
    
    // Reset all nodes
    [_nodes each :@selector(path)];
    
    // Initialise the start node
    [node path :nil :0.0];
    
    // Process until the 'to' node is found or no more nodes can be processed
    while ((node != to) && (node != nil))
    {
      DListIterator *iter  = [node outgoingEdges];
      DGraphEdge    *edge  = nil;
      
      // Process all outgoing edges
      edge = [iter first];
      while (edge != nil)
      {
        DGraphNode *target = [edge target];

        if (target != nil)
        {
          double sum = [node sum] + [edge weight];
        
          if (sum < [target sum])
          {
            [target path :node :sum];
          
            // Add target node if not already there
            if (![todo has :target])
            {
              [todo append :target];
            }
          }
        }
        
        edge = [iter next];
      }
      [iter free];
      
      // Determine the node with the lowest sum for next processing
      {
        DGraphNode *walk = nil;
        double      sum  = DBL_MAX;
        
        iter = [DListIterator alloc];
        
        [iter init :todo];
        
        node = nil;
        walk = [iter first];
        while (walk != nil)
        {
          if ([walk sum] < sum)
          {
            node = walk;
            sum  = [walk sum];
          }
          walk = [iter next];
        }
        [iter free];
        
        if (node != nil)
        {
          [todo remove :node];
        }
      }
    }
    [todo shallowFree];
    
    if (node == to)
    {
      if (weight != NULL)
      {
        *weight = [to sum];
      }
      
      // fill nodes with the shortest path
      path = [DList new];
     
      node = to;
      do
      {
        [path prepend :node];
        
        node = [node prev];
      }
      while (node != nil);
    }
  }
  
  return path;
}

//
// Check if the graph has a node
// 
// @param node      the graph node
// 
// @return has it ?
// 
- (BOOL) hasNode :(DGraphNode *) node
{
  return ((node != nil) && ([_nodes has :node]));
}

//
// Check if the graph has an edge
// 
// @param edge      the graph edge
// 
// @return has it ?
// 
- (BOOL) hasEdge :(DGraphEdge *) edge
{
  return ((edge != nil) && ([_edges has :edge]));
}

//
// Remove an node from the graph (node must be unconnected)
// 
// @param node      the graph node
// 
// @return the object stored in the node (or nil)
// 
- (id) removeNode :(DGraphNode *) node
{
  id object = nil;
  
  if (node == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "node");
  }
  else if ([node degree] > 0)
  {
    WARNING(DW_UNKNOWN_WARNING, "node still connected");
  }
  else
  {
    if ([_nodes remove :node])
    {
      object = [node object];
      
      [node shallowFree];
    }
    else
    {
      WARNING(DW_UNKNOWN_WARNING, "node not in graph");
    }
  }
  
  return object;
}

//
// Remove an edge from the graph (auto disconnect)
// 
// @param edge      the edge node
// 
// @return the object stored in the edge (or nil)
// 
- (id) removeEdge :(DGraphEdge *) edge
{
  id object = nil;
  
  if (edge == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "edge");
  }
  else if (![_edges has :edge])
  {
    WARNING(DW_UNKNOWN_WARNING, "edge not in graph");
  }
  else
  {
    if (([edge source] != nil) || ([edge target] != nil))
    {
      [edge disconnect];
    }
    
    if ([_edges remove :edge])
    {
      object = [edge object];
      
      [edge shallowFree];
    }
  }
  
  return object;
}


//// Iterators

//
// Return a (new) iterator on all the nodes in the graph
// 
// @return a (new) iterator
// 
- (DListIterator *) nodes
{
  DListIterator *iter = [DListIterator alloc];
  
  [iter init :_nodes];
  
  return iter;
}

//
// Return a (new) iterator on all the edges in the graph
// 
// @return a (new) list iterator
// 
- (DListIterator *) edges
{
  DListIterator *iter = [DListIterator alloc];
  
  [iter init :_edges];
  
  return iter;
}


//// Export methods

//
// Export to DOT 
// 
// @param writer    the output object
// 
// @return success ?
// 
- (BOOL) toDot :(id <DTextWritable>) writer
{
  BOOL ok = NO;
  
  if (writer == nil)
  {
    WARNING(DW_NIL_NOT_ALLOWED, "writer");
  }
  else
  {
    DListIterator *iter = nil;
    DGraphEdge    *edge = nil;
    DGraphNode    *node = nil;

    
    ok  = [writer writeLine :"digraph G {"];
    
    if ([self attributes] != NULL)
    {
      ok &= [writer writeLine :"  /* the graph attributes */"];
      ok &= [writer writeText :"  "];
      ok &= [writer writeText :[self attributes]];
      ok &= [writer writeLine :";"];
    }
    
    
    ok &= [writer writeLine :"  /* the edges */"];
    
    iter = [self edges];
    
    edge = [iter first];
    
    while (edge != nil)
    {
      if (([edge source] != nil) && ([edge target] != nil))
      {
        const char *edgeLabel = [edge label];
        const char *edgeAttr  = [edge attributes];
        
        // add the edge
        ok &= [writer writeText :"  "]; // ident
        ok &= [writer writeText :[[edge source] name]];
        ok &= [writer writeText :" -> "];
        ok &= [writer writeText :[[edge target] name]];
        
        // add the edge label and/or attributes
        if ((edgeLabel != NULL) || (edgeAttr != NULL))
        {
          ok &= [writer writeText :" ["];
          
          if (edgeLabel != NULL)
          {
            ok &= [writer writeText :"label=\""];
            ok &= [writer writeText :edgeLabel];
            ok &= [writer writeText :"\""];
          }
          
          if ((edgeLabel != NULL) && (edgeAttr != NULL))
          {
            ok &= [writer writeText :","];
          }
          
          if (edgeAttr != NULL)
          {
            ok &= [writer writeText :edgeAttr];
          }
          
          ok &= [writer writeText :"]"];
        }
        
        ok &= [writer writeLine :";"];
      }
      
      edge = [iter next];
    }
    [iter free];
    

    ok &= [writer writeLine :"/* the node attributes */"];

    iter = [self nodes];
    
    node = [iter first];
    
    while (node != nil)
    {
      const char *nodeLabel = [node label];
      const char *nodeAttr  = [node attributes];
      
      if ((nodeLabel != NULL) || (nodeAttr != NULL))
      {
        // add the node
        ok &= [writer writeText :"  "]; // ident
        ok &= [writer writeText :[node name]];
        
        // add the node label and/or the attributes
        ok &= [writer writeText :" ["];
          
        if (nodeLabel != NULL)
        {
          ok &= [writer writeText :"label=\""];
          ok &= [writer writeText :nodeLabel];
          ok &= [writer writeText :"\""];
        }
          
        if ((nodeLabel != NULL) && (nodeAttr != NULL))
        {
          ok &= [writer writeText :","];
        }
          
        if (nodeAttr != NULL)
        {
          ok &= [writer writeText :nodeAttr];
        }
          
        ok &= [writer writeText :"]"];
      }
      
      node = [iter next];
    }
    [iter free];
    
    ok &= [writer writeLine :"}"];
  }
  return ok; 
}

@end

/*==========================================================================*/
