#ifndef _DHTTPCLIENT_H
#define _DHTTPCLIENT_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DHTTPClient - the HTTP client class in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-13 17:14:32 $ $Revision: 1.14 $
//
//==============================================================================
//
// ToDo:
// 1. Chunked transfer
// 2. POST/PUT
// 3. Cookies
//

#include "ofc/config.h"

#include "ofc/DURL.h"
#include "ofc/DSocket.h"
#include "ofc/DHashTable.h"


#define DHTTP_PORT     (80)

#define DHC_UNKNOWN    (0)
#define DHC_HEAD       (1)
#define DHC_GET        (2)

//
// The DHTTPClient class implements a HTTP client. The class does not support
// the POST and PUT requests. Also the responses with a chunked transfer 
// encoding are not handled.
//
// @example
// #include <stdio.h>
// #include "ofc/DHTTPClient.h"
// 
// int main(int argc, char *argv[])
// {
//   DHTTPClient   *client = [DHTTPClient new];
//   DURL          *url    = [DURL        new];
//   DHashIterator *iter;
// 
//   [url url :"http://ofc.dvoudheusden.net/index.html"];
// 
//   printf("Send the request for a http page..\n");
//   if (![client sendStartRequest :DHC_GET :url])  // send a GET for index.html
//     printf("Error in sendStartRequest..\n");
// 
//   if (![client sendConnectionHeader :"close"])   // send that the connection should be closed
//     printf("Error in sendConnectionHeader..\n");
// 
//   if (![client sendUserAgentHeader :"libofc"])   // send the user agent
//     printf("Error in sendUserAgentHeader..\n");
// 
//   if (![client sendEndRequest])                  // send the end of the request
//     printf("Error in sendEndRequest..\n");
// 
//   printf("Wait for the response of the server..\n");
//   if ([client receiveReply])                     // wait for the response of the server
//   {
//     printf("Response reason:%d http version:%d.%d body length:%ld\n", 
//       [client reason], [client peerMajor], [client peerMinor], [client bodyLength]);
// 
//     iter = [client headers];                     // print the headers in the reply
// 
//     if (iter != nil)
//     {
//       DText *value = [iter first];
//     
//       printf("Headers in reply:\n");
//       while (value != nil)
//       {
//         DText *key = [iter key];
// 
//         printf("%s = %s\n", [key cstring], [value cstring]);
// 
//         value = [iter next];
//       }
//       [iter free];
//     }
//     else
//       printf("No headers in reply..\n");
// 
//   }
//   else
//     printf("Error in receiveReply..\n");
// 
//   [client free];                       // Cleanup
//   [url    free];
// 
//   return 0;
// }
// 

@interface DHTTPClient : Object
{
@private
  DSocket            *_client;       // the client socket
  BOOL                _shouldOpen;   // should the socket be opened?
  int                 _state;        // the state of the client

  DText              *_server;       // the server host name (or proxy)
  DInetSocketAddress *_address;      // the address of the server (or proxy)

  int                 _type;         // the request type (DC_HEAD, DHC_GET, ..)
  int                 _major;        // the major version number (def. 1)
  int                 _minor;        // the minor version number (def. 1)
  BOOL                _shouldClose;  // should the connection close after comm.?
  DText               *_request;     // the request

  DData               *_response;    // the respone
  int                  _respIndex;   // the index in the response string
  DText               *_line;        // one line of the response
  int                  _reason;      // the received reply reason
  DText               *_reasonText;  // the received reason text
  DHashTable          *_headers;     // the received headers
  DData               *_body;        // the received body in data
  BOOL                 _useChunked;  // is chunked used ?
  int                  _length;      // the length of the body
  int                  _peerMajor;   // the major version number of the peer
  int                  _peerMinor;   // the minor version number of the peer
}


//// Constructors

//
// Initialise the HTTP Client
//
// @return the object
//
- (DHTTPClient *) init;

//
// Initialise the HTTP Client with a client version
//
// @param major    the major version number
// @param minor    the minor version number
// 
// @return the object
//
- (DHTTPClient *) init :(int) major :(int) minor;


//// Deconstructor

//
// Free the HTTP client object
//
// @return the object
//
- free;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
//
- shallowCopy;


//// Member methods

//
// Return the local client object
// 
// @return a reference to the client object
//
- (DSocket *) socket;


//// Sending methods

// 
// Send the start of the request
//
// @param request      the request (DHC_HEAD, DHC_GET ..) 
// @param path         the (absolute) url to get
// 
// @return success
// 
- (BOOL) sendStartRequest :(int) request :(DURL *) path;

// 
// Send the start of the request via a proxy
//
// @param request      the request (DHTTP_HEAD, DHTTP_GET, ..)
// @param path         the (absolute) url to get
// @param proxy        the host name of the proxy (or NULL for no proxy)
// @param proxyPort    the proxy port number
// 
// @return success
// 
- (BOOL) sendStartRequest :(int) request :(DURL *) path :(const char *) proxy :(int) proxyPort;

//
// Send a (general) header after the request
// 
// @param header       the header to send
// @param argument     the argument for the header
// 
// @return success
// 
- (BOOL) sendHeader :(const char *) header :(const char *) argument;

//
// Send Accept-Language header
// 
// @param languages   a string containing the languages (en-us, en)
//
// @return success
// 
- (BOOL) sendAcceptLanguageHeader :(const char *) languages;

//
// Send User-Agent header
// 
// @param agent     a c-string with the user agent
// 
// @return success
//
- (BOOL) sendUserAgentHeader :(const char *) agent;

//
// Send Referer header
// 
// @param referer   a c-string with the referer link
// 
// @return success
// 
- (BOOL) sendRefererHeader :(const char *) referer;

//
// Send Connection header
// 
// @param connection   a c-string with "close" or "keep-alive"
// 
// @return success
// 
- (BOOL) sendConnectionHeader :(const char *) connection;

//
// Send the end of the request to the server/proxy
// 
// @return success
// 
- (BOOL) sendEndRequest;


//// Receiving methods

//
// Receive a reply from the server
// 
// @return success
// 
- (BOOL) receiveReply;

//
// Return the major version number of the peer (after receiveReply)
// 
// @return the major version number
//
- (int) peerMajor;

//
// Return the minor version number of the peer (after receiveReply)
// 
// @return the minor version number
//
- (int) peerMinor;

//
// Get the reason (after receiveReply)
// 
// @return the reason code (or -1)
// 
- (int) reason;

//
// Get the reason text (after receiveReply)
// 
// @return the reason text (or NULL for error)
// 
- (const char *) reasonText;

//
// Get the argument of a header (after receiveReply)
// 
// @param header   the header (all lower case)
// 
// @return a (new) text object with the argument (or nil)
// 
- (DText *) header :(const char *) header;

//
// Get all received headers (after receiveReply)
// 
// @return a (new) iterator on the hash table with all the received headers (or nil)
// 
- (DHashIterator *) headers;

//
// Get the body of the received message (after receiveReply)
// 
// @return a received data (or NULL for error)
// 
- (const unsigned char *) body;

//
// Get the length of the body of the received message (after receiveReply)
// 
// @return a received length
// 
- (unsigned long) bodyLength;


//// Response processing methods (private methods)

//
// Send the first line to the server
// 
// @param path      the path of the server
// @param proxy     the name of the proxy (or NULL)
// @param proxyPort the port of the proxy (or 0)
// @param request   the request for the server
// 
// @return success
//
- (BOOL) _sendFirstRequest :(DURL *) path :(const char *) proxy :(int) proxyPort :(const char *) request;

//printf("Shutdown\n");
//printf("Shutdown\n");
//printf("Connecting(%d) to %s port %d\n", ok, [[_address host] cstring], [_address port]);
//printf("Sending(%d):%s", ok, [_request cstring]);
//
// Start processing the response
// 
// @return none
// 
- (void) _startProcessingResponse;

//
// Receive a line from the client
// 
// @return success
// 
- (BOOL) _receiveLine;

//printf("Received:%s\n", [_line cstring]);
//
// Process the first line of the response
// 
// @return none
// 
- (void) _processFirstLine;

//printf("peer version : %d.%d\n", _peerMajor, _peerMinor);
//printf("Reason : %d Text: %s\n", _reason, [_reasonText cstring]);
//
// Process the headers in the response
// 
// @return none
// 
- (void) _processHeaders;

//printf("header : %s value : %s\n", [header cstring], [value cstring]);
//
// Process non-chunked data in the response
// 
// @return none
// 
- (void) _processNonChunkedData;

//printf("Body length: %d\n", _length);
@end
#endif
