#ifndef _DINTARRAY_H
#define _DINTARRAY_H

// This file was automatically generated by objch.

//==============================================================================
//
//  DIntArray - the one dimensional array of integers class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-02 05:58:14 $ $Revision: 1.8 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DComparable.h"
#include "ofc/DSizeable.h"
#include "ofc/DParsable.h"

#include "ofc/DText.h"

//
// The DIntArray class implements an one dimensional array of integers. It is of 
// course possible to create an array of integers with the classes DArray and DInt, 
// but this is not very efficient and a little bit clumsy. This class also implements some
// statistical methods.
// 
// @example
// #include <stdio.h>
// #include "ofc/DIntArray.h"
// 
// int main(int argc, char *argv[])
// {
//   DIntArray *arr1  = [DIntArray alloc];
//   DIntArray *arr2  = [DIntArray new  ];
//   DText     *str;
// 
//   int       int1[] = {9, 4, 2, 9, 3, 7};
//   int       int2[] = {3, 2, 1};
// 
//   [arr1 init :int1 :sizeof(int1)/sizeof(int)];  // Init with an array of integers
// 
//   printf("Length of array1:%ld.\n", [arr1 length]);
// 
//   str = [arr1 toText];                 // Convert the array to a string
//   printf("Array1 as string:%s.\n", [str cstring]);
//   [str free];
// 
//   [arr1 put :3 :8];                    // Change value on index 3 from 9 to 8
// 
//   [arr1 insert :2 :2 :int2 :sizeof(int2)/sizeof(int)]; // Replace value on index 2 with the int2 array
// 
//   [arr1 delete :4 :5];                 // Delete two values from the array
// 
//   str = [arr1 toText];
//   printf("Array1 as string:%s.\n", [str cstring]); // Print the modified string
//   [str free];
// 
//                                        // Searching in the array
//   printf("Value 3 is present %ld times in array1.\n", [arr1 count :3 :0 :-1]); // Count the occurences of an integer
// 
//   printf("First index of value 3:%ld.\n", [arr1 index :3 :0 :-1]);  // Find first index
// 
//                                        // Statistical methods
//   printf("Sum of array1:%ld\n",  [arr1 sum :0 :-1]);
//   printf("Max of array1:%d\n",   [arr1 max :0 :-1]);
//   printf("Min of array1:%d\n",   [arr1 min :0 :-1]);
//   printf("Avg of array1:%.1f\n", [arr1 average :0 :-1]);
//   printf("Std of array1:%.1f\n", [arr1 standardDeviation :0 :-1]);
//   
//   [arr1 sort :0 :-1];                  // Sort the array
// 
//   str = [arr1 toText];
//   printf("Array1 sorted as string:%s.\n", [str cstring]);
//   [str free];
// 
//                                        // Use array2 as lifo/fifo buffer
//   [arr2 size :50];                     // Insure the size of the buffer
// 
//   [arr2 push :1];
//   [arr2 push :2];
//   [arr2 push :3];
// 
//   printf("Length of lifo:%ld.\n", [arr2 length]);
//   printf("Top of lifo:%d.\n", [arr2 tos]);
//   printf("Pop value (%d) and another (%d).\n", [arr2 pop], [arr2 pop]);
// 
//   [arr2 clear];
// 
//   [arr2 enqueue :4];
//   [arr2 enqueue :5];
//   [arr2 enqueue :6];
//   [arr2 enqueue :7];
// 
//   printf("Length of fifo:%ld.\n", [arr2 length]);
//   printf("Dequeue value (%d) and another (%d).\n", [arr2 dequeue], [arr2 dequeue]);
// 
//                                        // Compare arrays
//   if ([arr1 compare :arr2] == 0)
//     printf("Array1 is equal to array2.\n");
//   else if ([arr1 compare :arr2] < 0)
//     printf("Array1 is smaller than array2.\n");
//   else 
//     printf("Array1 is greater than array2.\n");
// 
//   [arr1 free];                         // Cleanup
//   [arr2 free];
//     
//   return 0;
// }
// 

@interface DIntArray : Object <DTextable,DComparable,DSizeable,DParsable>
{
@private
  unsigned long  _size;     // the allocated size for the data
  unsigned       _extra;    // the extra size amount
  unsigned long  _length;   // the length of the data
  int           *_array;    // the array of integers
}


//// Constructors

//
// Initialise to an empty array
//
// @return the object
//
- (DIntArray *) init;

//
// Initialise the object with integers
//
// @param ints     the integers
// @param length   the length of the integers
//
// @return the string object
//
- (DIntArray *) init :(const int *) ints :(unsigned long) length;


//// Copy related methods

//
// Do a shallow copy of the array
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free;


//// Member methods

//
// Check if the array is empty
//
// @return Is it empty ?
//
- (BOOL) isEmpty;

//
// Return the length of the array
//
// @return the length
//
- (unsigned long) length;

//
// Return the pointer to the array of integers
//
// @return the pointer to the array
//
- (const int *) array;


//// Size methods

//
// Insure the size for the array
//
// @param size     the requested size (in elements) for the array
//
// @return the object
//
- (DIntArray *) size :(unsigned long) size;

//
// Set the amount of extra size increase if the size must be extended
//
// @param extra    the amount (def. 0)
//
// @return the object
//
- (DIntArray *) extra :(unsigned) extra;


//// DTextable protocol methods

//
// Convert the object to a text string 
// 
// @return a (new) text string object
//
- (DText *) toText;


//// DParsable protocol methods

//
// Parse a string for an array of integers: <int>,<int>;<int>
// 
// @param cstr     the string to be parsed (moved to the first non-parsable character)
// 
// @return the result (0, ERANGE, ENODATA)
//
- (int) fromString :(char **) cstr;


//// Array setters

//
// Clear the array
// 
// @return the object
// 
- (DIntArray *) clear;

//
// Set an array of integers in the object
//
// @param ints     the integers
// @param length   the length of the integers
//
// @return the object
//
- (DIntArray *) set :(const int *) ints :(unsigned long) length;


//// Element methods

//
// Put an integer at the given index
//
// @param index     the index (-length..+length)
// @param element   the integer to be placed
//
// @return the object
//
- (DIntArray *) put :(long) index :(int) element;

//
// Get the integer at the given index
//
// @param index    the index 
//
// @return the integer from the array
//
- (int) get :(long) index;

// 
// Insert an integer at the given index
// 
// @param index    the index
// @param value    the value to be inserted
// 
// @return the object
//
- (DIntArray *) insert :(long) index :(int) value;

//
// Delete the integer at the given index
//
// @param index    the index
//
// @return the object
//
- (DIntArray *) delete :(long) index;


//// Slice methods

//
// Insert an array of integers in a part of the array in the object
//
// @param from     the start index
// @param to       the end index (to >= (from-1))
// @param ints     the array with integers to be inserted
// @param length   the length of the integers
//
// @return the object
//
- (DIntArray *) insert :(long) from :(long) to :(const int *) ints :(unsigned long) length;

//
// Return a sub array
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return a (new) integer array object
//
- (DIntArray *) get :(long) from :(long) to;

//
// Delete a range in the array
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the object
//
- (DIntArray *) delete :(long) from :(long) to;


//// Array manipulation methods

//
// Append an array to the object
//
// @param ints     the integers to be appended
// @param length   the length of the integer array
//
// @return the object
//
- (DIntArray *) append :(const int *) ints :(unsigned long) length;

//
// Prepend an array in front of the object
//
// @param ints     the integers to be prepended
// @param length   the length of the integer array to be prepended
//
// @return the object
//
- (DIntArray *) prepend :(const int *) ints :(unsigned long) length;


//// Stack methods

//
// Push one integer at the end of the array
// 
// @param value    the integer that is placed at the end
//
// @return the object
//
- (DIntArray *) push :(int) value;

//
// Pop one integer from the end of the array
//
// @return the integer that is pulled (or 0 for empty stack)
//
- (int) pop;

//
// Return the integer at the end of the array
// 
// @return the integer (or 0 for empty stack)
// 
- (int) tos;


//// Queue methods (FIFO)

//
// Put an integer at the start of the array 
//
// @param value    the integer 
// 
// @return success
// 
- (BOOL) enqueue :(int) value;

//
// Get the integer at the end of the array
// 
// @return the integer (or 0 for empty queue)
// 
- (int) dequeue;


//// Comparable protocol methods

//
// Compare the integer array object with another integer array object
//
// @param other    the object that should be compared (!= nil)
//
// @return the compare result (-1, 0, 1)
//
- (int) compare :(DIntArray *) other;


//// Comparison methods

//
// Compare the object with another array of integers
//
// @param ints     the array of integers
// @param length   the length of the array
//
// @return the compare result (-1, 0, 1)
//
- (int) bcompare :(const int *) ints :(unsigned long) length;


//// Statistical methods

//
// Count the number of occurrences of an integer in the array
//
// @param search   the value that should be counted
// @param from     the start index 
// @param to       the end index (to >= from)
//
// @return the number of occurrences
//
- (unsigned long) count :(int) search :(long) from :(long) to  ;

//
// Return the first index where the search integer is found
//
// @param search   the integer that should be searched
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the first index or -1 = not found
//
- (long) index :(int) search :(long) from :(long) to ;

//
// Return the last index where the search integer is found
//
// @param search   the integer that should be searched
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the last index or -1 = not found
//
- (long) rindex :(int) search :(long) from :(long) to;

//
// Calculate the sum in a range of the integer array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the sum of the integers
// 
- (long) sum :(long) from :(long) to;

//
// Determine the maximum integer in a range of the integer array
//
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the maximum integer
// 
- (int) max :(long) from :(long) to;

//
// Determine the minimum integer in a range of the integer array
//
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the minimum integer
// 
- (int) min :(long) from :(long) to;

//
// Calculate the average in a range of the integer array
// 
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the average (or 0.0)
// 
- (double) average :(long) from :(long) to;

// 
// Calculate the variance in a range of the integer array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the variance
// 
- (double) variance :(long) from :(long) to;

//
// Calculate the standard deviation in a range of the integer array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the standard deviation
// 
- (double) standardDeviation :(long) from :(long) to;

//
// Sort a range in the the integer array
//
// @param from     the start index
// @param to       the end index (to > from)
// 
// @return the object
// 
- (DIntArray *) sort :(long) from :(long) to;

//
// Invert a range in the integer array
// 
// @param from     the start index
// @param to       the end index (to >= from)
// 
// @return the object
// 
- (DIntArray *) invert :(long) from :(long) to;

@end
#endif
