#ifndef _DLRND_H
#define _DLRND_H

// This file was automatically generated by objch.

//==============================================================================
//
//       DLRnd - the (long) pseudo random generator in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-02 05:58:14 $ $Revision: 1.6 $
//
//==============================================================================

#include <stdint.h>

#include "ofc/config.h"

#include "ofc/DRandomable.h"

//
// The DLRnd class implements a pseudo random generator bases on a 
// generalized feedback shift register random generator. It is not as 
// fast as the generator used in the @class(DRnd) but this one passes the 
// Marsaglia's 'Diehard' statistical tests and it has a long period 
// of 2^800. This class is 'inspired' by the code present on the site
// @link(http://remus.rutgers.edu/~rhoads)
//
// @example
// #include <stdio.h>
// #include "ofc/DLRnd.h"
// 
// int main(int argc, char *argv[])
// {
//   DLRnd *rnd1 = [DLRnd alloc];
//   DLRnd *rnd2 = [DLRnd new  ];
// 
//   [rnd1 init :12345678];               // Init with a seed value
// 
//                                        // Get pseudo random values
//   printf("Pseudo random long value:%ld.\n", [rnd1 nextLong]);
//   printf("Pseudo random int value with range [1,100]:%d.\n", [rnd1 nextInt :1 :100]);
// 
//   [rnd2 seed :87654321];              // Set a seed value
// 
//                                       // Get pseudo random values
//   printf("Pseudo random double value:%f.\n", [rnd2 nextDouble]);
//   printf("Pseudo random double value with range [0.1,0.3]:%f.\n", [rnd2 nextDouble :0.1 :0.3]);
// 
//   [rnd1 free];                        // Cleanup
//   [rnd2 free];
// 
//   return 0;
// }
// 

@interface DLRnd : Object <DRandomable>
{
@private
  uint32_t _seeds[25];  // the seeds for the generator
  int      _index;      // the index in the seeds
}


//// Constructors

//
// Initialise a non-seeded random generator
//
// @return the object
//
- (DLRnd *) init;

//
// Initialise a seeded random generator
//
// @param seed     the seed to be used
// 
// @return the object
//
- (DLRnd *) init :(unsigned long) seed;


//// Member methods

//
// Set the seed for the random generator
// 
// @param seed     the seed for the random generator
//
// @return the object
//
- (DLRnd *) seed :(unsigned long) seed;


//// Random generator methods

//
// Get the next long from the generator
//  
// @return the next long
//
- (long) nextLong ;

//
// Get the next ranged long from the generator
// 
// @param from     the start of the range
// @param to       the end of the range
// 
// @return the next long
// 
- (long) nextLong :(long) from :(long) to;

//
// Get the next integer from the generator
//  
// @return the next integer
//
- (int) nextInt;

//
// Get the next ranged integer from the generator
// 
// @param from     the start of the range
// @param to       the end of the range
// 
// @return the next integer
// 
- (int) nextInt :(int) from :(int) to;

//
// Get the next double from the generator (0 <= d < 1)
//  
// @return the next double
//
- (double) nextDouble;

//
// Get the next double from the generator (from <= dE < to)
// 
// @param from     the start of the range
// @param to       the end of the range
// 
// @return the next double
// 
- (double) nextDouble :(double) from :(double) to;

@end
#endif
