#ifndef _DLIST_H
#define _DLIST_H

// This file was automatically generated by objch.

//==============================================================================
//
//         DList - the double linked list class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-03 18:31:55 $ $Revision: 1.17 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DText.h"

#include "ofc/DStackable.h"
#include "ofc/DQueuable.h"


typedef struct _DListNode
{
  struct _DListNode  *_next;   // the next node (or NULL)
  struct _DListNode  *_prev;   // the previous node (or NULL)
  id                  _object; // the object in the node (or nil)
} DListNode;


//
// The DList class implements a number of methods for manipulating double linked
// lists. 
//
// @example
// #include <stdio.h>
// #include "ofc/DList.h"
// #include "ofc/DInt.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DList         *list = [DList new];
//   DListIterator *iter;
//   DInt          *nr;
//   DText         *str;
//   DInt          *nr7;
//   id             obj;
//                                        // Add strings and numbers to the list
//   nr = [DInt alloc]; [nr init :4];
//   [list append :nr];                   // Append objects in list
// 
//   nr = [DInt alloc]; [nr init :8];
//   [list append :nr];
// 
//   str = [DText alloc]; [str init :"hello"];
//   [list append :str];
// 
//   nr7 = [DInt alloc]; [nr7 init :7];
//   [list append :nr7];
// 
//   str = [ DText alloc]; [str init :"bye"];
//   [list append :str];
// 
//   printf("List has %ld elements.\n", [list length]); // Get the lenght of the list
// 
//   nr = [DInt alloc]; [nr init :3];
//   [list insert :3 :nr];                // Insert an object in the list
// 
//   str = [DText alloc]; [str init :"replace"];
//   nr = [list set :1 :str];             // Index 2 is replaced by "replace"
//   [nr free];
// 
//   str = [list get :2];                 // Get an object from the list
//   printf("Element 2 is \"%s\".\n", [str cstring]);
// 
//   printf("List contains object nr7 %ld times.\n", [list count :nr7]); // Count occurences
// 
//   printf("Index of object nr7 is %ld.\n", [list index :nr7]); // Find index of object
// 
//   if ([list remove :nr7])              // Remove an object from the list
//   {
//     printf("Object nr7 is removed from the list.\n");
//     [nr7 free];
//   }
//   else
//     printf("Object nr7 could not be found in the list.\n");
// 
//   str = [list join :','];              // Join all objects text to one text object
//   printf("Join of all objects text in the list:%s.\n", [str cstring]);
//   [str free];
// 
//                                        // Iterate the list
//   iter = [DListIterator new];
// 
//   [iter list :list];                   // Set the list in the iterator
// 
//   obj = [iter last];                   // Iterate the list backwards
//   while (obj != nil)
//   {
//     if ([obj isKindOf :[DText class]]) // Find out the type of the object and ..
//     {
//       str = obj;
//       printf("Element: %s.\n", [str cstring]); // .. print the contents
//     }
//     else
//       if ([obj isKindOf :[DInt class]])
//       {
//         nr = obj;
//         printf("Element: %d\n", [nr get]);
//       }
//       else 
//         printf("Element: unknown.\n");
//     obj = [iter prev];
//   }
// 
//   [iter free];                         // Cleanup
//   [list free];
// 
//   return 0;
// }
// 

@interface DList : Object <DStackable,DQueuable,DTextable>
{
@private
  DListNode      *_first;   // the start of the list
  DListNode      *_last;    // the end of the list
  long            _count;   // the number of nodes in the list
}


//// Constructors

//
// Initialise an empty list
//
// @return the list
//
- (DList *) init;


//// Copy related methods

//
// Do a shallow copy of the object
// 
// @return the copy
// 
- shallowCopy;

//
// Deepen a copied object
//
// @return the object
//
- deepen;


//// Deconstructor

//
// Free the list (the stored objects are *NOT* freed)
//
// @return the instance
//
- shallowFree;

//
// (Deep) free the list (the stored objects are *FREED*)
//
// @return the instance
//
- free;


//// Member methods

//
// Check if the list is empty
//
// @return is the list empty ?
//
- (BOOL) isEmpty;

// Return the length of the list
//
// @return the number of nodes in the list
//
- (long) length;


//// List manipulation methods

//
// Prepend an object to the list
//
// @param object   the object that must be prepended
//
// @return the list
//
- (DList *) prepend :(id) object;

//
// Append an object to the list
//
// @param object   the object that must be appended
//
// @return the instance
//
- (DList *) append :(id) object;

//
// Insert an object in the list on the given index
//
// @param index    the index in the list (-n..+n)
// @param object   the object to be inserted
//
// @return the list
//
- (DList *) insert :(long) index :(id) object;

//
// Set the object in a list node
//
// @param index    the index in the list
// @param object   the object to be set
// 
// @return the previous stored object
//
- (id) set :(long) index :(id) object;

//
// Get the object in a list node
//
// @param index    the index in the list
//
// @return the stored object
//
- (id) get :(long) index;

//
// Get a range of objects from the list
//
// @param from     the start index from the range
// @param to       the end index from the range
//
// @return a (new) list with (references) to the objects
//
- (DList *) get :(long) from :(long) to;

//
// Delete a list node
//
// @param index    the index in the list
//
// @return the stored object
//
- (id) delete :(long) index;

//
// Delete a range of list nodes
//
// @param from    the first index in the list
// @param to      the last index in the list (to >= from)
//
// @return a (new) list with the deleted list nodes
//
- (DList *) delete :(long) from :(long) to;

//
// Count the number of occurrences of an object in the list
//
// @param object   the object that must be counted (incl. nil)
//
// @return the number of occurrences
//
- (long) count :(id) object;

//
// Find the smallest index where an object is located
//
// @param object   the object that must be located
//
// @return the index in the list (or -1 if not found)
//
- (long) index :(id) object;

//
// Find the largest index where an object is located
//
// @param object   the object that must be located in the list
//
// @return the index in the list (or -1 if not found)
//
- (long) rindex :(id) object;

//
// Check if the object is in the list
//
// @param object   the object that must be located in the list
//
// @return is the object in the list ?
//
- (BOOL) has :(id) object;

//
// Replace in the list max times every occurence of the original object
// by the replacement (orig in NOT freed)
//
// @param orig     the original object
// @param repl     the replacement object
// @param max      the maximum number of replacements (-1 = all)
//
// @return the number of actual replacements
//
- (long) replace :(id) orig :(id) repl :(long) max;

//
// Remove the first occurrence of an object in the list
//
// @param object   the object that must be removed from the list
//
// @return success
//
- (BOOL) remove :(id) object;

//
// Reverse the list
//
// @return the list object
//
- (DList *) reverse;

//
// Perform a method for every object in the list
//
// @param sel      the selector of the method
//
// @return the object
//
- (DList *) each :(SEL) sel;


//// Stackable protocol methods

//
// Push a new node with the object at the end of the list
//
// @param obj      the object to be pushed
//
// @return success
//
- (BOOL) push :(id) obj;

//
// Pop the last node from the list and returns the object
//
// @return the object from the list (or nil)
//
- (id) pop;

//
// Return the object in the last node of the list
//
// @return the object (or nil if empty)
//
- (id) tos;


//// Queuable protocol methods

//
// Enqueue the object at the end of the list
//
// @param obj      the object 
//
// @return success
//
- (BOOL) enqueue :(id) obj;

//
// Dequeue the first object from the list
//
// @return the object (or nil for buffer empty)
//
- (id) dequeue;


//// String methods

//
// Split a string in a list of strings
//
// @param string    the string to be splitted
// @param seperator the seperator character
// @param maxsplits the maximum number of splits (or -1 for all)
//
// @return a (new) list with string objects
//
+ (DList *) split :(const char *) string :(char) seperator :(int) maxsplits;

//
// Join the list of strings to one string
//
// @param seperator the seperator to be used
//
// @return the new string with the joined strings
//
- (DText *) join :(char) seperator;

//
// Convert the list to a string
//
// @return the new text string
//
- (DText *) toText;

@end
  
//
// The DListIterator class implements a number of methods for iterating lists.
//
// // @example
// #include <stdio.h>
// #include "ofc/DList.h"
// #include "ofc/DInt.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DList         *list = [DList new];
//   DListIterator *iter;
//   DInt          *nr;
//   DText         *str;
//   DInt          *nr7;
//   id             obj;
//                                        // Add strings and numbers to the list
//   nr = [DInt alloc]; [nr init :4];
//   [list append :nr];                   // Append objects in list
// 
//   nr = [DInt alloc]; [nr init :8];
//   [list append :nr];
// 
//   str = [DText alloc]; [str init :"hello"];
//   [list append :str];
// 
//   nr7 = [DInt alloc]; [nr7 init :7];
//   [list append :nr7];
// 
//   str = [ DText alloc]; [str init :"bye"];
//   [list append :str];
// 
//   printf("List has %ld elements.\n", [list length]); // Get the lenght of the list
// 
//   nr = [DInt alloc]; [nr init :3];
//   [list insert :3 :nr];                // Insert an object in the list
// 
//   str = [DText alloc]; [str init :"replace"];
//   nr = [list set :1 :str];             // Index 2 is replaced by "replace"
//   [nr free];
// 
//   str = [list get :2];                 // Get an object from the list
//   printf("Element 2 is \"%s\".\n", [str cstring]);
// 
//   printf("List contains object nr7 %ld times.\n", [list count :nr7]); // Count occurences
// 
//   printf("Index of object nr7 is %ld.\n", [list index :nr7]); // Find index of object
// 
//   if ([list remove :nr7])              // Remove an object from the list
//   {
//     printf("Object nr7 is removed from the list.\n");
//     [nr7 free];
//   }
//   else
//     printf("Object nr7 could not be found in the list.\n");
// 
//   str = [list join :','];              // Join all objects text to one text object
//   printf("Join of all objects text in the list:%s.\n", [str cstring]);
//   [str free];
// 
//                                        // Iterate the list
//   iter = [DListIterator new];
// 
//   [iter list :list];                   // Set the list in the iterator
// 
//   obj = [iter last];                   // Iterate the list backwards
//   while (obj != nil)
//   {
//     if ([obj isKindOf :[DText class]]) // Find out the type of the object and ..
//     {
//       str = obj;
//       printf("Element: %s.\n", [str cstring]); // .. print the contents
//     }
//     else
//       if ([obj isKindOf :[DInt class]])
//       {
//         nr = obj;
//         printf("Element: %d\n", [nr get]);
//       }
//       else 
//         printf("Element: unknown.\n");
//     obj = [iter prev];
//   }
// 
//   [iter free];                         // Cleanup
//   [list free];
// 
//   return 0;
// }
//

@interface DListIterator : Object
{
@private
  DList          *_list;    // the list on which the iterator operates
  DListNode      *_node;    // the current node in the list
}


//// the constructors

//
// Initialise a list iterator without a list
//
// @return the iterator
//
- (DListIterator *) init;

//
// Initialise a list iterator with a list (the iterator is moved to the first node)
//
// @param list     the list on which the iterator operates
//
// @return the iterator
//
- (DListIterator *) init :(DList *) list;


//// the iterator methods

//
// Set the list (iterator is moved to the first object)
//
// @return the iterator
//
- (DListIterator *) list :(DList *) list;

//
// Get the first object in the list
//
// @return the first object (or nil)
//
- (id) first;

//
// Get the next object in the list
//
// @return the next object (or nil)
//
- (id) next;

//
// Get the previous object in the list
//
// @return the previous object (or nil)
//
- (id) prev;

//
// Get the last object in the list
//
// @return the last object (or nil)
//
- (id) last;

//
// Set the current object to another object
//
// @param object   the object to be placed in current node
//
// @return the previous stored object
//
- (id) object :(id) object;

//
// Get the current object
//
// @return the current object (or nil)
//
- (id) object;

//
// Insert a new object before the current object (iterator is moved to the new object)
// 
// @param obj      the object to be added before the current object
// 
// @return the iterator
// 
- (DListIterator *) before :(id) obj;

//
// Insert a new object after the current object (iterator is moved to the new object)
// 
// @param obj      the object to be added after the current object
// 
// @return the iterator
// 
- (DListIterator *) after :(id) obj;

@end
#endif
