#ifndef _DREGEX_H
#define _DREGEX_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DRegEx - the regular expression matcher in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-03 18:31:55 $ $Revision: 1.18 $
//
//==============================================================================

#include "ofc/config.h"

#if defined(HAVE_REGEX_H) && (defined(HAVE_RE_MATCH) || defined(HAVE_LIBREGEX))
#define HAVE_DREGEX (1)
#elif defined(HAVE_GNU_REGEX_H) && defined(HAVE_LIBGNUREGEX)
#define HAVE_DREGEX (1)
#endif


#ifdef HAVE_DREGEX

#if defined(HAVE_GNU_REGEX_H)
#include "gnu/regex.h"
#else
#include "regex.h"
#endif

#include "ofc/DArray.h"


#define DRE_NO_MATCH (-1)
#define DRE_ERROR    (-2)

//
// The RegEx class implements methods for using regular expressions. After
// compiling a regular expressing, it can be used to match and search strings.
// <pre>
//
//     This class uses the Extended POSIX syntax:
//      .   Match any char (incl. newline)   *         Match zero or more
//      +   Match one or more                ?         Match zero or one
//      {c} Match exactly c times            {min,max} Match min..max times
//      \|   Match alternatives               []        Match one in the list
//      [^] Match any except in list         [::]      Match a class in a list
//      ()  Group or subexpression
//      ^   Match begin of line              $         Match end of line
//
//     Classes: alnum, alpha, blank, cntrl, digit, graph, lower, print, punct, 
//              space, upper, xdigit
//
//     Examples: 
//           (aab\|zef)   matches aab or zef
//           aa(b\|z)ef   matches aabef or aazef
//           [a-c]       matches a or b or c
//           []de]       matches ] or d or e
//           [^ad]       match any except a and b
//           [[:alpha:]] matches letters (*NOTE*: use ccompile and not icompile)
//           [.*]        matches . and *
// </pre>
//
// @example
// #include <stdio.h>
// #include "ofc/DRegEx.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DRegEx *expr1 = [DRegEx alloc];
//   DRegEx *expr2 = [DRegEx new  ];
//   
//   int     matched;
//   int     index;
// 
//   [expr1 init :"a+b"];                 // Init with a regular expression
// 
//   matched = [expr1 match :"aaba"];     // Match a string with the regular expression
//   if (matched == DRE_NO_MATCH)
//     printf("String1 did not match the regular expression.\n");
//   else
//     printf("The first %d characters of string1 matched the regular expression.\n", matched);
// 
//   matched = [expr1 match :"baaba"];    // Match another string with the regular expression
//   if (matched == DRE_NO_MATCH)
//     printf("String2 did not match the regular expression.\n");
//   else
//     printf("The first %d characters of the string2 matched the regular expression.\n", matched);
// 
//   matched = [expr1 match :"baab" :1];  // Match a string with offset with the regular expression
//   if (matched == DRE_NO_MATCH)
//     printf("String3 did not match the regular expression.\n");
//   else
//     printf("The offseted %d characters of the string3 matched the regular expression.\n", matched);
// 
//   index = [expr1 search :"zyxab"];     // Search for the regular expression in the string
//   if (index == DRE_NO_MATCH)
//     printf("The regular expression could not be found in string4.\n");
//   else
//     printf("The regular expression started at %d in the string.\n", index);
// 
// 
//   [expr2 icompile :"a(b+)a"];          // Compile a regular expression, case insensitive
// 
//   matched = [expr2 match :"abbba"];    // Match the string with the regular expression
//   if (matched == DRE_NO_MATCH)
//     printf("String5 did not match the regular expression.\n");
//   else
//   {
//     DArray *matches = [expr2 matches :"abbba"]; // Get the matches from last match
// 
//     if (matches != nil)
//     {
//       for (index = 0; index < [matches length]; index++)
//       {
//         DText *str = [matches get :index];
// 
//         printf("Match %d:%s.\n", index, [str cstring]);
//       }
//       [matches free];
//     }
//   }
// 
//   [expr1 free];                        // Cleanup
//   [expr2 free];
// 
//   return 0;
// }
// 

@interface DRegEx : Object
{
@private
  regex_t             _pattern;  // the compiled pattern
  struct re_registers _regs;     // the match registers
  int                 _result;   // the result of the last match/search
  size_t              _length;   // the length of the last string
}

#endif
#ifdef HAVE_DREGEX

//// Constructors

//
// Initialise to an empty expression
//
// @return the object
//
- (DRegEx *) init;

//
// Initialise with a pattern
//
// @param pattern  the (case sensitive) c-string pattern
//
// @return the object
//
- (DRegEx *) init :(const char *) pattern;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
// @return the object
//
- shallowCopy;


//// Destructor

//
// Free the expression
//
// @return the object
//
- free;


//// Compile methods

//
// Compile a case sensitive pattern (POSIX Extended syntax)
// 
// @param pattern  the c-string pattern
//
// @return success
//
- (BOOL) ccompile :(const char *) pattern;

//
// Compile a case insensitive pattern (POSIX Extended syntax)
// 
// @param pattern  the c-string pattern
//
// @return success
//
- (BOOL) icompile :(const char *) pattern;


//// Matching and searching methods

//
// Match the string for the pattern
//
// @param str      the string to be matched
//
// @return the number of characters matched (or DRE_NO_MATCH, DRE_ERROR)
//
- (int) match :(const char *) str;

//
// Match the string with an offset for the pattern
// 
// @param str    the string to be matched
// @param from   the start in the string for the matching
// 
// @return the number of characters matched (or DRE_NO_MATCH, DRE_ERROR)
// 
- (int) match :(const char *) str :(int) from;

//
// Match the data string with an offset for the pattern
// 
// @param data     the data string to be matched
// @param length   the length of the data string
// @param from     the start in the data string for the matching
// 
// @return the number of characters matched (or DRE_NO_MATCH, DRE_ERROR)
// 
- (int) match :(const unsigned char *) data :(int) length :(int) from;

//
// Search in a string for the pattern
//
// @param str      the string to be searched
// 
// @return the offset in the string where the pattern matched (or DRE_NO_MATCH, DRE_ERROR)
// 
- (int) search :(const char *) str;

//
// Search in a range of a string for the pattern
// 
// @param str      the string to be searched
// @param from     the start of the range for searching
// @param to       the end of the range for searching
// 
// @return the offset in the string where the pattern matched (or DRE_NO_MATCH, DRE_ERROR)
// 
- (int) search :(const char *) str :(int) from :(int) to;

//
// Search in a data string for the pattern
//
// @param data     the data string to be searched
// @param length   the length of the data string
// 
// @return the offset in the data string where the pattern matched (or DRE_NO_MATCH, DRE_ERROR)
// 
- (int) search :(const unsigned char *) data :(int) length;

//
// Search in a range of a data string for the pattern
// 
// @param data     the data string to be searched
// @param length   the length of the data string
// @param from     the start of the range for searching
// @param to       the end of the range for searching
// 
// @return the offset in the data string where the pattern matched (or DRE_NO_MATCH, DRE_ERROR)
// 
- (int) search :(const unsigned char *) str :(int) length :(int) from :(int) to;


//// Result methods

//
// Return the subexpression indices from the last match/search. The indices
// are returned in a DArray with (new) DInt objects. The sequence in the array
// is: index 0 = start of whole matched string, index 1 = end of whole matched
// string, index 2 = start of first subexpression, index 3 = end of first 
// subexpression, and so on. Warning: if a group matches more than once as a
// result of a repetition operator, only the last match is stored in the indices.
// 
// @return a (new) array (or nil for no match)
// 
- (DArray *) indices;

//
// Return the matches from the last match/search. The matches are
// returned in a (new) DArray with (new) DText objects.
// 
// @param str      the string that was matched/searched
// 
// @return a (new) array (or nil for no match)
// 
- (DArray *) matches :(const char *) str;

//
// Return the matches from the last match/search. The matches
// are returned in a (new) DArray with (new) DData objects.
// 
// @param data     the data string that was matched/searched
// @param length   the length of the data string
// 
// @return a (new) array (or nil for no match)
// 
- (DArray *) matches :(const unsigned char *) data :(int) length;

@end
#endif
#endif
