#ifndef _DRND_H
#define _DRND_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DRnd - the pseudo random generator in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-02 05:58:14 $ $Revision: 1.5 $
//
//==============================================================================

#include <stdint.h>

#include "ofc/config.h"

#include "ofc/DRandomable.h"

//
// The DRnd class implements a pseudo random generator bases on a linair
// congruential generator. This is the most commonly used type, it is fast 
// but it has some well-known flaws. See @class(DMRnd) and @class(DLRnd) 
// for a better, but slower pseudo random generator.
//
// @example
// #include <stdio.h>
// #include "ofc/DRnd.h"
// 
// int main(int argc, char *argv[])
// {
//   DRnd *rnd1 = [DRnd alloc];
//   DRnd *rnd2 = [DRnd new  ];
// 
//   [rnd1 init :12345678];               // Init with a seed value
// 
//                                        // Get pseudo random values
//   printf("Pseudo random long value:%ld.\n", [rnd1 nextLong]);
//   printf("Pseudo random int value with range [1,100]:%d.\n", [rnd1 nextInt :1 :100]);
// 
//   [rnd2 seed :87654321];              // Set a seed value
// 
//                                       // Get pseudo random values
//   printf("Pseudo random double value:%f.\n", [rnd2 nextDouble]);
//   printf("Pseudo random double value with range [0.1,0.3]:%f.\n", [rnd2 nextDouble :0.1 :0.3]);
// 
//   [rnd1 free];                        // Cleanup
//   [rnd2 free];
// 
//   return 0;
// }
// 

@interface DRnd : Object <DRandomable>
{
@private
  uint32_t _seed;   // the seed for the generator
}


//// Constructors

//
// Initialise a non-seeded random generator
//
// @return the object
//
- (DRnd *) init;

//
// Initialise a seeded random generator
//
// @param seed     the seed to be used
// 
// @return the object
//
- (DRnd *) init :(unsigned long) seed;


//// Member methods

//
// Set the seed for the random generator
// 
// @param seed     the seed for the random generator
//
// @return the object
//
- (DRnd *) seed :(unsigned long) seed;


//// Random generator methods

//
// Get the next long from the generator
//  
// @return the next long
//
- (long) nextLong ;

//
// Get the next long inside a range from the generator
// 
// @param from     the start of the range
// @param to       the end of the range
// 
// @return the next long
// 
- (long) nextLong :(long) from :(long) to;

//
// Get the next integer from the generator
//  
// @return the next integer
//
- (int) nextInt;

//
// Get the next integer inside a range from the generator
// 
// @param from     the start of the range
// @param to       the end of the range
// 
// @return the next integer (inside the range)
// 
- (int) nextInt :(int) from :(int) to;

//
// Get the next double from the generator (0 <= v < 1)
//  
// @return the next double
//
- (double) nextDouble;

//
// Get the next double from the generator (from <= v < to)
// 
// @param from     the start of the range
// @param to       the end of the range
// 
// @return the next double
// 
- (double) nextDouble :(double) from :(double) to;

@end
#endif
