#ifndef _DRNDDIST_H
#define _DRNDDIST_H

// This file was automatically generated by objch.

//==============================================================================
//
//   DRndDist - the distributed pseudo random generator in the ofc-library
//
//               Copyright (C) 2007  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-03 18:31:55 $ $Revision: 1.6 $
//
//==============================================================================

#include <stdint.h>

#include "ofc/config.h"

#include "ofc/DRandomable.h"


//
// The DRndDist class implements pseudo random generators distributions.
//
// @example
// #include <stdio.h>
// #include "ofc/DRndDist.h"
// #include "ofc/DMRnd.h"
// 
// int main(int argc, char *argv[])
// {
//   DMRnd *rnd = [DMRnd new];
//   
//                                        // Uniform distributed random number
//   printf("Random number, uniform distributed, range 0.0..10.0    : %f.\n", 
//     [DRndDist nextUniform :rnd :0.0 :10.0]);
// 
//                                        // Normal distributed random number
//   printf("Random number, normal distributed,  range 0.0..10.0    : %f.\n",
//     [DRndDist nextNormal :rnd :0.0 :10.0]);
// 
//                                        // Exponential distributed random number
//   printf("Random number, exponential distributed, mean 5.0       : %f.\n",
//     [DRndDist nextExponential :rnd :5.0]);
// 
//                                        // Gamma distributed random number
//   printf("Random number, gamma distributed, alpha 2.0, beta 3.0  : %f.\n",
//     [DRndDist nextGamma :rnd :2.0 :3.0]);
// 
//                                        // Beta distributed random number
//   printf("Random number, beta distributed, alpha 2.0, beta 3.0   : %f.\n",
//     [DRndDist nextBeta :rnd :2.0 :3.0]);
// 
//                                        // Binomial distributed random number
//   printf("Random number, binomial distributed, prob .5 runs 10   : %u.\n",
//     [DRndDist nextBinomial :rnd :0.5 :10]);
// 
//                                        // Poisson distributed random number
//   printf("Random number, poisson distributed, mu 5.0             : %u.\n",
//     [DRndDist nextPoisson :rnd :5.0]);
// 
//                                        // Weibull distributed random number
//   printf("Random number, weibull distributed, alpha 2.0, beta 3.0: %f.\n",
//     [DRndDist nextWeibull :rnd :2.0 :3.0]);
// 
//   [rnd free];                          // Cleanup
// 
//   return 0;
// }
//

@interface DRndDist : Object
{
}


//// Random Generator Distribution class methods

//
// Generate a positive uniform random number <0,1>
// 
// @param rnd      the uniform random genetor
// 
// @param the positive uniform random number
// 
+ (double) nextPosDouble :(id <DRandomable>) rnd;

//
// Generate a random number with a uniform distribution in the range of [start,end>
// 
// @param rnd      the uniform random generator
// @param start    the start of the range
// @param end      the end of the range
//
// @return the random number
//
+ (double) nextUniform :(id <DRandomable>) rnd  :(double) start :(double) end;

//
// Generate a random number with a normal or gaussian distribution
// 
// @param rnd      the uniform random generator
// @param mu       the mean
// @param sigma    the standard deviation
//
// @return the random number
//
+ (double) nextNormal :(id <DRandomable>) rnd  :(double) mu :(double) sigma;

//
// Generate a random number with an exponential distribution
// 
// @param rnd      the random generator
// @param mu       the mean (>0)
//
// @return the random number
//
+ (double) nextExponential :(id <DRandomable>) rnd :(double) mu;

//
// Generate a random number with a gamma distribution
// 
// @param rnd      the uniform random generator
// @param alpha    the alpha parameter (>0)
// @param beta     the beta parameter (>0)
//
// @return the random number
//
+ (double) nextGamma :(id <DRandomable>) rnd :(double) alpha :(double) beta;

//
// Generate a random number with a beta distribution
// 
// @param rnd      the uniform random generator
// @param alpha    the alpha parameter (>0)
// @param beta     the beta parameter (>0)
//
// @return the random number
//
+ (double) nextBeta :(id <DRandomable>) rnd :(double) alpha :(double) beta;

//
// Generate a random number with a binomial distribution
// 
// @param rnd      the uniform random generator
// @param p        the probability [0,1]
// @param n        the number of trails (0>)
//
// @return the random number
//
+ (unsigned int) nextBinomial :(id <DRandomable>) rnd :(double) p :(unsigned int) n;

//
// Generate a random number with a poisson distribution
//  
// @param rnd      the uniform random generator
// @param mu       the mean (>=0)
//
// @return the random number
//
+ (unsigned int) nextPoisson :(id <DRandomable>) rnd :(double) mu;

//
// Generate a random number with a pareto distribution
//  
// @param rnd      the uniform random generator
// @param alpha    the scale parameter (>0)
// @param beta     the shape parameter (>0)
//
// @return the random number
//
+ (double) nextPareto :(id <DRandomable>) rnd :(double) alpha : (double) beta;

//
// Generate a random number with a weibull distribution
//  
// @param rnd      the uniform random generator
// @param alpha    the scale parameter (0>)
// @param beta     the shape parameter (0>)
//
// @return the random number
//
+ (double) nextWeibull :(id <DRandomable>) rnd :(double) alpha :(double) beta;

@end
#endif
