//==============================================================================
//
//      DSHA1 - the NIST's Secure Hash Algorithm class in the ofc-library
//
//               Copyright (C) 2002  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-07-28 05:16:31 $ $Revision: 1.7 $

//==============================================================================

#include "ofc/DSHA1.h"

#include <string.h>


#if _INTERFACE_

#include <stdint.h>

#include "ofc/config.h"

#include "ofc/DText.h"
#include "ofc/DData.h"

//
// The DSHA1 class implements a number of methods for using the NIST's SHA
// algorithm. The code is based on 'SHA-1 in C' by Steve Reid (steve[at]edmweb.com)
// which is in Public Domain.
//
// @example
// #include <stdio.h>
// #include "ofc/DSHA1.h"
// 
// int main(int argc, char *argv[])
// {
//   DSHA1 *sha1  = [DSHA1 alloc];
//   DSHA1 *sha2  = [DSHA1 new  ];
//   DText *str;
//   int    i;
// 
//   [sha1 init :"Test string for the SHA1 class"]; // Init with c-string
// 
//   str = [sha1 hexdigest];              // Convert to string and print
//   printf("SHA1 for the test string:%s.\n", [str cstring]);
//   [str free];
// 
//   for (i = 0; i < 10; i++)
//   {
//     [sha2 update :"abcdefghijklmnopqrstuvwxyz"]; // Repeated adding strings to sha
//   }
//   str = [sha2 hexdigest];              // Convert to string and print
//   printf("SHA1 for the repeated test string:%s.\n", [str cstring]);
//   [str free];
// 
//   [sha1 free];                         // Cleanup
//   [sha2 free];
// 
//   return 0;
// }
//

@interface DSHA1 : Object
{
@private
  uint32_t _state[5];
  uint32_t _count[2];
  unsigned char _buffer[64];
}

#endif



@implementation DSHA1

  
#if _PRIVATE_

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

// blk0() and blk() perform the initial expand.
// I (steve) got the idea of expanding during the round function from SSLeay
#ifndef WORDS_BIGENDIAN
#define blk0(i)     (block->l[i] = (rol(block->l[i],24)&0xFF00FF00) \
                  | (rol(block->l[i], 8) & 0x00FF00FF))
#else
#define blk0(i)      block->l[i]
#endif
#define blk(i)      (block->l[i&15] = rol(block->l[(i+13)&15]^ \
                    block->l[(i+8)&15]^block->l[(i+2)&15]^block->l[i&15],1))

// (R0+R1), R2, R3, R4 are the different operations used in SHA1
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);


// Hash a single 512-bit block. This is the core of the algorithm.

static void SHA1Transform(uint32_t state[5], const unsigned char buffer[64])
{
  uint32_t a, b, c, d, e;
  typedef union 
  {
    unsigned char c[64];
    uint32_t l[16];
  } 
  CHAR64LONG16;
  
  CHAR64LONG16* block;
  static unsigned char workspace[64];
  
  block = (CHAR64LONG16*)workspace;
  memcpy(block, buffer, 64);

   // Copy context->state[] to working vars
   a = state[0];
   b = state[1];
   c = state[2];
   d = state[3];
   e = state[4];
  
   // 4 rounds of 20 operations each. Loop unrolled.
   R0(a,b,c,d,e, 0); R0(e,a,b,c,d, 1); R0(d,e,a,b,c, 2); R0(c,d,e,a,b, 3);
   R0(b,c,d,e,a, 4); R0(a,b,c,d,e, 5); R0(e,a,b,c,d, 6); R0(d,e,a,b,c, 7);
   R0(c,d,e,a,b, 8); R0(b,c,d,e,a, 9); R0(a,b,c,d,e,10); R0(e,a,b,c,d,11);
   R0(d,e,a,b,c,12); R0(c,d,e,a,b,13); R0(b,c,d,e,a,14); R0(a,b,c,d,e,15);
   R1(e,a,b,c,d,16); R1(d,e,a,b,c,17); R1(c,d,e,a,b,18); R1(b,c,d,e,a,19);
   R2(a,b,c,d,e,20); R2(e,a,b,c,d,21); R2(d,e,a,b,c,22); R2(c,d,e,a,b,23);
   R2(b,c,d,e,a,24); R2(a,b,c,d,e,25); R2(e,a,b,c,d,26); R2(d,e,a,b,c,27);
   R2(c,d,e,a,b,28); R2(b,c,d,e,a,29); R2(a,b,c,d,e,30); R2(e,a,b,c,d,31);
   R2(d,e,a,b,c,32); R2(c,d,e,a,b,33); R2(b,c,d,e,a,34); R2(a,b,c,d,e,35);
   R2(e,a,b,c,d,36); R2(d,e,a,b,c,37); R2(c,d,e,a,b,38); R2(b,c,d,e,a,39);
   R3(a,b,c,d,e,40); R3(e,a,b,c,d,41); R3(d,e,a,b,c,42); R3(c,d,e,a,b,43);
   R3(b,c,d,e,a,44); R3(a,b,c,d,e,45); R3(e,a,b,c,d,46); R3(d,e,a,b,c,47);
   R3(c,d,e,a,b,48); R3(b,c,d,e,a,49); R3(a,b,c,d,e,50); R3(e,a,b,c,d,51);
   R3(d,e,a,b,c,52); R3(c,d,e,a,b,53); R3(b,c,d,e,a,54); R3(a,b,c,d,e,55);
   R3(e,a,b,c,d,56); R3(d,e,a,b,c,57); R3(c,d,e,a,b,58); R3(b,c,d,e,a,59);
   R4(a,b,c,d,e,60); R4(e,a,b,c,d,61); R4(d,e,a,b,c,62); R4(c,d,e,a,b,63);
   R4(b,c,d,e,a,64); R4(a,b,c,d,e,65); R4(e,a,b,c,d,66); R4(d,e,a,b,c,67);
   R4(c,d,e,a,b,68); R4(b,c,d,e,a,69); R4(a,b,c,d,e,70); R4(e,a,b,c,d,71);
   R4(d,e,a,b,c,72); R4(c,d,e,a,b,73); R4(b,c,d,e,a,74); R4(a,b,c,d,e,75);
   R4(e,a,b,c,d,76); R4(d,e,a,b,c,77); R4(c,d,e,a,b,78); R4(b,c,d,e,a,79);

   // Add the working vars back into context.state[]
   state[0] += a;
   state[1] += b;
   state[2] += c;
   state[3] += d;
   state[4] += e;
  
   // Wipe variables
   a = b = c = d = e = 0;
}

#endif


//// Constructors

//
// Initialise the sha object
//
// @return the object
//

- (DSHA1 *) init
{
  [super init];

  _state[0] = 0x67452301;
  _state[1] = 0xEFCDAB89;
  _state[2] = 0x98BADCFE;
  _state[3] = 0x10325476;
  _state[4] = 0xC3D2E1F0;
  _count[0] = 0;
  _count[1] = 0;
  
  return self;
}

//
// Initialise the sha object with a string
//
// @param cstring  the c-string
//
// @return the object
//

- (DSHA1 *) init :(const char *) cstring
{
  [self init];

  [self update :cstring];

  return self;
}

//
// Initialise the sha object with data
//
// @param data     the data ('\0' allowed)
// @param length   the length of the data
//
// @return the object
//

- (DSHA1 *) init :(const unsigned char *) data :(unsigned long) length
{
  [self init];

  [self update :data :length];

  return self;
}


//// Update methods

//
// Update the object with data. Repeated calls can be compared
// to a single call in which the strings are concatted
//
// @param data     the data to be feeded to sha
// @param length   the length of the data
//
// @return the object
//

- (DSHA1 *) update :(const unsigned char *) data :(unsigned long) length
{
  if (length > 0)
  {
    unsigned int i, j;

    j = (_count[0] >> 3) & 63;
    
    if ((_count[0] += length << 3) < (length << 3)) 
      _count[1]++;
    
    _count[1] += (length >> 29);
    if ((j + length) > 63) 
    {
      memcpy(&_buffer[j], data, (i = 64-j));
      SHA1Transform(_state, _buffer);
      for ( ; i + 63 < length; i += 64) 
      {
        SHA1Transform(_state, &data[i]);
      }
      j = 0;
    }
    else 
      i = 0;
    
    memcpy(&_buffer[j], &data[i], length - i);
  }

  return self;
}

//
// Update the object with a c-string. Repeated calls can be compared
// to a single call in which the strings are concatted
//
// @param cstring  the c-string to be feeded to sha1
//
// @return the object
//

- (DSHA1 *) update :(const char *) cstring
{
  int length = (cstring != NULL) ? strlen(cstring) : 0;
  
  if (length > 0)
  {
    [self update :(unsigned char *) cstring :length];
  }

  return self;
}

//// Digest methods

//
// Return the (current) digest of all data passed to the object. 
//
// @return the (new) data string object with the digest
//

- (DData *) digest
{
  unsigned long  i;
  unsigned char  finalcount[8];
  unsigned char  digest[20];
  DData         *data = [DData alloc];
  
  uint32_t  state[5]; // copy of state
  uint32_t  count[2];
  unsigned char  buffer[64];

  // save the current state
  memcpy(state,  _state,  sizeof(_state));
  memcpy(count,  _count,  sizeof(_count));
  memcpy(buffer, _buffer, sizeof(_buffer));
  
  for (i = 0; i < 8; i++) 
  {
    finalcount[i] = (unsigned char)((_count[(i >= 4 ? 0 : 1)]
                    >> ((3-(i & 3)) * 8) ) & 255);  // Endian independent
  }

  [self update :(unsigned char *)"\200" :1];
  
  while ((_count[0] & 504) != 448) 
  {
    [self update :(unsigned char *)"\0"  :1];
  }
  [self update :finalcount :8];  // Should cause a SHA1Transform()
  
  for (i = 0; i < 20; i++) 
  {
    digest[i] = (unsigned char) ((_state[i>>2] >> ((3-(i & 3)) * 8) ) & 255);
  }

  // restore the state
  memcpy(_buffer, buffer, sizeof(_buffer));
  memcpy(_state,  state,  sizeof(_state));
  memcpy(_count,  count,  sizeof(_count));
  
  return [data init :digest :20];
}

//
// Returns the (current) digest of all data passed to the object. The
// returning string contains the digest in hexadecimal ascii characters.
//
// @return the (new) string object with the digest
//

- (DText *) hexdigest
{
  DData *data   = [self digest     ];
  DText *string = [data tohexString];
  
  [data free];
  
  return string;
}

@end

/*===========================================================================*/

