#ifndef _DSHA256_H
#define _DSHA256_H

// This file was automatically generated by objch.

//==============================================================================
//
//   DSHA256 - the NIST's Secure Hash Algorithm 256 class in the ofc-library
//
//               Copyright (C) 2007  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-02 05:58:14 $ $Revision: 1.4 $
//==============================================================================

#include <stdint.h>

#include "ofc/config.h"

#include "ofc/DText.h"
#include "ofc/DData.h"

//
// The DSHA256 class implements a number of methods for using the NIST's SHA256
// algorithm. 
//
// @example
// #include <stdio.h>
// #include "ofc/DSHA256.h"
// 
// int main(int argc, char *argv[])
// {
//   DSHA256 *sha1 = [DSHA256 alloc];
//   DSHA256 *sha2 = [DSHA256 new  ];
//   DText   *str;
//   DData   *data;
// 
//   [sha1 init :"This is a test string for SHA256"]; // Init with a test string
// 
//   str = [sha1 hexdigest];              // Convert sha256 value to hexstring
//   printf("SHA256 of \"This is a test string for SHA256\" in hex-string is:%s.\n", [str cstring]);
//   [str free];
// 
// 
//   [sha2 update :"abcdefghijklmnopqrstuvwxyz"]; // Feed the sha2 variable with a c-string
//   [sha2 update :"0123456789" :10];     // Feed the sha2 variable with data string
// 
//   data = [sha2 digest];
//   str  = [data toBase64];              // Get SHA256 value and convert to base 64
//   printf("SHA256 of test string in base64 is:%s.\n", [str cstring]);
//   [str  free];
//   [data free];
// 
//   [sha1 free];                         // Cleanup
//   [sha2 free];
// 
//   return 0;
// }
// 

@interface DSHA256 : Object
{
@private
  uint32_t      _h[8];
  uint32_t      _count[2];
  unsigned char _input[64];
}


//// Constructors

//
// Initialise the sha256 object
//
// @return the object
//
- (DSHA256 *) init;

//
// Initialise the sha256 object with a string
//
// @param cstring  the c-string
//
// @return the object
//
- (DSHA256 *) init :(const char *) cstring;

//
// Initialise the sha256 object with data
//
// @param data     the data ('\0' allowed)
// @param length   the length of the data
//
// @return the object
//
- (DSHA256 *) init :(const unsigned char *) data :(unsigned long) length;


//// Reset method

//
// Reset the sha256 object 
//
// @return the object
//
- (DSHA256 *) reset;


//// Update methods

//
// Update the sha256 object with data. Repeated calls can be compared
// to a single call in which the strings are concatenated
//
// @param data     the data to be feeded to sha256
// @param length   the length of the data
//
// @return the object
//
- (DSHA256 *) update :(const unsigned char *) data :(unsigned long) length;

//
// Update the object with a c-string. Repeated calls can be compared
// to a single call in which the strings are concatenated
//
// @param cstring  the c-string to be feeded to sha256
//
// @return the object
//
- (DSHA256 *) update :(const char *) cstring;


//// Digest methods

//
// Return the (current) digest of all data passed to the object. 
//
// @return the (new) data string object with the digest
//
- (DData *) digest;

//
// Returns the (current) digest of all data passed to the object. The
// returning string contains the digest in hexadecimal ascii characters.
//
// @return the (new) string object with the digest
//
- (DText *) hexdigest;

@end
#endif
