#ifndef _DSOCKET_H
#define _DSOCKET_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DSocket - the socket wrapper classes in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-13 17:14:32 $ $Revision: 1.21 $
//==============================================================================
#ifdef WIN32
#endif
#ifndef O_NONBLOCK
#endif
#ifdef WIN32
#endif

#include "ofc/config.h"

#include "ofc/DData.h"
#include "ofc/DSocketAddress.h"

#ifdef WIN32
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#endif



//// Socket types

#define DSK_STREAM     (SOCK_STREAM)     // Stream socket 
#define DSK_DGRAM      (SOCK_DGRAM)      // Datagram socket


//// Send/Receive flags

#ifdef  MSG_OOB
#define DSK_MSG_OOB    (MSG_OOB)         // Out of band message
#endif
#ifdef  MSG_DONTROUTE
#define DSK_MSG_DONTROUTE (MSG_DONTROUTE) // Do not route the message
#endif
#ifdef  MSG_PEEK
#define DSK_MSG_PEEK   (MSG_PEEK)        // Peek data
#endif
#ifdef  MSG_WAITALL
#define DSK_MSG_WAITALL (MSG_WAITALL)    // Wait for the full message
#endif


//// Shutdown flags

#ifdef WIN32
#define DSK_SHUT_RD    (SD_RECEIVE)      // Shutdown receive
#define DSK_SHUT_WR    (SD_SEND)         // Shutdown send
#define DSK_SHUT_RDWR  (SD_BOTH)         // Shutdown both send and receive
#else
#define DSK_SHUT_RD    (SHUT_RD)         // Shutdown receive
#define DSK_SHUT_WR    (SHUT_WR)         // Shutdown send
#define DSK_SHUT_RDWR  (SHUT_RDWR)       // Shutdown both send and receive
#endif


//
//
// The DSocket class implements a number of methods for using network sockets.
//
// @example
// #include <stdio.h>
// #include "ofc/DSocket.h"
// #include "ofc/DSocketAddress.h"
// #include "ofc/DData.h"
// 
// int main(int argc, char *argv[])
// {
//   DSocket            *client = [DSocket new];
//   DInetSocketAddress *addr   = [DInetSocketAddress new];
// 
//                                        // Setup a client
//   [addr loopback :7000];               // Setup an address to localhost, port 7000
// 
//   printf("Open the socket ..\n");
//   [client open :[addr family] :DSK_STREAM :[DSocket protocol :"tcp"]]; // Open the socket for tcp
// 
//   [client reuseAddr :YES];             // Socketoption: Reuse the socket
// 
//   printf("Wait for connection with the server..\n");
//   if ([client connect :addr])          // Try to connect to the server
//   {
//     DData *received;
// 
//     printf("Send data to the server..\n");
//     if (![client send :"Hello there .." :14 :0]) // Send data to server
//       printf("Error sending data to server:%d\n", [client error]);
// 
//     printf("Receive data from the server..\n");
//     received = [client recv :256 :0];            // Receive data from server
// 
//     if (received != nil)
//     {
//       printf("Length of received data:%ld\n", [received length]);
// 
//       [received free];
//     }
//     else
//       printf("Error receiving data from server:%d\n", [client error]);
// 
//     printf("Close the connection..\n");
//     [client shutdown :DSK_SHUT_RDWR];  // Shutdown connection
// 
//     [client close];                    // Close socket
//   }
//   else
//     printf("No connection to the server:%d\n", [client error]);
// 
//   [client free];                       // Cleanup
//   [addr   free];
// 
//   return 0;
// }
// 

@interface DSocket : Object
{
@private
  int                     _socket;     // the socket
  id <DSocketAddressable> _address;    // the address (reference)
  int                     _family;     // the socket family (DSA_AF_INET,DSA_AF_UNIX)
  int                     _type;       // the socket type
  int                     _protocol;   // the socket protocol 
  BOOL                    _blocking;   // is the socket blocking ?
  int                     _errno;      // the last error
}


//// Constructors

//
// Initialise the socket
//
// @return the object
//
- (DSocket *) init;

#ifdef WIN32
#endif
//
// Initialise the socket with an already opened socket file descriptor
//
// @param fileno    the opened socket file descriptor
// @param address   the address for the socket file descriptor
// @param type      the socket type (DSK_STREAM, DSK_DGRAM)
// @param protocol  the protocol
//
// @return the object
//
- (DSocket *) init :(int) fileno :(id <DSocketAddressable>) address :(int) type :(int) protocol;

//
// Initialise the socket with a family, type and protocol
// 
// @param family    the socket family (DSA_AF_INET, DSA_AF_UNIX)
// @param type      the socket type (DSK_STREAM, DSK_DGRAM)
// @param protocol  the protocol
//
// @return the object
//
- (DSocket *) init :(int) family :(int) type :(int) protocol;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the socket object
//
// @return the object
//
- free;

#ifdef WIN32
#endif

//// Class methods

//
// Determine the protocol for a protocol name
// 
// @param name      the name of the protocol
// 
// @return the protocol number (or -1 if not found)
// 
+ (int) protocol :(const char *) name;


//// Member methods

//
// Get the sockets file descriptor
// 
// @return the file descriptor (or -1)
// 
- (int) fileno;

//
// Get the last error
// 
// @return the last error number
// 
- (int) error;

//
// Get the socket family
// 
// @return the family (DSA_AF_INET, DSA_AF_UNIX, DSA_AF_INET6)
- (int) family;

//
// Get the socket type
// 
// @return the type (DSK_STREAM, DSK_DGRAM)
//
- (int) type;


//// Socket option methods

//
// Set a socket option
//
// @param level     the level for the socket option (e.g. SOL_SOCKET)
// @param optname   the name for the socket option
// @param optval    the pointer to the option value
// @param optsize   the size of the option value
// 
// @return success
//
- (BOOL) setSocketOption :(int) level :(int) optname :(void *) optval :(int) optsize;

//
// Get a socket option
//
// @param level     the level for the socket option (e.g. SOL_SOCKET)
// @param optname   the name for the socket option
// @param optval    the pointer to the option value
// @param optsize   the size of the option value
// 
// @return success
// 
- (BOOL) getSocketOption :(int) level :(int) optname :(void *) optval :(int) optsize;

//
// Set the socket blocking
// 
// @param block     should the socket be blocking ?
// 
// @return success
// 
- (BOOL) blocking :(BOOL) block;

#ifdef WIN32
#endif
//
// Get the socket blocking state
// 
// @return is the socket blocking ?
// 
- (BOOL) blocking;

// 
// Set the reuse address socket option
//
// @param reuse    can the address be reused ?
//
// @return success
//
- (BOOL) reuseAddr :(BOOL) reuse;

//
// Get the reuse address socket option
//
// @return can the address be reused ?
//
- (BOOL) reuseAddr;

//
// Set the send buffer size socket option
//
// @param size      the size for the send buffer
//
// @return success
//
- (BOOL) sendBufferSize :(int) size;

//
// Get the send buffer size socket option
//
// @return the size
//
- (int) sendBufferSize;

//
// Set the receive buffer size socket option
//
// @param size      the size for the receive buffer
//
// @return success
//
- (BOOL) receiveBufferSize :(int) size;

//
// Get the receive buffer size socket option
//
// @return the size
//
- (int) receiveBufferSize;

//
// Set the keep alive socket option
//
// @param keep      should the connnection keep alive ?
//
// @return success
//
- (BOOL) keepAlive :(BOOL) keep;

//
// Get the keep alive socket option
//
// @return the keep alive setting
//
- (BOOL) keepAlive;

//
// Set the linger time out value
// 
// @param secs      the number of seconds for linger (or 0 for off)
// 
// @return success
//
- (BOOL) linger :(unsigned) secs;

//
// Get the linger time out value
// 
// @return the linger time out value (or 0 for off)
// 
- (unsigned) linger;


//// Connection methods

//
// Open a socket
// 
// @param family    the socket family (DSA_AF_INET, DSA_AF_UNIX)
// @param type      the socket type (DSK_STREAM, DSK_DGRAM)
// @param protocol  the protocol (see [protocol :name])
// 
// @return success
// 
- (BOOL) open :(int) family :(int) type :(int) protocol;

#ifdef WIN32
#endif
#ifdef USE_NOSIGPIPE
#endif
//
// Bind the socket to the address (inet:server only, unix:both)
// 
// @param address      the address to bind to
// 
// @return success
// 
- (BOOL) bind :(id <DSocketAddressable>) address;

//
// Listen for incoming connnections by a server
// 
// @param backlog   the number of pending connection before refuse
// 
// @return success
// 
- (BOOL) listen :(int) backlog;

//
// Accept a connection (server only)
// 
// @return the (new) socket object for the connection (or nil for error)
// 
- (DSocket *) accept;

#ifdef WIN32
#endif
//
// Connect a client to a server
// 
// @param address   the address of the server
// 
// @return success
// 
- (BOOL) connect :(id <DSocketAddressable>) address;

//
// Close a socket
// 
// @return success
// 
- (BOOL) close;

#ifdef WIN32
#endif
//
// Shutdown the connection
// 
// @param what      what part of the connection must be shutdown (DSK_SHUT_RD,DSK_SHUT_WR,DSK_SHUT_RDWR)
// 
// @return success
// 
- (BOOL) shutdown :(int) what;


//// Data transfer methods

//
// Send data (UDP)
// 
// @param address   the address to send to
// @param data      the data to be sent
// @param length    the length of the data
// @param flags     the message flags (DSK_MSG.. or 0)
// 
// @return the number of bytes sent (or -1 for error)
// 
- (int) sendto :(id <DSocketAddressable>) address :(void *) data :(int) length :(int) flags;

#ifndef WIN32
#endif
//
// Receive data (UDP)
// 
// @param address   the destination for the peer address
// @param length    the length of the data to be received
// @param flags     the receive flags (DSK_MSG.. or 0)
// 
// @return a new DData object (or nil)
// 
- (DData *) recvfrom :(id <DSocketAddressable>) address :(int) length :(int) flags;

#ifndef WIN32
#endif
//
// Send data (TCP)
// 
// @param data      the data to be sent
// @param length    the length of the data
// @param flags     the message flags (DSK_MSG.. or 0)
// 
// @return the number of bytes sent (or -1 for error)
// 
- (int) send :(void *) data :(int) length :(int) flags;

#ifndef WIN32
#endif
//
// Receive data (TCP)
// 
// @param length    the length of the data to be received
// @param flags     the receive flags (DSK_MSG.. or 0)
// 
// @return a new DData object (or nil)
// 
- (DData *) recv :(int) length :(int) flags;

#ifndef WIN32
#endif

//// Text transfer methods

//
// Send text (UDP)
// 
// @param address   the address to send to
// @param cstring   the cstring to be sent
// @param flags     the message flags (DSK_MSG.. or 0)
// 
// @return the number of bytes sent (or -1 for error)
// 
- (int) sendto :(id <DSocketAddressable>) address :(const char *) cstring :(int) flags;

#ifndef WIN32
#endif
//
// Receive text (UDP)
// 
// @param dest      the destination text object
// @param address   the destination for the peer address
// @param length    the length of the data to be received
// @param flags     the receive flags (DSK_MSG.. or 0)
// 
// @return the number of bytes received (or -1)
// 
- (int) recvfrom :(DText *) dest :(id <DSocketAddressable>) address :(int) length :(int) flags;

#ifndef WIN32
#endif
//
// Send text (TCP)
// 
// @param cstring   the c-string to be sent
// @param flags     the message flags (DSK_MSG.. or 0)
// 
// @return the number of bytes sent (or -1 for error)
// 
- (int) send :(const char *) cstring :(int) flags;

#ifndef WIN32
#endif
//
// Receive text (TCP)
// 
// @param dest     the destination text object
// @param length   the length of the data to be received
// @param flags    the receive flags (DSK_MSG.. or 0)
// 
// @return the number of characters received (or -1 for error)
// 
- (int) recv :(DText *) dest :(int) length :(int) flags;

#ifndef WIN32
#endif
//
// Receive (ASCII) data (TCP) till an end of line
// 
// @param dest      the destination for the received data
// @param eofl      the end of line string
// @param length    the maximum length of the data to be received (or -1 for no maximum)
// @param flags     the receive flags (DSK_MSG.. or 0)
// 
// @return the number of bytes received (or -1 for error)
// 
- (long) recv :(DText *) dest :(const char *) eofl :(long) length :(int) flags;

#ifndef WIN32
#endif
@end
#endif
