#ifndef _DTCPCLIENT_H
#define _DTCPCLIENT_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DTCPClient - the TCP client class in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-11 16:59:46 $ $Revision: 1.10 $
//==============================================================================

#include "ofc/config.h"

#include "ofc/DSocketAddress.h"
#include "ofc/DSocket.h"



//
// The DTCPClient class implements a number of methods for implementing a TCP client.
//
// @example
// #include <stdio.h>
// #include "ofc/DTCPClient.h"
// #include "ofc/DSocketAddress.h"
// 
// int main(int argc, char *argv[])
// {
//   DInetSocketAddress *addr = [DInetSocketAddress new];
//   DTCPClient         *clnt = [DTCPClient         new];
//   DData              *data;
// 
//   [addr host :"localhost" :7000];
// 
//   printf("Open the client..\n");
// 
//   if ([clnt open :[addr family] :[DSocket protocol :"tcp"]])
//   {
//     printf("Open connection to the server..\n");
// 
//     if ([clnt start :addr])
//     {
//       printf("Send request to the server..\n");
// 
//       data = [clnt doRequest :"Hello" :5 :256];
// 
//       if (data != nil)
//       {
//         DText *str = [data toText];
// 
//         printf("Response from the server:%s.\n", [str cstring]);
// 
//         [str  free];
//         [data free];
//       }
//       else
//         printf("No data received from the server.\n");
//       
//       [clnt stop];
//     }
//     else
//       printf("Cound not open connection to the server.\n");
//   }
//   else
//     printf("Could not open the client.\n");
// 
//   [clnt free];                         // Cleanup
//   [addr free];
// 
//   return 0;
// }
// 

@interface DTCPClient : Object
{
@private
  DSocket          *_local;        // the local socket
  int               _sendFlag;     // the send flags (def. 0)
  int               _recvFlag;     // the receive flags (def. 0)
  BOOL              _connected;    // is there a connection to the server ?
}


//// Constructors

//
// Initialise the TCP Client
//
// @return the object
//
- (DTCPClient *) init;

//
// Initialise the TCP Client for a family and protocol
//
// @param family       the family (DSA_AF_INET, DSA_AF_UNIX, DSA_AF_INET6)
// @param protocol     the protocol to be used
//
// @return the object
//
- (DTCPClient *) init :(int) family :(int) protocol;


//// Deconstructor

//
// Free the TCP client object
//
// @return the object
//
- free;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
// @return the object
//
- shallowCopy;


//// Member methods

//
// Return the local socket
// 
// @return a reference to the socket object
//
- (DSocket *) socket;

//
// Check if there is a connection to the server
// 
// @return is there a connection ?
// 
- (BOOL) isConnected;

//
// Set the send flag
// 
// @param flag         the send flag
// 
// @return the object
// 
- (DTCPClient *) sendFlag :(int) flag;

//
// Return the send flag
// 
// @return the send flag
// 
- (int) sendFlag;

//
// Set the receive flag
// 
// @param flag         the receive flag
// 
// @return the object
// 
- (DTCPClient *) recvFlag :(int) flag;

//
// Return the receive flag
// 
// @return the receive flag
// 
- (int) recvFlag;


//// Main methods

//
// Open the TCP client
// 
// @param family       the family (DSA_AF_INET, DSA_AF_UNIX, DSA_AF_INET6)
// @param protocol     the protocol to be used
// 
// @return success
// 
- (BOOL) open :(int) family :(int) protocol;

//
// Start the communication with the server (blocking till connection)
// 
// @param address      the address of the server
// 
// @return success
// 
- (BOOL) start :(id <DSocketAddressable>) address;

#ifdef DSA_AF_UNIX
#endif
//
// Send a request to the server and wait for the response (using send-
// and recvFlag)
// 
// @param request      the request to be sent
// @param length       the length of the request
// @param respLength   the (maximum) length of the response
//
// @return a (new) data object with the response (or nil for error)
//
- (DData *) doRequest :(const unsigned char *) request :(unsigned) length :(unsigned) respLength;

//
// Stop the communication with the server
// 
// @return none
// 
- (void) stop;

@end
#endif
