#ifndef _DTCPSERVER_H
#define _DTCPSERVER_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DTCPServer - the tcp server class in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-11 16:59:46 $ $Revision: 1.14 $
//==============================================================================

#include "ofc/config.h"

#include "ofc/DSocketAddress.h"
#include "ofc/DSocket.h"
#include "ofc/DData.h"

#define DTS_BLOCKING  (0)
#define DTS_THREADING (1)
#define DTS_FORKING   (2)

//
// The DTCPServer class implements a number of methods for implementing a TCP server.
// (Note: on WIN32 DTS_FORKING is not possible; when DTS_FORKING is used, it is switched
// to DTS_THREADING)(Note: this class assumes that the socket is blocking.)(Note: on 
// FreeBSD DTS_THREADING results in errors; when DTS_THREADING is used, it is 
// switched to DTS_FORKING)
//
// @example
// #include <stdio.h>
// #include "ofc/DTCPServer.h"
// #include "ofc/DSocketAddress.h"
// 
// @interface MyServer : DTCPServer
// {
// }
// 
// - (BOOL) handleRequest :(DData *) request :(DData *) response;
// 
// @end
// 
// @implementation MyServer
// 
// - (BOOL) handleRequest :(DData *) request :(DData *) response
// {
//   DText *str;
//  
//   str = [request toText];              // Print the response
//   printf("Server receives the request:%s\n", [str cstring]);
//   [str free];
// 
//   printf(".. and sends: Nice weather.\n");
// 
//   [response set :"Nice weather" :12];  // Send the response
// 
//   return YES;                          // Example stop the server
// }
// 
// @end
// 
// int main(int argc, char *argv[])
// {
//   MyServer           *srvr = [MyServer       new];
//   DInetSocketAddress *addr = [DInetSocketAddress new];
// 
//   [addr any :7000];                    // Setup the address for the server
// 
//   printf("Open the server..\n");
// 
//   if ([srvr open :[addr family] :[DSocket protocol :"tcp"]])
//   {
//     printf("Start the server (and wait for client to connect)..\n");
// 
//     if (![srvr start :addr :2 :DTS_BLOCKING :1])
//       printf("Could not start the server ..\n");
//   }
//   else
//     printf("Could not open the server ..\n");
// 
//   [srvr free];                         // Cleanup
//   [addr free];
// 
//   return 0;
// }
// 

@interface DTCPServer : Object
{
@private
  DSocket          *_local; // the local socket
  int               _mode;  // the running mode (DTS_THREADING, DTS_FORKING, DTS_BLOCKING)
  int               _sendFlag; // the send flag (def. 0)
  int               _recvFlag; // the receive flag (def. 0)
  unsigned          _recvLength; // the length used during receiving of requests (def. 1024)
}


//// Constructors

//
// Initialise the Tcp Server
//
// @return the object
//
- (DTCPServer *) init;

//
// Initialise the Tcp Server for a family and protocol
//
// @param family       the family (DSA_AF_INET, DSA_AF_UNIX, DSA_AF_INET6)
// @param protocol     the protocol to be used
// 
// @return the object
//
- (DTCPServer *) init :(int) family :(int) protocol;


//// Deconstructor

//
// Free the tcp server object
//
// @return the object
//
- free;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
//
- shallowCopy;


//// Member methods

//
// Return the local socket for the TCP server
// 
// @return a reference to the local socket 
//
- (DSocket *) socket;

//
// Set the send flag
// 
// @param flag         the send flag
// 
// @return the object
// 
- (DTCPServer *) sendFlag :(int) flag;

//
// Return the send flag
// 
// @return the send flag
// 
- (int) sendFlag;

//
// Set the receive flag
// 
// @param flag         the receive flag
// 
// @return the object
// 
- (DTCPServer *) recvFlag :(int) flag;

//
// Return the receive flag
// 
// @return the receive flag
// 
- (int) recvFlag;

//
// Set the length for receiving of requests
// 
// @param length       the length
// 
// @return the object
// 
- (DTCPServer *) recvLength :(unsigned) length;

//
// Return the length for receiving of requests
// 
// @return the length
// 
- (unsigned) recvLength;


//// Main methods

//
// Open the TCP server
// 
// @param family       the family (DSA_AF_INET, DSA_AF_UNIX, DSA_AF_INET6)
// @param protocol     the protocol to be used
// 
// @return success
// 
- (BOOL) open :(int) family :(int) protocol;

//
// Wait for connections and call 'handleConnection' on connection. Note: DTS_FORKING does not work
// on WIN32; it is converted to DTS_THREADING.
// 
// @param address      the address for the server
// @param backlog      the number of pending connnections before refuse
// @param mode         the running mode (DTS_THREADING, DTS_FORKING, DTS_BLOCKING)
// @param connections  the number of connections to be made before returning (or -1 for unlimited)
// 
// @return success
// 
- (BOOL) start :(id <DSocketAddressable>) address :(int) backlog :(int) mode :(int) connections;

#ifdef WIN32
#endif
#ifdef __FreeBSD__
#endif
#ifndef WIN32                
#endif                
//
// Handle a new connection with a peer (called by start)
//
// @param peer         the connected socket
// 
// @return the object
//
- (void) handleConnection :(DSocket *) peer;

//
// Handle a new request by the client (called by handleConnection, to be overridden)
// 
// @param request      the request made by the client
// @param response     the destination for the response
// 
// @return should the connection be stopped ?
//
- (BOOL) handleRequest :(DData *) request :(DData *) response;

@end
#endif
