#ifndef _DTEXT_H
#define _DTEXT_H

// This file was automatically generated by objch.

//==============================================================================
//
//            DText - the text string class in the ofc-library
//
//             Copyright (C) 2003-2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-05 16:29:04 $ $Revision: 1.33 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DDatable.h"
#include "ofc/DTextable.h"
#include "ofc/DComparable.h"
#include "ofc/DSizeable.h"
#include "ofc/DTextReadable.h"
#include "ofc/DTextScannable.h"
#include "ofc/DTextWritable.h"
#include "ofc/DParsable.h"

#include "ofc/DData.h"

//
// The text string class implements methods for storing and manipulating text strings.
//
// @example
// #include <stdio.h>
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DText *str1 = [DText alloc];
//   DText *str2 = [DText new  ];
//   DText *str3 = [DText new  ];
//   DText *str4 = [DText new  ];
//   DText *str5 = [DText new  ];
// 
//                                        // String setters
//   [str1 init :"  Initial  "];          // Init with a c-string
// 
//   [str2 format :"Form%ctted", 'a'];    // Set the string with a format
// 
//   [str3 set :"hello, this is a nice day"]; // Set the string
// 
//   [str4 set :"Substring" :2 :4];       // Set with substring
// 
//   [str5 set :'-' :15];                 // Make a string with 15 dashes
// 
//   printf("String setters: %s %s %s %s %s.\n", 
//       [str1 cstring], [str2 cstring], [str3 cstring], [str4 cstring], [str5 cstring]);
// 
//                                        // Character methods
//   [str5 put :1 :'+'];                  // Put on index 1 a '+'
//   
//   printf("Last char in string1:%c.\n", [str1 get :-1]);
// 
//   [str5 delete :0];                    // Remove the first character from string5
// 
//                                        // Slice methods
//   [str4 insert :1 :-1 :"string"];      // Replace st to string in string4
// 
//   [str5 insert :2 :4 :'+' :4];         // Replace [2,4] to ++++ in string5
// 
//   [str5 delete :6 :10];                // Remove a substring from string5
// 
//   printf ("String lengths: %ld %ld %ld %ld %ld.\n",
//       [str1 length], [str2 length], [str3 length], [str4 length], [str5 length]);
// 
//                                        // String manipulation
//   [str4 prepend :"Su"];                // Prepend to make Substring
// 
//   [str4 multiply :2];                  // Multiply substring twice
// 
//   [str3 capwords];                     // Capitalize all words in string3
// 
//   [str1 lower];                        // Lower all caps in string1
// 
//   [str1 strip];                        // Strip all leading and trailing spaces
// 
//   [str2 rjust :15];                    // Right justify string2
// 
//                                        // Results
//   printf("Results: %s %s %s %s %s.\n", 
//       [str1 cstring], [str2 cstring], [str3 cstring], [str4 cstring], [str5 cstring]);
// 
//                                        // String compare 
//   if ([str1 compare :str4] == 0)
//     printf("String1 and string4 are equal.\n");
//   else
//     if ([str1 compare :str4] < 0)
//       printf("String1 is smaller than string4.\n");
//     else
//       printf("String1 is bigger than string4.\n");
//   
//   return 0;
// }
//

@interface DText : Object <DDatable,DTextable,DComparable,DSizeable,DTextReadable,DTextScannable,DTextWritable,DParsable>
{
@private
  unsigned long  _size;     // the allocated size for the string
  unsigned       _extra;    // the extra amount for sizing
  unsigned long  _length;   // the length of the string
  unsigned long  _pntr;     // the pointer for the DTextReadable,-Writable protocols
  char          *_string;   // the string data
  int            _error;    // the last error (or 0)
}


//// Constructors

//
// Initialise an empty string
//
// @return the string object
//
- (DText *) init;

//
// Initialise the object with a c-string
//
// @param cstring  the zero terminated ascii string
//
// @return the string object
//
- (DText *) init :(const char *) cstring;


//// Copy related methods

//
// Do a shallow copy of the string object
//
// @return the string object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return nothing
//
- free;


//// DSizeable protocol methods

//
// Set the size (storage space) for the string
//
// @param size   the request size for the string
//
// @return the string object
//
- (DText *) size :(unsigned long) size;

//
// Set the amount of extra size increase if the size must be extended
//
// @param extra    the amount (def. 0)
//
// @return the object
//
- (DText *) extra :(unsigned) extra;


//// Conversion methods

//
// Convert the string to a floating point number
//
// @return the floating point number
//
- (double) toDouble;

//
// Convert the string to a integer number
//
// @return the integer number
//
- (int) toInt;

//
// Convert the string to a long number
//
// @return the long number
//
- (long) toLong;


//// String manipulation

//
// Format the string
//
// @param format   the format string
//
// @return the string object
//
- (DText *) format :(const char *) format,...;

//
// Clear the text string
// 
// @return the object
// 
- (DText *) clear;

//
// Set a c-string in the string
//
// @param cstring  the c-string
//
// @return the string object
//
- (DText *) set :(const char *) cstring;

//
// Set a substring of a c-string in the object
//
// @param cstring  the c-string
// @param from     the start in the c-string
// @param to       the end in the c-string
//
// @return the object
//
- (DText *) set :(const char *) cstring :(long) from :(long) to;

//
// Set number of characters in the string
//
// @param ch        the character to be set
// @param number    the number of characters
// 
// @return the object
// 
- (DText *) set :(char) ch :(long) number;

// Set a character at the given index (must fit in string)
//
// @param index    the index in the string
// @param ch       the character to be placed in the string
//
// @return the string object
//
- (DText *) put :(long) index :(char) ch;

//
// Get the character at the given index
//
// @param index    the index in the string
//
// @return the character in the string or \0
//
- (char) get :(long) index;

//
// Delete the character at the given index
//
// @param index    the index in the string
//
// @return the string object
//
- (DText *) delete :(long) index;

//
// Insert a cstring in a part of the string.
//
// @param from     the start index
// @param to       the end index (to >= (from-1))
// @param cstring  the c-string
//
// @return the string object
//
- (DText *) insert :(long) from :(long) to :(const char *) cstring;

//
// Insert a number of characters in a part of the string
//
// @param from     the start index
// @param to       the end index (to >= (from-1))
// @param ch       the character to be inserted
// @param number   the number of characters to insert
//
// @return the string object
//
- (DText *) insert :(long) from :(long) to :(char) ch :(long) number;

//
// Return a substring 
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return a new text string object
//
- (DText *) get :(long) from :(long) to;

//
// Delete a substring from the string
//
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the string object
//
- (DText *) delete :(long) from :(long) to;

//
// Split the string in two: the part before the separator character is
// stored in the returning new string object; the part after the separator
// remains in the string object.
//
// @param separator the character
//
// @return a new text object (or nil if character not found)
//
- (DText *) split :(char) separator;


//// String properties methods

//
// Check if the string is empty
//
// @return Is the string empty ?
//
- (BOOL) isEmpty;

//
// Return the length of the string
//
// @return the length of the string
//
- (unsigned long) length;

//
// Return the pointer to a c-string notation of the string
//
// @return the pointer to the c-string
//
- (const char *) cstring;

//
// Return the last occurred error
//
// @return the last error (or 0)
//
- (int) error;


//// DTextable protocol methods

//
// Convert to a text string object
//
// @return a (new) text string object
//
- (DText *) toText;


//// DDatable protocol methods

//
// Convert to a data string
//
// @return a (new) data string object
//
- (DData *) toData;


//// DTextReadable protocol methods

//
// Read text string (all available text)
//
// @return a (new) text string
//
- (DText *) readText;

//
// Read text string 
//
// @param length   the length of the text to be read
//
// @return a (new) text string
//
- (DText *) readText :(long) length;

//
// Read a line of text (till Eof or \n, not including in text string)
//
// @return a (new) text string
//
- (DText *) readLine;

//
// Read character
//
// @return the character read
//
- (char) readChar;

//
// Tell the current position in the text string
//
// @return the current position
//
- (unsigned long) tell;

//
// Move the current position in the text string
//
// @param offset   the offset from the origin (in bytes)
// @param origin   the origin for the offset (0=begin, 1=current, 2=end)
//
// @return success
//
- (BOOL) seek :(unsigned long) offset :(int) origin;

//
// Skip a number of positions
//
// @param offset   the number of bytes to skip
//
// @return success
//
- (BOOL) skip :(unsigned long) offset;

//
// Check if the end of the string is reached
//
// @return is it?
//
- (BOOL) isEof;


//// TextScannable protocol methods

//
// Scan for a text string until a separator character. The separator character is 
// is not part of the returned string and is skipped in the text object.
//
// @param separator  the separator character
//
// @return a (new) text string (or nil if the separator is not found)
//
- (DText *) scanText :(char) separator;

//
// Scan for a text string until one of the separator characters in the 
// separator string.  The matched separator character is is not part of 
// the returned string, is skipped in the text object and is returned in 
// the separator argument.
//
// @param separators  the c-string with the separators
// @param separator   the destination for the actual matched separator character (or NULL)
//
// @return a (new) text string (or nil if none of the separators is found)
//
- (DText *) scanText :(const char *) separators :(char *) separator;

//
// Match a string case sensitive; if the string matched, the current position in
// the readable is moved to the location after the matched string.
//
// @param cstring   the c-string to be matched
//
// @return does the string match ?
//
- (BOOL) cmatch :(const char *) cstring;

//
// Match a string case insensitive; if the string matched, the current position in
// the readable is moved to the location after the matched string.
//
// @param cstring   the c-string to be matched
//
// @return does the string match ?
//
- (BOOL) imatch :(const char *) cstring;

//
// Scan for a (decimal) integer value
//
// @param wrong   the value returned if there is no int to scan
//
// @return the integer read (or the wrong value)
//
- (int) scanInt :(int) wrong;

//
// Skip characters
//
// @param ch    the character to be skipped
//
// @return the number of characters skipped
//
- (int) skipChar :(char) ch;

//
// Skip whitespace
//
// @return the number of characters skipped
//
- (int) skipWhiteSpace;


//// TextWritable protocol methods

//
// Write text string 
//
// @param text     the text to be written
//
// @return success
//
- (BOOL) writeText :(const char *) text;;

//
// Write line (appending a '\n')
//
// @param text     the text to be written as line
//
// @return success
//
- (BOOL) writeLine :(const char *) text;

//
// Write character
//
// @param ch       the character to be written
//
// @return success
//
- (BOOL) writeChar :(char) ch;;


//// DParsable methods

//
// Parse a string for a DText string
// 
// @param cstr     the string to be parsed (moved to the first non-parsable char)
// 
// @return the result (0, ENODATA)
// 
- (int) fromString :(char **) cstr;


//// String manipulation methods

//
// Append a c-string to the string
//
// @param cstring  the string to be appended
//
// @return the string object
//
- (DText *) append :(const char *) cstring;

//
// Append a formatted string
//
// @param format   the format string
//
// @return the string object
//
- (DText *) appendFormat :(const char *) format,...;

//
// Prepend a c-string in front of the string
//
// @param cstring  the string to be prepended
//
// @return the string object
//
- (DText *) prepend :(const char *) cstring;

//
// Push one character at the end of the string
// 
// @param ch       the character
//
// @return the object
//
- (DText *) push :(char) ch;

//
// Pop the last character from the string
//
// @return the character  (or \0)
//
- (char) pop;

//    
// Multiply the string a number of times
//
// @param times    the number of times the string must be multiplied
//
// @return the string object
//
- (DText *) multiply :(unsigned) times;

//
// Capitalize the first character of the string
//
// @return the string object
//
- (DText *) capitalize;

//
// Capitalize the first character of every word, replaces repeated whitespace
// with one space and strips leading and trailing spaces.
//
// @return the string object
//
- (DText *) capwords;

//
// Expand all tabs with tabsize spaces
//
// @param tabsize  the size of the tab in spaces
//
// @return the string object
//
- (DText *) expandtabs :(unsigned) tabsize;

//
// Convert all uppercase characters in the string to lowercase
//
// @return the string object
//
- (DText *) lower;

//
// Convert all lowercase characters in the string to uppercase
//
// @return the string object
//
- (DText *) upper;

//
// Convert all lowercase to uppercase and uppercase to lowercase
//
// @return the string object
//
- (DText *) swapcase;

//
// Strip leading spaces from the string
//
// @return the string object
//
- (DText *) lstrip;

//
// Strip tailing spaces from the string
//
// @return the string object
//
- (DText *) rstrip;

//
// Strip both leading and tailing spaces from the string
//
// @return the string object
//
- (DText *) strip;

//
// Align the string to the left
//
// @param width    the length of the resulting string
//
// @return the string object
//
- (DText *) ljust :(unsigned long) width;

//
// Align the string to the right
//
// @param width    the length of the resulting string
//
// @return the string object
//
- (DText *) rjust :(unsigned long) width;

//
// Center the string
//
// @param width    the length of the resulting string
//
// @return the string object
//
- (DText *) center :(unsigned long) width;

//
// Pad a string to the left with zeros 
//
// @param width    the length of the resulting string
//
// @return the string object
//
- (DText *) zfill  :(unsigned long) width;

//
// Insert linefeeds in a string to create a column
// 
// @param width    the width of the column
// 
// @return the object
// 
- (DText *) column :(int) width;


//// Comparable protocol methods

//
// Compare the string with another string
//
// @param other    the string that should be compared
//
// @return the compare result (-1, 0, 1)
//
- (int) compare :(DText *) other;


//// Comparison methods

//
// Compare case sensitive the string with a c-string
//
// @param other    the c-string that should be tested
//
// @return the test result (-1, 0, 1)
//
- (int) ccompare :(const char *) other;

//
// Compare case insensitive the string with a c-string
//
// @param other    the c-string that should be compared
//
// @return the compare result (-1, 0, 1)
//
- (int) icompare :(const char *) other;

//
// Compare case sensitive the string with the first number of characters 
// of a c-string
//
// @param other    the c-string that should be tested
// @param length   the length of the string to be tested
//
// @return the test result (-1, 0, 1)
//
- (int) ccompare :(const char *) other :(unsigned long) length;

//
// Compare case insensitive the string with the first number of characters
// of a c-string
//
// @param other    the c-string that should be compared
// @param length   the length of the string to be tested
//
// @return the compare result (-1, 0, 1)
//
- (int) icompare :(const char *) other :(unsigned long) length;


//// Search and replace methods

//
// Count the number of occurrences of the c-string in the string
//
// @param search   the c-string that should be counted
// @param from     the start index 
// @param to       the end index (to >= from)
//
// @return the number of occurrences
//
- (unsigned long) count :(const char *) search :(long) from :(long) to;

//
// Return the first index where the search c-string is found
//
// @param search   the c-string that should be searched
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the first index or -1 = not found
//
- (long) index  :(const char *) search :(long) from :(long) to;

//
// Return the last index where the search c-string is found
//
// @param search   the c-string that should be searched
// @param from     the start index
// @param to       the end index (to >= from)
//
// @return the last index or -1 = not found
//
- (long) rindex :(const char *) search :(long) from :(long) to;

//
// Replace in the string the occurrences of old with new, max times
//
// @param old      the c-string that is searched
// @param new      the c-string that is inserted
// @param max      the maximum number of replaces (-1 = all)
//
// @return the string object
//
- (DText *) replace :(const char *) old :(const char *) new :(long) max;

@end
#endif
