#ifndef _DTEXTWRITABLELOGGER_H
#define _DTEXTWRITABLELOGGER_H

// This file was automatically generated by objch.

//==============================================================================
//
//        DTextWritableLogger - the TextWritable logger in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-07 17:38:49 $ $Revision: 1.5 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DLogger.h"
#include "ofc/DTextWritable.h"

#include "ofc/DText.h"
#include "ofc/DDateTime.h"

//
// The DTextWritableLogger class implements methods for a continuous logging
// to a class implementing  the @proto(DTextWritable). The logger changes
// identical sequential log messages to two lines: first the message and then a line
// with the number of occurences. The layout of the log line is: 
// yyyy-mm-dd hh:mm:ss.mmm <level> message.
// 
// @example
// #include <stdio.h>
// #include "ofc/DTextWritableLogger.h"
// #include "ofc/DLog.h"
// #include "ofc/DFile.h"
// 
// int main(int argc, char *argv[])
// {
//   DTextWritableLogger *logger = [DTextWritableLogger new];
//   DFile               *file   = [DFile new];
//   char                 name[] = "example.log";
//   int                  i;
// 
//   if ([file open :name :"w"])
//   {
//     [logger writer :file];             // Give to the textwriteable logger the destination file
// 
//     dlogger(logger);                   // Give to the base logger the textwritable logger as destination
// 
//     dlog(DLOG_ALERT, "Alert!!");       // Log an alert
// 
//     for (i = 0; i < 10; i++)           // Add a repeated error message
//     {
//       dlog(DLOG_ERROR, "Errors");
//     }
// 
//     dlogmask(DLOG_NOTICE|DLOG_DEBUG);  // Mask some log levels
//     
//     dlog(DLOG_WARNING, "Warning..masked"); // Masked warning
//     dlog(DLOG_DEBUG,   "Debug:%d",7);  // Debug message
//     dlog(DLOG_NOTICE,  "Notice.");     // Notice message
// 
//     dlogger(nil);                      // Stop logging via the textwritable logger
// 
//     [file close];                      // Close the destination file
// 
//     printf("Events succesfully logged in \"%s\".\n", name);
//   }
//   else
//     printf("Could not open \"%s\" for writing:%d.\n", name, [file error]);
// 
//   [logger free];                       // Cleanup
//   [file   free];
// 
//   return 0;
// }
// 

@interface DTextWritableLogger : Object <DLogger>
{
  unsigned            _mask;     // the active mask
  id<DTextWritable>   _writer;   // the reference to the writer
  DText              *_previous; // the previous message
  DDateTime          *_time;     // the time
  long                _repeated; // the number of identical messages
  DText              *_output;   // the output for the logger
}

//// Constructors

//
// Initialise an empty logger 
//
// @return the object
//
- (DTextWritableLogger *) init;

//
// Initialise a logger with a writter
// 
// @param writer     the writer for the logger
//
// @return the object
//
- (DTextWritableLogger *) init :(id <DTextWritable>) writer;


//// Copy constructor

//
// Do a shallow copy of the object
// 
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free;


//// Member methods

//
// Set the writer for the logger
// 
// @param writer    the writer for the logger
// 
// @return the object
// 
- (DTextWritableLogger *) writer :(id<DTextWritable>) writer;


//// Implementation of interface DLogger

//
// Log an event by the logger
// 
// @param level      the level of the event
// @param message    the message to be logged
// 
// @return success
// 
- (BOOL) doLog :(int) level :(const char *) message;

//
// Set the active levels 
// 
// @param level      the bitwise or'ed active levels 
// 
// @return the previous mask
// 
- (int) mask :(int) levels;

//
// Set the active level range
// 
// @param high        the highest active level
// @param low         the lowest active level
// 
// @return the previous mask
//
- (int) mask :(int) high :(int) low;

//
// Write a event in the writer
// 
// @param level       the level of the event   
// @param message     the message of the event
// 
// @return success
// 
- (BOOL) writeEvent :(int) level :(const char *) message;

//
// Format the output of the event (to be overridden for different format)
// 
// @param output      the output of the formatting of the event
// @param dateTime    the date/time of the event
// @param level       the level of the event   
// @param message     the message of the event
// 
// @return success
// 
- (BOOL) format :(DText *) output :(DDateTime *) dateTime :(int) level :(const char *) message;

@end
#endif
