#ifndef _DTOKENIZER_H
#define _DTOKENIZER_H

// This file was automatically generated by objch.

//==============================================================================
//
//           DTokenizer - the tokenizer class in the ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-06 17:23:51 $ $Revision: 1.5 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DSource.h"

#include "ofc/DText.h"
#include "ofc/DList.h"

#include "ofc/DTextReadable.h"


#define DTK_UNKNOWN     (0)
#define DTK_EOFF        (1)
#define DTK_EXTRA1      (2)
#define DTK_EXTRA2      (3)
#define DTK_EXTRA3      (4)
#define DTK_WHITESPACE  (5)
#define DTK_COMMENT     (6)
#define DTK_NUMBER      (7)
#define DTK_KEYWORD     (8)
#define DTK_STRING      (9)
#define DTK_OPERATOR    (10)

//
// The class DTokenizer implements a (simple) lexical scanner. The 
// tokenizer scans a text stream for whitespace, strings, keywords, 
// comments, operators, numbers and three to be implemented formats. The 
// scanner scans the text on a line basis.
// The tokenizer does not use regular expressions, but it uses for every token
// a method to scan for the token. The method can be overridden for a specific
// implementation of the token scanner.
// There are two methods: nextToken and checkToken. The first method checks
// for a token and if there is a match, the scanner location is moved to
// the next location in the source. The second method only returns 
// the result of the match, the client must call the service
// 'next' to move the scanner location.
// The token methods return one of the following: DTK_UNKNOWN, DTK_EOFF,
// DTK_EXTRA1, DTK_EXTRA2, DTK_EXTRA3, DTK_WHITESPACE, DTK_COMMENT, 
// DTK_NUMBER, DTK_KEYWORD, DTK_STRING, DTK_OPERATOR.
//
// @example
// #include <stdio.h>
// #include <ctype.h>
// #include "ofc/DTokenizer.h"
// #include "ofc/DFile.h"
// 
// @interface MyTokenizer : DTokenizer           // MyTokenizer overrides comment and extraToken1
// {
// @private
// }
// 
// 
// - (DText *) comment :(const char *) text;     // Override the comment checker
// - (DText *) extraToken1 :(const char *) text; // Override the extra token checker
// @end
// 
// @implementation MyTokenizer 
// 
// - (DText *) comment :(const char *) text      // Override the default comment checker ..
// {
//   DText *scanned = nil;
//   
//   if ((*text == '/') && (*(text+1) == '/'))   // .. comment is '//' 
//   {
//     scanned = [DText new];
//     
//     while (*text != EOS)
//     {
//       [scanned push :*text++];
//     }
//   }
//   return scanned;
// }
// 
// - (DText *) extraToken1 :(const char *) text  // Override the extra token checker for ..
// {
//   DText *scanned = nil;
// 
//   if (*text == '#')                           // .. scanning of directives: #...
//   {
//     scanned = [DText new];
// 
//     while ((!isspace(*text)) && (*text != EOS))
//     {
//       [scanned push :*text++];
//     }
//   }
//   return scanned;
// }
// 
// @end
// 
// 
// int main(int argc, char *argv[])
// {
//   MyTokenizer *srce = [MyTokenizer new];
//   DFile       *file = [DFile       new];
// 
//   if ([file open :"example.tok" :"r"]) // Open the file
//   { 
//     if ([srce source :file :"example.tok"]) // Give the file to the tokenizer
//     {
//       [srce skipWhiteSpace :YES];      // Skip whitespace
// 
//       while (![srce isEof])            // Do for the whole
//       {
//         int token = [srce nextToken];  // Get the next token and ..
// 
//         switch (token)                 // .. print the tokenized info
//         {
//           case DTK_EXTRA1    : printf("Directive: %s\n", [srce text]); break;
//           case DTK_COMMENT   : printf("Comment  : %s\n", [srce text]); break;
//           case DTK_NUMBER    : printf("Number   : %s\n", [srce text]); break;
//           case DTK_KEYWORD   : printf("Keyword  : %s\n", [srce text]); break;
//           case DTK_STRING    : printf("String   : %s\n", [srce text]); break;
//           case DTK_OPERATOR  : printf("Operator : %s\n", [srce text]); break;
//           default:             printf("Unkown   : %s\n", [srce text]); break;
//         }
//       }
//     }
//     else
//       printf("File \"example.tok\" could not be used by tokenizer.\n");
// 
//     [file close];
//   }
//   else
//     printf("File \"example.tok\" could not be opened:%d.\n", [file error]);
// 
//   [file free];                         // Cleanup
//   [srce free];
// 
//   return 0;
// }
// 

@interface DTokenizer : Object
{
@private
  DList       *_sources;    // the stack with the sources
  DSource     *_source;     // the source for the tokenizer
   
  DText       *_text;       // the last scanned text
  BOOL         _eoff;       // is end of file reached ?
  unsigned     _scanned;    // the length of the scanned text
  BOOL         _skipWhiteSpace; // should whitespace be skipped ? (def. YES)
}


//// Constructors

//
// Initialise a tokenizer
//
// @return the object
//
- (DTokenizer *) init;

//
// Initialise a tokenizer with a file
//
// @param source   the source
// @param name     the name for the source
// 
// @return the object
//
- (DTokenizer *) init :(id <DTextReadable>) source :(const char *) name;


//// Copy related methods

// 
// Do a shallow copy of the object (not implemented)
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the tokenizer (but *NOT* the TextReadables)
//
// @return the object
//
- free;


//// Member methods

//
// Get the last scanned text
// 
// @return the last scanned text (or NULL)
// 
- (const char *) text;

//
// Get the name of the current file
// 
// @return the name of the file (or NULL)
// 
- (const char *) name;

//
// Get the current line number
//
// @return the current line number
//
- (int) lineNumber;

//
// Get the current column number
//
// @return the current column number
//
- (int) columnNumber;

//
// Is the end of file reached ?
// 
// @return is it ?
// 
- (BOOL) isEof;

//
// Return whether white space is be skipped 
// 
// @return is it ?
// 
- (BOOL) skipWhiteSpace;

//
// Set whether white space should be skipped
// 
// @param skip          should white space be skipped
// 
// @return the object
// 
- (DTokenizer *) skipWhiteSpace :(BOOL) skip;


//// Basic scanner methods (to be overridden)

//
// Check for white space (based on isspace)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) whiteSpace :(const char *) text ;

//
// Check for an comment (start with '#' till eofl)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) comment :(const char *) text ;

//
// Check for a keyword ([A-Za-z_][A-Za-z0-9_]*)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) keyword :(const char *) text ;

//
// Check for a string (".." or '..', \" and \' allowed)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) string :(const char *) text ;

//
// Check for an operator (ispunct)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) operator :(const char *) text;

//
// Check for a number (isdigit)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) number :(const char *) text;

//
// Check for extra token 1 (to be overridden)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) extraToken1 :(const char *) text;

//
// Check for extra token 2 (to be overridden)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) extraToken2 :(const char *) text ;

//
// Check for extra token 3 (to be overridden)
// 
// @param text          the text to be scanned
// 
// @return the scanned text (or nil)
// 
- (DText *) extraToken3 :(const char *) text ;


//// Source methods

//
// Start using a source (an already open source is pushed on the sources stack)
// 
// @param source   the source
// @param name     the name for the source
// 
// @return success
// 
- (BOOL) source :(id <DTextReadable>) source :(const char *) name;

//
// Pop a source from the sources stack
// 
// @return is a source popped from the sources stack ?
// 
- (BOOL) popSource;


//// Scanner methods

//
// Scan the source for the next token (white space is skipped if
// this is active)
// 
// @return the next token
// 
- (int) nextToken ;

//
// Check the source for the next token (white space is not skipped)
//
// @return the found token
// 
- (int) checkToken;

//
// Move the scanned indicator after a checkToken method
// 
// @return the object
// 
- (DTokenizer *) next ;

//
// Move the scanned indicator a number of positions (private)
//
// @param positions  the number of position to move
//  
// @return the object
// 
- (DTokenizer *) next :(unsigned) positions;


//// Error methods

//
// Generate an error on stderr for the current file and line
// 
// @param msg      the error message
// 
// @return the object
// 
- (DTokenizer *) error :(const char *) msg;

@end
#endif
