#ifndef _DTREE_H
#define _DTREE_H

// This file was automatically generated by objch.

//==============================================================================
//
//              DTree - the tree collection in ofc-library
//
//               Copyright (C) 2003  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-05 16:29:04 $ $Revision: 1.13 $
//
//==============================================================================

#include "ofc/config.h"


typedef struct _DTreeNode
{
  id                     _object;        // the stored object
  struct _DTreeNode     *_parent;        // the parent node
  struct _DTreeNode     *_children;      // the (first) child node
  struct _DTreeNode     *_next;          // the next node 
  struct _DTreeNode     *_prev;          // the previous ode
  int                    _path;          // the path during iterating
} DTreeNode;

//
// The tree class implements a n-tree collection. It stores objects in a 
// tree of nodes. Because there is no key or index available to store and
// remove the objects, the objects can only be stored, inserted and removed 
// by the tree iterator: @class(DTreeIterator). 
//
// @example
// #include <stdio.h>
// #include "ofc/DTree.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DTree         *tree = [DTree         new  ];
//   DTreeIterator *iter = [DTreeIterator alloc];
//   DText         *str1;
//   DText         *str2;
//   int            level;
// 
//   [iter init :tree];                   // Init the iterator with the tree
// 
//   str1 = [DText new]; [str1 set :"root"];
//   [iter append :str1];                 // Set the root
//     
//   str1 = [DText new]; [str1 set :"child1"];
//   [iter append :str1];                 // Append a child to the root and ..
//   [iter parent];                       // .. move the iterator back to the root
// 
//   str1 = [DText new]; [str1 set :"child2"]; // Append again a child to the root and ..
//   [iter append :str1];                 // .. keep the iterator on the child
// 
//   str1 = [DText new]; [str1 set :"child21"]; // Append a child for the chilld
//   [iter append :str1];
// 
//   str1 = [DText new]; [str1 set :"child22"]; // Append another child
//   [iter after :str1];
// 
//   [iter parent];                       // Move iter back to level 2
// 
//   str1 = [DText new]; [str1 set :"child4"];
//   [iter after :str1];                  // Append after the child a new child
// 
//   str2 = [DText new]; [str2 set :"child3"];
//   [iter before :str2];                 // Prepend before the child a new child
// 
//   printf("Elements in the tree:%ld.\n", [tree length]);
// 
//   printf("Tree %s child3.\n", ([tree has :str2] ? "has" : "has not")); 
// 
//   level = 1;
// 
//   str1 = [iter root];                  // Move the iter to the root
//   while (str1 != nil)
//   {
//     printf("Level:%d Element:%s.\n", level, [str1 cstring]);
// 
//     if ([iter hasChildren])            // If the node has children, than move to the first child
//     {
//       level++;
//       str1 = [iter child];
//     }
//     else if ([iter isLast])            // If all children done, than move to the next of the parent
//     {
//       level--;
//       str1 = [iter parent];            
//       if (str1 != nil)
//         str1 = [iter next];
//     }
//     else
//     {
//       str1 = [iter next];              // Else next child
//     }
//   }
// 
//   [iter free];                         // Cleanup
//   [tree free];
// 
//   return 0;
// }
//

@interface DTree : Object
{
@private
  DTreeNode     *_root;          // the root of the tree
  long           _length;        // the number of nodes in the tree
}


//// Constructors

//
// Initialise an empty tree
//
// @return the object
//
- (DTree *) init ;


//// Copy related methods

//
// Do a shallow copy of the tree object
//
// @return the object
//
- shallowCopy;

//
// Deepen a copied tree object
//
// @return the object
//
- deepen;


//// Deconstructor

//
// Free the object (without the stored objects)
//
// @return the object
//
- shallowFree;

//
// (Deep) free the object (including the stored objects)
//
// @return the object
//
- free;


//// Member methods

//
// Return the number of objects in the tree
//
// @return the number of objects
//
- (long) length;

//
// Check if the tree is empty
//
// @return is it
//
- (BOOL) isEmpty;


//// Collection methods

//
// Check if the collection has a certain object
//
// @param obj      the object to be found
//
// @return has it ?
//
- (BOOL) has :(id) obj;

//
// Count the number of occurences of an object
//
// @param obj      the object to be counted
//
// @return the number of occurences
//
- (long) count :(id) obj;

//
// Perform a method for each object in the tree
//
// @param sel      the selector 
//
// @return the object
//
- (DTree *) each :(SEL) sel;

@end

//
// The tree iterator class implements an iterator on a n-tree collection. 
// Because there is no key or index available to store and remove the objects, 
// the objects can only be stored and removed by this iterator.
//
// @example
// #include <stdio.h>
// #include "ofc/DTree.h"
// #include "ofc/DText.h"
// 
// int main(int argc, char *argv[])
// {
//   DTree         *tree = [DTree         new  ];
//   DTreeIterator *iter = [DTreeIterator alloc];
//   DText         *str1;
//   DText         *str2;
//   int            level;
// 
//   [iter init :tree];                   // Init the iterator with the tree
// 
//   str1 = [DText new]; [str1 set :"root"];
//   [iter append :str1];                 // Set the root
//     
//   str1 = [DText new]; [str1 set :"child1"];
//   [iter append :str1];                 // Append a child to the root and ..
//   [iter parent];                       // .. move the iterator back to the root
// 
//   str1 = [DText new]; [str1 set :"child2"]; // Append again a child to the root and ..
//   [iter append :str1];                 // .. keep the iterator on the child
// 
//   str1 = [DText new]; [str1 set :"child21"]; // Append a child for the chilld
//   [iter append :str1];
// 
//   str1 = [DText new]; [str1 set :"child22"]; // Append another child
//   [iter after :str1];
// 
//   [iter parent];                       // Move iter back to level 2
// 
//   str1 = [DText new]; [str1 set :"child4"];
//   [iter after :str1];                  // Append after the child a new child
// 
//   str2 = [DText new]; [str2 set :"child3"];
//   [iter before :str2];                 // Prepend before the child a new child
// 
//   printf("Elements in the tree:%ld.\n", [tree length]);
// 
//   printf("Tree %s child3.\n", ([tree has :str2] ? "has" : "has not")); 
// 
//   level = 1;
// 
//   str1 = [iter root];                  // Move the iter to the root
//   while (str1 != nil)
//   {
//     printf("Level:%d Element:%s.\n", level, [str1 cstring]);
// 
//     if ([iter hasChildren])            // If the node has children, than move to the first child
//     {
//       level++;
//       str1 = [iter child];
//     }
//     else if ([iter isLast])            // If all children done, than move to the next of the parent
//     {
//       level--;
//       str1 = [iter parent];            
//       if (str1 != nil)
//         str1 = [iter next];
//     }
//     else
//     {
//       str1 = [iter next];              // Else next child
//     }
//   }
// 
//   [iter free];                         // Cleanup
//   [tree free];
// 
//   return 0;
// }
//

@interface DTreeIterator : Object
{
@private
  DTree         *_tree;          // the tree
  DTreeNode     *_node;          // the current node
}


//// Constructors

//
// Initialise an empty iterator
//
// @return the object
//
- (DTreeIterator *) init ;

//
// Initialise the iterator with a tree (iterator is moved to the root)
//
// @param tree     the tree for the iterator (<> nil)
//
// @return the object
//
- (DTreeIterator *) init :(DTree *) tree;


//// Movement methods

//
// Move to the first child (in the list of childs)
//
// @return the object of the first child (or nil)
//
- (id) first;

//
// Move to the next child (in the list of childs)
//
// @return the object of the next child (or nil)
//
- (id) next;

//
// Move to the previous child (in the list of childs)
//
// @return the object of the previous child (or nil)
//
- (id) prev;

//
// Move to the last child (in the list of childs)
//
// @return the object in the last node (or nil)
//
- (id) last;

//
// Move the iterator to a certain object. The object is searched
// from the root. If the object is not found, the iterator
// remains on the same location as before calling this method.
// If the object is found, the iterator is moved to this object.
//
// @param obj      the object to be found
//
// @return is it found ?
//
- (BOOL) move :(id) obj;

//
// Move to the root
//
// @return the object in the root (or nil)
//
- (id) root;

//
// Move to the parent (for the current child)
//
// @return the object in the parent (or nil)
//
- (id) parent;

//
// Move to the first child (for the current parent)
//
// @return the object in the first child (or nil)
//
- (id) child;


//// Test methods

//
// Test if the iterator is on the first child (in the list of childs)
//
// @return is it ?
//
- (BOOL) isFirst;

//
// Test if the iterator is on the last child (in the list of childs)
//
// @return is it ?
//
- (BOOL) isLast;

//
// Test if the iterator is on the root 
//
// @return is it ?
//
- (BOOL) isRoot;

//
// Test if the iterator is on a parent with children
//
// @return has it ?
//
- (BOOL) hasChildren;


//// Stored object methods

//
// Set the object in the current node
//
// @param obj      the object to be set
//
// @return the previous object in the node
//
- (id) object :(id) obj;

//
// Get the object in the current node
//
// @return the object in the node (or nil)
//
- (id) object;


//// Add/Insert/Remove methods

//
// Prepend a child to the current (parent) node (iterator is moved to this new child)
//
// @param obj      the object to be added as child
//
// @return the object
//
- (DTreeIterator *) prepend :(id) obj;

//
// Append a child to current (parent) node (iterator is moved to new child)
//
// @param obj      the object to be added as child
//
// @return the object
//
- (DTreeIterator *) append :(id) obj;

//
// Insert a new child before current (child) node (iterator is moved to new child)
//
// @param obj      the object to be added as child
//
// @return the object
//
- (DTreeIterator *) before :(id) obj;

//
// Insert a new child after current (child) node (iterator is moved to new child)
//
// @param obj      the object to be added as child
//
// @return the object
//
- (DTreeIterator *) after :(id) obj;

//
// Remove the current node 
// Note: only if current node has no children
// Note: iterator moves to 1. the next child or 2. the previous child or 3. the parent
//
// @return the object from the current node (or nil)
//
//
- (id) remove;

@end
#endif
