#ifndef _DUDPCLIENT_H
#define _DUDPCLIENT_H

// This file was automatically generated by objch.

//==============================================================================
//
//          DUDPClient - the UDP client class in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-11 16:59:46 $ $Revision: 1.10 $
//==============================================================================

#include "ofc/config.h"

#include "ofc/DSocketAddress.h"
#include "ofc/DSocket.h"



//
// The DUDPClient class implements a number of methods for implementing an UDP client.
//
// @example
// #include <stdio.h>
// #include "ofc/DUDPClient.h"
// #include "ofc/DSocketAddress.h"
// 
// int main(int argc, char *argv[])
// {
//   DInetSocketAddress *addr = [DInetSocketAddress new];
//   DUDPClient         *clnt = [DUDPClient         new];
//   DData              *data;
// 
//   [addr host :"localhost" :7000];
// 
//   printf("Open the client..\n");
// 
//   if ([clnt open :[addr family] :[DSocket protocol :"udp"]])
//   {
//     printf("Open connection to the server..\n");
// 
//     if ([clnt start :addr])
//     {
//       printf("Send request to the server..\n");
// 
//       data = [clnt doRequest :addr :"Hello" :5 :256];
// 
//       if (data != nil)
//       {
//         DText *str = [data toText];
// 
//         printf("Response from the server:%s.\n", [str cstring]);
// 
//         [str  free];
//         [data free];
//       }
//       else
//         printf("No data received from the server.\n");
//       
//       [clnt stop];
//     }
//     else
//       printf("Cound not open connection to the server.\n");
//   }
//   else
//     printf("Could not open the client.\n");
// 
//   [clnt free];                         // Cleanup
//   [addr free];
// 
//   return 0;
// }
// 

@interface DUDPClient : Object
{
@private
  DSocket          *_local;        // the local socket
  int               _sendFlag;     // the send flags (def. 0)
  int               _recvFlag;     // the receive flags (def. 0)
  BOOL              _started;      // is the client started ?
}


//// Constructors

//
// Initialise the UDP Client
//
// @return the object
//
- (DUDPClient *) init;

//
// Initialise the UDP Client for a family and protocol
//
// @param family       the family (DSA_AF_INET,DSA_AF_UNIX,DSA_AF_INET6)
// @param protocol     the protocol to be used
//
// @return the object
//
- (DUDPClient *) init :(int) family :(int) protocol;


//// Copy related methods

//
// Do a shallow copy of the object (not implemented)
//
// @return the objec
//
- shallowCopy;


//// Deconstructor

//
// Free the UDP client object
//
// @return the object
//
- free;


//// Member methods

//
// Return the local socket
// 
// @return a reference to the socket object
//
- (DSocket *) socket;

//
// Set the send flag
// 
// @param flag         the send flag
// 
// @return the object
// 
- (DUDPClient *) sendFlag :(int) flag;

//
// Return the send flag
// 
// @return the send flag
// 
- (int) sendFlag;

//
// Set the receive flag
// 
// @param flag         the receive flag
// 
// @return the object
// 
- (DUDPClient *) recvFlag :(int) flag;

//
// Return the receive flag
// 
// @return the receive flag
// 
- (int) recvFlag;


//// Main methods

//
// Open the UDP client
// 
// @param family       the family (DSA_AF_INET,DSA_AF_UNIX,DSA_AF_INET6)
// @param protocol     the protocol to be used
// 
// @return success
// 
- (BOOL) open :(int) family :(int) protocol;

//
// Start the communication 
// 
// @param address      the local address of the client
// 
// @return success
// 
- (BOOL) start :(id <DSocketAddressable>) address;

#ifdef DSA_AF_UNIX
#endif
//
// Send a request to a server and wait for a response (using send-
// and recvFlag)
// 
// @param server       the address of the server
// @param request      the request to be sent
// @param length       the length of the request
// @param respLength   the (maximum) length of the response
//
// @return a (new) data object with the response (or nil for error)
//
- (DData *) doRequest :(id <DSocketAddressable>) server :(const unsigned char *) request :(unsigned) length :(unsigned) respLength;

//
// Stop the communication
// 
// @return none
// 
- (void) stop;

@end
#endif
