#ifndef _DURL_H
#define _DURL_H

// This file was automatically generated by objch.

//==============================================================================
//
//         DURL - the Uniform Resource Locator class in the ofc-library
//
//               Copyright (C) 2004  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-05 16:29:04 $ $Revision: 1.5 $
//==============================================================================

#include "ofc/config.h"

#include "ofc/DText.h"
#include "ofc/DInt.h"

#include "ofc/DParsable.h"
#include "ofc/DTextable.h"

//
// The DURL class implements a number of methods for using URLs accordingly the
// RFC1738 specification. Note: this class is not a validator; the url:/fromString:
// methods check only a few rules. ToDo: relative urls, factory method for protocol
// client classes.
//
// @example
// #include <stdio.h>
// #include "ofc/DURL.h"
// 
// int main(int argc, char *argv[])
// {
//   DURL  *url1 = [DURL alloc];
//   DURL  *url2 = [DURL new  ];
//   DText *str;
// 
//   [url1 init :"http://ofc.dvoudheusden.net/index.html"]; // Init with an URL
// 
//                                        // Print the elements of the URL
//   printf("URL consists of protocol (%s), host (%s), port(%d) and path (%s).\n",
//       [url1 protocol], [url1 host], [url1 port], [url1 path]);
// 
//                                        // Build a URL
//   [url2 scheme   :"ftp"];
//   [url2 host     :"www.example.org"];
//   [url2 user     :"me"             ];
//   [url2 password :"secret"         ];
//   [url2 port     :67               ];
//   [url2 path     :"/pub"           ];
// 
//   str = [url2 url];
//   printf("URL: %s.\n", [str cstring]);
//   [str free];
// 
//   [url1 free];                         // Cleanup
//   [url2 free];
//   
//   return 0;
// }
// 

@interface DURL : Object <DTextable,DParsable>
{
@private
  DText            *_scheme;     // the scheme (or protocol) of the url
  DText            *_user;       // the username in the url
  BOOL              _noUser;     // there is no user name in the url
  DText            *_password;   // the password in the url
  BOOL              _noPassword; // there is no password in the url
  DText            *_host;       // the host
  DInt             *_port;       // the port number
  DText            *_path;       // the path
}


//// Constructors

//
// Initialise an empty URL
//
// @return the object
//
- (DURL *) init;

//
// Initialise the URL with an url
//
// @param url          the url
// 
// @return the object
//
- (DURL *) init :(const char *) url;


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the url object
//
// @return the object
//
- free;


//// Member methods

//
// Return the scheme
//
// @return the scheme
// 
- (const char *) scheme;

//
// Return the protocol (identical to the scheme)
//
// @return the scheme (or NULL)
// 
- (const char *) protocol;

//
// Set the scheme
// 
// @param scheme   the scheme for the url (or NULL)
// 
// @return the object
// 
- (DURL *) scheme :(const char *) scheme;

//
// Return the user name
// 
// @return the user name (or NULL)
// 
- (const char *) user;

//
// Set the user name
// 
// @param user     the user name (or NULL)
// 
// @return the object
// 
- (DURL *) user :(const char *) user;

//
// Return the password
// 
// @return the password (or NULL)
// 
- (const char *) password;

//
// Set the password for the url
// 
// @param password the password
// 
// @return the object
// 
- (DURL *) password :(const char *) password;

//
// Return the host
// 
// @return the host (or NULL)
// 
- (const char *) host;

//
// Set the host
// 
// @param host     the host (or NULL)
// 
// @return the object
// 
- (DURL *) host :(const char *) host;

//
// Return the port
// 
// @return the port (or 0)
// 
- (int) port;

//
// Set the port
// 
// @param port     the port (or 0)
// 
// @return the object
// 
- (DURL *) port :(int) port;

//
// Return the path of the url
// 
// @return the path (or NULL)
// 
- (const char *) path;

//
// Set the path of the url
// 
// @param path     the path (or NULL)
// 
// @return the object
// 
- (DURL *) path :(const char *) path;


//// Main methods

//
// Clear the url
// 
// @return the object
// 
- (DURL *) clear;

//
// Return the url 
// 
// @return a (new) text object with the url
//
- (DText *) url;

//
// Set the url
// 
// @param url      the url to be set
// 
// @return success
// 
- (BOOL) url :(const char *) url;

//
// Set an url with a reference url (missing scheme, user,
// password, host and port in url are used from reference)
// 
// @param url       the url to be set
// @param reference the reference url
// 
// @return success
// 
- (BOOL) url :(const char *) url :(DURL *) reference;


//// Parsable protocol

//
// Parse a string to set an url
//
// @param cstr     the string to be parsed (moved to first non-parsable char)
// 
// @return the result (0, ERANGE, ENODATA)
//
- (int) fromString :(char **) cstr;


//// Textable protocol methods

//
// Convert the url object to a text object
//
// @return the (new) text string object
//
- (DText *) toText;

@end
#endif
