#ifndef _DVALUE_H
#define _DVALUE_H

// This file was automatically generated by objch.

//==============================================================================
//
//             DValue - the value class in the ofc-library
//
//               Copyright (C) 2005  Dick van Oudheusden
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free
// Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//==============================================================================
// 
//  $Date: 2008-08-13 17:14:32 $ $Revision: 1.3 $
//
//==============================================================================

#include "ofc/config.h"

#include "ofc/DTextable.h"
#include "ofc/DText.h"
#include "ofc/DBool.h"
#include "ofc/DInt.h"
#include "ofc/DLong.h"
#include "ofc/DDouble.h"


// the value types
#define DVL_EMPTY    0
#define DVL_CLASS    1 // Protocol, category ??
#define DVL_OBJECT   2
#define DVL_SEL      3
#define DVL_BOOL     4
#define DVL_INT      5
#define DVL_LONG     6
#define DVL_DOUBLE   7
#define DVL_STRING   8

//
// The value class implements an union of basic data types. The class stores
// the current type of the storage and the value. This class also implements
// type conversion between the different types of storage.
//
// @example
// #include <stdio.h>
// #include "ofc/DValue.h"
// 
// int main(int argc, char *argv[])
// {
//   DValue *value = [DValue new];
//   DText  *str   = [DText  new];
// 
//   printf("Initial the value %s empty.\n", ([value isEmpty] ? "is" : "is not"));
// 
//   [value setInt :7];                   // Set the value to integer 7
// 
//   printf("Type of value:%s\n", [value typeString]);
// 
//   [str set :"0.07"];                  // Set the value to DText:0.07
//   [value setObject :str];
// 
//   printf("Type of value:%s\n", [value typeString]);
// 
//   printf("Double of value:%f\n", [value toDouble]); // Convert the value to a double
// 
//   printf("Type of value:%s\n", [value typeString]);
// 
//   [value free];                        // Cleanup
//   [str   free];
// 
//   return 0;
// }
//

@interface DValue : Object <DTextable>
{
@private
  int          _type;  // the value type
  union
  {
    Class      _class;  // the class
    id         _object; // the (general) object reference
    SEL        _sel;    // the selector
    BOOL       _bool;   // the boolean
    int        _int;    // the integer
    long       _long;   // the long integer
    double     _double; // the double
    DText     *_text;   // the string
  } _value;
}


//// Constructors

//
// Initialise to empty value
//
// @return the object
//
- (DValue *) init;


//// Copy related methods

//
// Do a shallow copy of the object
//
// @return the object
//
- shallowCopy;


//// Deconstructor

//
// Free the object
//
// @return the object
//
- free;


//// Member methods

//
// Empty the value
// 
// @return the object
// 
- (DValue *) empty;

//
// Check if the value is without a value
// 
// @return is the value empty ?
// 
- (BOOL) isEmpty;

//
// Return the type of the value
// 
// @return the type of the value (DVL_...)
// 
- (int) type;

//
// Return a string specifying the type of the value
// 
// @return the type string
// 
- (const char *) typeString;


//// Set methods

// 
// Set the value to a class
// 
// @param value   the class
// 
// @return the object
// 
- (DValue *) setClass :(Class) value;

// 
// Set the value to a reference of an object. The object will NOT be
// freed when the value is freed.
// 
// @param value   the reference to the object
// 
// @return the object
// 
- (DValue *) setObject :(id) value ;

// 
// Set the value to a selector
// 
// @param value   the selector
// 
// @return the object
// 
- (DValue *) setSel :(SEL) value;

// 
// Set the value to a boolean
// 
// @param value   the boolean 
// 
// @return the object
// 
- (DValue *) setBool :(BOOL) value;

// 
// Set the value to an integer
// 
// @param value   the integer
// 
// @return the object
// 
- (DValue *) setInt :(int) value;

// 
// Set the value to a long
// 
// @param value   the long
// 
// @return the object
// 
- (DValue *) setLong :(long) value;

// 
// Set the value to a double
// 
// @param value   the double
// 
// @return the object
// 
- (DValue *) setDouble :(double) value;

// 
// Set the value to a c-string
// 
// @param value   the c-string
// 
// @return the object
// 
- (DValue *) setString :(const char *) value;


//// Get methods

// 
// Return the class in the value
// 
// @return the class (or Nil)
// 
- (Class) getClass ;

// 
// Return the reference in the value
// 
// @return the object (or nil)
// 
- (id) getObject;

// 
// Return the selector in the value
// 
// @return the selector (or NULL)
// 
- (SEL) getSel;

// 
// Return the boolean in the value
// 
// @return the boolean (or NO)
// 
- (BOOL) getBool;

// 
// Return the integer in the value
// 
// @return the integer (or 0)
// 
- (int) getInt;

// 
// Return the long in the value
// 
// @return the long (or 0)
// 
- (long) getLong;

// 
// Return the double in the value
// 
// @return the double (or 0.0)
// 
- (double) getDouble;

// 
// Return the c-string in the value
// 
// @return the c-string  (or NULL)
// 
- (const char *) getString ;


//// Conversion methods

// 
// Convert the value to a class: Nil for selector, boolean, integer, long, 
// double and string.
// 
// @return the class (or Nil)
// 
- (Class) toClass ;

// 
// Convert the value to a (new) object: a copy for an object reference, nil for selector,
// DBool object for boolean, DInt object for integer, DLong object for long, DDouble object
// for double and DText for string.
// 
// @return the (new) object (or nil)
// 
- (id) toObject;

// 
// Convert the value to a selector: NULL is returned if value is not a selector
// 
// @return the selector (or NULL)
// 
- (SEL) toSel;

// 
// Convert the value to a boolean: class and sel returns NO, object tries toBool, int, long and
// double return true if not 0/NULL, string is scanned for contents
// 
// @return the boolean (or NO)
// 
- (BOOL) toBool;

// 
// Convert the value to an integer: class and sel return 0, object tries toInt,
// bool, long and double are type casted to int, string is read for an int
// 
// @return the integer (or 0)
// 
- (int) toInt;

// 
// Convert the value to a long: class and sel return 0, object tries toLong,
// bool, int and double are type casted to long, string is read for an int
// 
// @return the long (or 0)
// 
- (long) toLong;

// 
// Convert the value to a double: class and sel return 0, object tries toDouble,
// bool, int and long are type casted to double, string is read for an double
// 
// @return the double (or 0.0)
// 
- (double) toDouble;

// 
// Convert the value to a string: class and sel return nil, object tries toText,
// bool, int, long and double do toText, string is copied
// 
// @return a (new) text objecct
// 
- (DText *) toText;

@end
#endif
